#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int LINELENGTH = 80;
#define READLENGTH 30000
#define KInputs 6
static void print_string (char *str_ptr, int *column, FILE *fpout);
static int  lookforToken( char ioBuf[KInputs+1][50], char* token );
int main( int argc, char *argv[] )
{
	FILE* in;
	FILE* out;
	int length;
	int count,count2,i,j,k;
	char buffer[READLENGTH]; /* EXTRA large buffer for input and outputs......  */
	                    /* GNL is bad because it doesn't limit line length */
	char* token;
	char seps[]  = " ,\t\n\\";
	char** types;
	int* sequential;  /* 0 for combinational element, 1 for sequential */
	int clocked,in_count;
	char iobuf[KInputs+1][50]; // A buffer for the IO pins
	int found;
	/* Welcome string */
	printf( "UBC hnl2blif converter\n" );
	printf( "Usage: hnl2blif in.hnl out.blif\n" );
	if( argc < 2 )
		return;

	/* File open */
	in = fopen( argv[1], "r" );
	out = fopen( argv[2], "w" );
	
	/* First Pass, pick up the element name */
	count = 0;
	while( 1 )
	{
		if( fgets( buffer, READLENGTH, in ) == NULL )
			break; /* EOF */

		token = strtok( buffer, seps );
		/* If it was a combinational or sequential element */
		if( token )
		{
			if( strcmp( token, "combinational" ) == 0 || strcmp( token, "sequential" ) == 0 )
			{
				count++;
			}
		}
	}

	/* Second Pass, pick up the circuit itself */
	rewind(in);
	types = (char**) malloc( sizeof( char* ) * count ); /* How many pointers? */
	sequential = (int*) malloc( sizeof( int ) * count );
	for( i=0; i<count; ++i )
	{
		types[i] = malloc( 30* sizeof( char ) );  /* alloc a 30 byte memory storage */
	}
	count2 = 0;
	while( 1 )
	{
		if( fgets( buffer, READLENGTH, in ) == NULL )
			break; /* EOF */

		token = strtok( buffer, seps );
		
		if( token == NULL ) /* blank lines*/
			continue;
		
		if( strcmp( token, "circuit" ) == 0 )
		{
			token = strtok( NULL, seps );
			fprintf( out, ".model %s\n", token ); 
		}
		else if ( strcmp( token, "end" ) == 0 )
		{
			fprintf( out, ".end\n" ); /* End of the blif file */
			break; /* End of the circuit declaration */
		}
		else if (strcmp( token, "combinational" ) == 0 )
		{
			token = strtok( NULL, seps );
			strcpy( types[count2], token );  /*grab the type name*/
			sequential[count2] = 0;
			count2++;

			/*fast foward till we find the end declaration */
			strcpy( token, " " );
			while( strcmp( token, "end" ) != 0 )
			{
				fgets( buffer, READLENGTH, in );
				token = strtok( buffer, seps );
			}
		}
		else if (strcmp( token, "sequential" ) == 0 )
		{
			token = strtok( NULL, seps );
			strcpy( types[count2], token );  /* grab the type name */
			sequential[count2] = 1;
			count2++;
			clocked = 1;                     /* set a flag that we have a clock */
			/*fast foward till we find the end declaration */
			strcpy( token, " " );
			while( strcmp( token, "end" ) != 0 )
			{
				fgets( buffer, READLENGTH, in );
				token = strtok( buffer, seps );
			}
		}
		else if( strcmp( token, "input" ) == 0 )
		{
			length = 8; /* Used for the writing function */
			fprintf(out, ".inputs " );
			token = strtok( NULL, seps );  
			while( token != NULL )
			{
				print_string( token, &length, out ); /* Convert to .blif format, 80 char per line*/ 
				token = strtok( NULL, seps );
			}
			if( clocked )
				fprintf(out, "pclk\n"); /* Clocked circuit */
			else
				fprintf(out, "\n");
		}
		else if ( strcmp( token, "output" ) == 0 )
		{
			length = 9;
			fprintf(out, ".outputs ");
			token = strtok( NULL, seps );
			while( token != NULL )
			{
				print_string( token, &length, out );
				token = strtok( NULL, seps ); 
			}
			fprintf(out, "\n");
		}
		else  /* We try to match the name to a type */ 
		{
			for( i=0; i<count; ++i )
			{
				if( strcmp( token, types[i] ) == 0 ) /* We found a logic element */
				{
					// Reset buffer to 0, assume we have at most 4 inputs
					for( k=0; k<(KInputs+1); ++k )
					{
						strcpy( iobuf[k], "");
					}
					/* Print convert the net list into .blif format */ 
					k=0;
					if( sequential[i] == 0 )
					{
						/* Non sequential element */
						/* Since TVPack strips out all logic info anyways, we just make it all nand gates*/
						fprintf( out, ".names " );
						in_count = 0;
						token = strtok( NULL, seps );
						while( token != NULL )
						{
							found = lookforToken( iobuf, token );
							if( !found )
								{
								fprintf( out, "%s ", token );
								strcpy( iobuf[in_count], token );
								in_count++;
								}
							token = strtok( NULL, seps );
						}
						/* Print the mask */
						fprintf( out, "\n" );
						for( j=0; j<in_count-1; ++j )
							fprintf( out, "1" );
						fprintf( out, " 1\n" ); 
					}
					else
					{
						/* Latch element, we only want DFF, no QFF*/
						fprintf(out, ".latch " );
						token = strtok( NULL, seps );
						fprintf(out, "%s ", token ); /* Input */
						token = strtok( NULL, seps );
						fprintf(out, "%s re pclk 2\n", token ); /* Output*/
					}
					break; /* No need to continue searching if we've found it*/
				}
			}
		}
	}
	/* Clean up */
	fclose( in );
	fclose( out );
	for( i=0; i<count; ++i )
	{
		free(types[i]);
	}
	free(types);
	free(sequential);
	return 0;
}

static int  lookforToken( char ioBuf[KInputs+1][50], char* token )
{
	int i;
	for( i=0; i<(KInputs+1); ++i )
	{
		if( strcmp( ioBuf[i], token ) == 0 )
			return 1;
	}
	return 0;
}

/* FUNCTION I borrowed from TVPACK, very handy for writing blifs
*/
static void print_string (char *str_ptr, int *column, FILE *fpout) {

/* Prints string without making any lines longer than LINELENGTH.  Column  *
 * points to the column in which the next character will go (both used and *
 * updated), and fpout points to the output file.                          */

 int len;

 len = (int)strlen(str_ptr);
 if (len + 3 > LINELENGTH) {
     printf("Error in print_string: String %s is too long for desired\n"
            "maximum line length.\n", str_ptr);
     exit(1);
 }
 
 if (*column + len + 2 > LINELENGTH) {
    fprintf(fpout,"\\\n");
    *column = 1;
 }
 
 fprintf(fpout,"%s ",str_ptr);
 *column += len + 1;
}