#include <stdio.h>
#include <math.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "draw.h"
#include "Compaction.h"
#include <string.h>

static int done[8];  /* Local variable to see which sections are done or not */

typedef struct 
{
	int type_num;    /* floorplan type lookup number */
	int averageX;    /* Median X, Y */
	int averageY;    
	int num_blocks;  /* Number of blocks of this type*/
	void* next;
	int x_min,x_max;
	int y_min,y_max;
} Compaction_Data;

/* Quick count of an area for any occupied blocks */
int QuickCount( int x1, int x2, int y1, int y2 )
{
	int x,y;
	int num_occupied;
	num_occupied = 0;

	/* Count the blocks not empty */
	for( x=x1; x<=x2; x++ )
		for( y=y1; y<=y2; y++ )
	{
		if( clb[x][y].occ == 1 )
		{
			num_occupied ++;
		}
	}
	return num_occupied;
}

/* Update the median, number of blocks etc.
*/
void UpdateCompactionInfo( Compaction_Data** list, int x, int y )
{
	Compaction_Data* cur;
	int found;

	/* Variable initialization */
	cur = *list;
	found = 0;

	/* Not floorplanned, don't care */
	//if( block[clb[x][y].u.block].lookup == -1 )
	//	return;

	if( *list == NULL )  // First element
	{
		cur = ( Compaction_Data*) malloc( sizeof( Compaction_Data ) );
		cur->averageX = x;
		cur->averageY = y;
		cur->next = NULL;
		cur->num_blocks = 1;
		cur->type_num = block[clb[x][y].u.block].lookup;
		cur->x_min = x;
		cur->x_max = x;
		cur->y_min = y;
		cur->y_max = y;
		*list = cur;
	}
	else 
	{   /* At least 1 element */
		while( cur != NULL )
		{
			/* Try to find the information*/
			if( cur->type_num == block[clb[x][y].u.block].lookup ) 
			{
				cur->averageX += x;
				cur->averageY += y;
				cur->num_blocks ++;
				found = 1;

				if( x < cur->x_min )
					cur->x_min = x;
				else if( x > cur->x_max )
					cur->x_max = x;
				if( y < cur->y_min )
					cur->y_min = y;
				else if ( y > cur->y_max )
					cur->y_max = y;
			}
			cur = cur->next;
		}

		if( !found )
		{
			/* bug fix, have to traverse to the end of the list again */
			cur = *list;
			while( cur->next != NULL )
				cur = cur->next;

			/* if not found, append an element to the end */
			cur->next = ( Compaction_Data*) malloc( sizeof( Compaction_Data ) );
			cur = cur->next;
			cur->averageX = x;
			cur->averageY = y;
			cur->next = NULL;
			cur->num_blocks = 1;
			cur->type_num = block[clb[x][y].u.block].lookup;
			cur->x_min = x;
			cur->x_max = x;
			cur->y_min = y;
			cur->y_max = y;
		}
	}
}
/* Counts the number of blocks occupied inclusive x1 <= x <= x2, y1 <= y <= y2
*/
int CountNumBlocksInArea(int x1, int x2, int y1, int y2, Compaction_Data** types)
{
	/* Variables */
	int x,y;
	int num_occupied;
	Compaction_Data* first;
	Compaction_Data* cur;

	/* Variable initialization*/
	num_occupied = 0;
	first = NULL;

	/* Count the blocks not empty */
	for( x=x1; x<=x2; x++ )
		for( y=y1; y<=y2; y++ )
	{
		if( clb[x][y].occ == 1 )
		{
			num_occupied ++;

			UpdateCompactionInfo( &first, x, y );
		}
	}

	/* Set all the average numbers */
	if( first != NULL )
	{
		cur = first;
		while( cur != NULL )
		{
			cur->averageX /= cur->num_blocks;
			cur->averageY /= cur->num_blocks;
			cur = cur->next;
		}
	}
	/* Return */
	*types = first;
	return num_occupied;
}

/* Free a Compaction data list*/
void FreeCompactionDataList( Compaction_Data* list )
{
	Compaction_Data* cur;
	Compaction_Data* next;
	cur = list;
	next = NULL;

	while( cur != NULL )
	{
		next = cur->next;
		free( cur );
		cur = next;
	}
}

/* Check all 8 regions for any stranglers */
int Is_sGrid_Done(struct s_placer_opts* p_opts, int* num_outliers )
{
	int count;
	int num;
	count = 0;

	/*   Effective compaction could be subdivided into 8 regions 
	*    012
	*    7x3
	*    654
	*/

	/* 0 */
	num = QuickCount(1,p_opts->grid_x1,1,p_opts->grid_y1);                                   
	if(num) done[0] = 0; else  done[0] = 1;
	count += num;
	
	/* 1 */
	num = QuickCount(p_opts->grid_x1+1,p_opts->grid_x2,1,p_opts->grid_y1);                                   
	if(num) done[1] = 0; else  done[1] = 1;
	count += num;
	
	/* 2 */
	num = QuickCount(p_opts->grid_x2+1, nx, 1, p_opts->grid_y1);                                   
	if(num) done[2] = 0; else done[2] = 1;
	count += num;
	
	/* 3 */
	num = QuickCount(p_opts->grid_x2+1, nx, p_opts->grid_y1+1, p_opts->grid_y2);                                   
	if(num) done[3] = 0; else done[3] = 1;
	count += num;
	
	/* 4 */
	num = QuickCount(p_opts->grid_x2+1, nx, p_opts->grid_y2+1, ny);                                   
	if(num) done[4] = 0; else done[4] = 1;
	count += num;
	
	/* 5 */
	num = QuickCount(p_opts->grid_x1+1, p_opts->grid_x2, p_opts->grid_y2+1, ny);                                   
	if(num) done[5] = 0; else done[5] = 1;
	count += num;
	
	/* 6 */
	num = QuickCount(1,p_opts->grid_x1,p_opts->grid_y2+1,ny);                                   
	if(num) done[6] = 0; else done[6] = 1;
	count += num;
	
	/* 7 */
	num = QuickCount(1,p_opts->grid_x1, p_opts->grid_y1+1, p_opts->grid_y2 );                                   
	if(num) done[7] = 0; else done[7] = 1;
	count += num;
	
	/* Optionally return the number of outliers*/
	if( num_outliers != NULL )
		{
		*num_outliers = count;
		}
	/* Are we done? */
	if( count != 0 )
		return 0;
	else 
		return 1;
}

/* Try to find a target */
int FindTarget()
{
	int rand;
	int found;
	found = 0;

	do
	{
		rand= my_irand(7);
		if( !done[rand] )
			found = 1;
	} while( !found );

	return rand;
}

/* 4 functions to compact the blocks to gain empty space.
*/
void CompactDown( struct s_placer_opts* p_opts )
{
	/* Algorithm, loop through all columns one by one */
	/* Group the blocks and replace them compacted */
	int x,y,i,j;
	int* blocks;
	
	/* Local storage for block numbers*/
	blocks = (int*) malloc( p_opts->norm_Y * sizeof(int) ); 
	
	/* Horizontal, one column at a time */
	/* O(n^2) any better? */
	for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
	{
		i=0; j=0;
		/* First is to get a list of blocks occupying this column */
		for( y=p_opts->grid_y2; y > p_opts->grid_y1; --y )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;  /* i keeps a running count of the number of elements that are occupied */
			}
		}
		
		/* Now that we know how many blocks we have, we retile them compacted */
		if( i != p_opts->norm_Y )
		{
			/* Only retile for columns w/ empty space, don't waste */
			for( y=p_opts->grid_y2; y>p_opts->grid_y1; --y )
			{
				if( j < i )
				{
					/* Relocate the clb */
					clb[x][y].occ = 1;
					clb[x][y].u.block = blocks[j];
					block[blocks[j]].x = x;
					block[blocks[j]].y = y;
					j++; 
				}
				else
				{
					/* Block is now empty */
					clb[x][y].occ = 0;
					clb[x][y].u.block = -1;
				}
			}
		}
	}

	/* Cleanup */
	free( blocks );

	//update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}

void CompactUp( struct s_placer_opts* p_opts )
{
	/* Algorithm, loop through all columns one by one */
	/* Group the blocks and replace them compacted */
	int x,y,i,j;
	int* blocks;
	
	/* Local storage for block numbers*/
	blocks = (int*) malloc( p_opts->norm_Y * sizeof(int) ); 
	
	/* Horizontal, one column at a time */
	/* O(n^2) any better? */
	for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
	{
		i=0; j=0;
		/* First is to get a list of blocks occupying this column */
		for( y=p_opts->grid_y1+1; y <= p_opts->grid_y2; ++y )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;  /* i keeps a running count of the number of elements that are occupied */
			}
		}
		
		/* Now that we know how many blocks we have, we retile them compacted */
		if( i != p_opts->norm_Y )
		{
			/* Only retile for columns w/ empty space, don't waste */
			for( y=p_opts->grid_y1+1; y <= p_opts->grid_y2; ++y )
			{
				if( j < i )
				{
					/* Relocate the clb */
					clb[x][y].occ = 1;
					clb[x][y].u.block = blocks[j];
					block[blocks[j]].x = x;
					block[blocks[j]].y = y;
					j++ ;
				}
				else
				{
					/* Block is now empty */
					clb[x][y].occ = 0;
					clb[x][y].u.block = -1;
				}
			}
		}
	}

	/* Cleanup */
	free( blocks );

	//update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}

void CompactLeft( struct s_placer_opts* p_opts )
{
	/* Algorithm, loop through all columns one by one */
	/* Group the blocks and replace them compacted */
	int x,y,i,j;
	int* blocks;
	
	/* Local storage for block numbers*/
	blocks = (int*) malloc( p_opts->norm_X * sizeof(int) ); 
	
	/* Vertical, one column at a time */
	/* O(n^2) any better? */
	for( y=p_opts->grid_y1+1; y<=p_opts->grid_y2; ++y )
	{
		i=0; j=0;
		/* First is to get a list of blocks occupying this column */
		for( x=p_opts->grid_x1+1; x <= p_opts->grid_x2; ++x )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;  /* i keeps a running count of the number of elements that are occupied */
			}
		}
		
		/* Now that we know how many blocks we have, we retile them compacted */
		if( i != p_opts->norm_X )
		{
			/* Only retile for columns w/ empty space, don't waste */
			for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
			{
				if( j < i )
				{
					/* Relocate the clb */
					clb[x][y].occ = 1;
					clb[x][y].u.block = blocks[j];
					block[blocks[j]].x = x;
					block[blocks[j]].y = y;
					j++; 
				}
				else
				{
					/* Block is now empty */
					clb[x][y].occ = 0;
					clb[x][y].u.block = -1;
				}
			}
		}
	}

	/* Cleanup */
	free( blocks );

	//update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}

void CompactRight( struct s_placer_opts* p_opts )
{
	/* Algorithm, loop through all columns one by one */
	/* Group the blocks and replace them compacted */
	int x,y,i,j;
	int* blocks;
	
	/* Local storage for block numbers*/
	blocks = (int*) malloc( p_opts->norm_X * sizeof(int) ); 
	
	/* Vertical, one column at a time */
	/* O(n^2) any better? */
	for( y=p_opts->grid_y1+1; y<=p_opts->grid_y2; ++y )
	{
		i=0; j=0;
		/* First is to get a list of blocks occupying this column */
		for( x=p_opts->grid_x2; x > p_opts->grid_x1; --x )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;  /* i keeps a running count of the number of elements that are occupied */
			}
		}
		
		/* Now that we know how many blocks we have, we retile them compacted */
		if( i != p_opts->norm_X )
		{
			/* Only retile for columns w/ empty space, don't waste */
			for( x=p_opts->grid_x2; x > p_opts->grid_x1; --x )
			{
				if( j < i )
				{
					/* Relocate the clb */
					clb[x][y].occ = 1;
					clb[x][y].u.block = blocks[j];
					block[blocks[j]].x = x;
					block[blocks[j]].y = y;
					j++; 
				}
				else
				{
					/* Block is now empty */
					clb[x][y].occ = 0;
					clb[x][y].u.block = -1;
				}
			}
		}
	}

	/* Cleanup */
	free( blocks );

	//update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}

/* Compact left and right as specified by the cut location */
/* updown is do we start at top or bottom? 1 is down, 0 is up */
void SplitCompactHorizontal( int updown, Compaction_Data* list, struct s_placer_opts* p_opts, int* num )
{
	/* Local Variables */
	int x,y,i,j,k;
	int* blocks;
	int numfree, done;
	int temp;

	/* Initialization*/
	blocks = (int*) malloc( p_opts->norm_X * sizeof(int));
	numfree = 0;
	done = 0;
	
	//for( y=p_opts->grid_y1+1; y<=p_opts->grid_y2; ++y )
	if(updown) y=p_opts->grid_y1+1; else y=p_opts->grid_y2;
	while( !done )
	{
		/* Compact left */
		i=0; j=0;
		for( x=p_opts->grid_x1+1; x<=list->averageX; ++x )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;
			}
		}
		
		/* Retile the blocks compacted */
		for( x=p_opts->grid_x1+1; x<=list->averageX; ++x )
		{
			if( j<i)
			{
				clb[x][y].occ = 1;
				clb[x][y].u.block = blocks[j];
				block[blocks[j]].x = x;
				block[blocks[j]].y = y;
				++j;
			}
			else
			{
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
			}
		}

		/* Compact right */
		i=0; j=0;
		for( x=p_opts->grid_x2; x>list->averageX; --x )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;
			}
		}

		/* Retile the blocks compacted */
		for( x=p_opts->grid_x2; x>list->averageX; --x )
		{
			if( j<i )
			{
				clb[x][y].occ = 1;
				clb[x][y].u.block = blocks[j];
				block[blocks[j]].x = x;
				block[blocks[j]].y = y;
				++j;
			}
			else
			{
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
			}
		}
		/* Count the number free in this row */
		temp = 0;
		//for(k=list->x_min; k<=list->x_max; ++k)
		for(k=p_opts->grid_x1+1; k<=p_opts->grid_x2; ++k)
		{
			if( clb[k][y].occ == 0 )
			{
				temp ++;
			}
		}
		if( temp > 0 )
		{
			numfree += temp;
			if( numfree >= list->num_blocks  )
			{   /* Enough blocks to fit this element */
			//	done = 1;
			}
		}
		else
		{
			/* This row was 100% compacted!! not enough room in target area */
			done = 1;
		}

		/* Exit condition */
		if(updown) y++; else y--;
		if( y==p_opts->grid_y1 || y==p_opts->grid_y2+1 )
		{
			done = 1;
		}
	}

	/* Cleanup */
	free( blocks );
	*num = numfree;

	done = 0;
	if( done ) 
		update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}


/* Split compaction up/down at a cut location */
/* left right, 1 is start from left, 0 is start from right */
void SplitCompactVertical( int leftright, Compaction_Data* list, struct s_placer_opts* p_opts, int* num )
{
	/* Local Variables */
	int x,y,i,j,k;
	int* blocks;
	int numfree, done;
	int temp;

	/* Initialization*/
	blocks = (int*) malloc( p_opts->norm_X * sizeof(int));
	numfree = 0;
	done = 0;

	//for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
	if(leftright) x=p_opts->grid_x1+1; else x=p_opts->grid_x2;
	while(!done)
	{
		/* Compact up */
		i=0; j=0;
		for( y=p_opts->grid_y1+1; y<=list->averageY; ++y )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;
			}
		}
		
		/* Retile the blocks compacted */
		for( y=p_opts->grid_y1+1; y<=list->averageY; ++y )
		{
			if( j<i)
			{
				clb[x][y].occ = 1;
				clb[x][y].u.block = blocks[j];
				block[blocks[j]].x = x;
				block[blocks[j]].y = y;
				++j;
			}
			else
			{
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
			}
		}

		/* Compact down */
		i=0; j=0;
		for( y=p_opts->grid_y2; y>list->averageY; --y )
		{
			if( clb[x][y].occ == 1 )
			{
				blocks[i] = clb[x][y].u.block;
				++i;
			}
		}

		/* Retile the blocks compacted */
		for( y=p_opts->grid_y2; y>list->averageY; --y )
		{
			if( j<i )
			{
				clb[x][y].occ = 1;
				clb[x][y].u.block = blocks[j];
				block[blocks[j]].x = x;
				block[blocks[j]].y = y;
				++j;
			}
			else
			{
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
			}
		}
		/* Count the number free in this row */
		temp = 0;
		//for(k=list->y_min; k<=list->y_max; ++k)
		for(k=p_opts->grid_y1+1; k<=p_opts->grid_y2; ++k)
		{
			if( clb[x][k].occ == 0 )
			{
				temp ++;
			}
		}
		if( temp > 0 )
		{
			numfree += temp;
			if( numfree >= list->num_blocks || done )
			{   /* Enough blocks to fit this element */
			//	done = 1;
			}
		}
		else
		{
			/* This row was 100% compacted!! not enough room in target area */
			done = 1;
		}

		if(leftright) x++; else x--;
		if( x==p_opts->grid_x1 || x==p_opts->grid_x2+1)
		{
			done = 1;
		}
	}

	/* Cleanup */
	free( blocks );
	*num = numfree;
    
	done = 0;
	if( done )
		update_screen(MAJOR, "Test", PLACEMENT, FALSE);
}

/*
*   Insertion algorithm for the 8 areas
*/
void InsertBlocks0( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks1( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks2( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks3( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks4( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks5( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks6( struct s_placer_opts* p_opts, Compaction_Data* area );
void InsertBlocks7( struct s_placer_opts* p_opts, Compaction_Data* area );
void FindHome( int bb_x, int bb_x2, int bb_y, int bb_y2, 
			   int x_from, int y_from, struct s_placer_opts* p_opts );

/* The actual compaction algorithm 
   Do not forget to run a re-timing after a compaction!
*/
void DoCompaction( struct s_placer_opts* p_opts )
{
	/* Variables */
	int rand;
	Compaction_Data* area;
	Compaction_Data dummy;
    int num_free;
	int num_needed;

	/* Now we can try to compact one of the 8 regions */
	rand = FindTarget();

	/*   Effective compaction could be subdivided into 8 regions 
	*    012
	*    7x3
	*    654
	*/

	/* Now try to compact the 8 cases */
	switch( rand )
		{
		case 0:
			num_needed = CountNumBlocksInArea(1,p_opts->grid_x1,1,p_opts->grid_y1,&area);
			CompactDown(p_opts);
			dummy.averageX = p_opts->grid_x1+1;
			dummy.num_blocks = num_needed;
			SplitCompactHorizontal(1,&dummy,p_opts,&num_free);
			InsertBlocks0( p_opts, area );
		break;

		case 1:
			CountNumBlocksInArea(p_opts->grid_x1+1,p_opts->grid_x2,1,p_opts->grid_y1,&area);
			CompactDown(p_opts);
			SplitCompactHorizontal(1, area,p_opts,&num_free);
			InsertBlocks1( p_opts, area );
		break;

		case 2:
			num_needed = CountNumBlocksInArea(p_opts->grid_x2+1, nx, 1, p_opts->grid_y1,&area);
			CompactLeft(p_opts);
			dummy.averageY = p_opts->grid_y1+1;
			dummy.num_blocks = num_needed;
			SplitCompactVertical(0,&dummy,p_opts,&num_free);
			InsertBlocks2( p_opts, area );
		break;
		
		case 3:
			CountNumBlocksInArea(p_opts->grid_x2+1, nx, p_opts->grid_y1+1, p_opts->grid_y2,&area);
			CompactLeft(p_opts);
			SplitCompactVertical(0,area,p_opts,&num_free);
			InsertBlocks3( p_opts, area );
		break;

		case 4:
			num_needed = CountNumBlocksInArea(p_opts->grid_x2+1, nx, p_opts->grid_y2+1, ny,&area);
			CompactUp(p_opts);
			dummy.averageX = p_opts->grid_x2;
			dummy.num_blocks = num_needed;
			SplitCompactHorizontal(0,&dummy,p_opts,&num_free);
			InsertBlocks4( p_opts, area );
		break;

		case 5:
			CountNumBlocksInArea(p_opts->grid_x1+1, p_opts->grid_x2, p_opts->grid_y2+1, ny,&area);
			CompactUp(p_opts);
			SplitCompactHorizontal(0,area,p_opts,&num_free);
			InsertBlocks5( p_opts, area );
		break;

		case 6:
			num_needed = CountNumBlocksInArea(1,p_opts->grid_x1,p_opts->grid_y2+1,ny,&area);
			CompactRight(p_opts);
			dummy.averageY = p_opts->grid_y2;
			dummy.num_blocks = num_needed;
			SplitCompactVertical(1,&dummy,p_opts,&num_free);
			InsertBlocks6( p_opts, area );
		break;
		
		case 7:
			CountNumBlocksInArea(1,p_opts->grid_x1, p_opts->grid_y1+1, p_opts->grid_y2,&area);
			CompactRight(p_opts);
			SplitCompactVertical(1,area,p_opts,&num_free);
			InsertBlocks7( p_opts, area );
		break;

		default:
		printf("ERROR\n");
		break;
		}

		//update_screen(MAJOR, "Test", PLACEMENT, FALSE);

		/* Cleanup */
		FreeCompactionDataList( area );
}

/* Actual implementation for the insertion functions 
*
*  The Layout is as follows:
*  012
*  7x3
*  654
*/
/* Compacted down, then to the right */
void InsertBlocks0( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	int x,y;  /* X and Y of where we are from*/
	int i,j;  /* I. J is where we are going */
	
	/* Set up where we want to look */
	i=p_opts->grid_x1+1;
	j=p_opts->grid_y1+1;

	/* Check top left hand corner */
	for( y=1; y<=p_opts->grid_y1; ++y )
	{
		for( x=1; x<=p_opts->grid_x1; ++x )
		{
			/* If the CLB was occupied, then we find it a new home */
			if( clb[x][y].occ == 1 )
			{
				
				while( clb[i][j].occ == 1 )
				{
					++i;
					if( i > p_opts->grid_x2 )
					{   /* Splill over to next row */
						j++;
						i = p_opts->grid_x1 + 1;
					}
				}
#ifdef _DEBUG
				if( clb[i][j].occ != 0 )
					printf("Overwriting a NON-empty clb\n");
#endif 
				
				/* Now perform the move */
				clb[i][j].occ = 1;
				clb[i][j].u.block = clb[x][y].u.block;
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
				block[clb[i][j].u.block].x = i;
				block[clb[i][j].u.block].y = j;
				i++;

				if( i > p_opts->grid_x2 )
					{
						/* Splill over to next row */
						j++;
						i = p_opts->grid_x1 + 1;
					}
			}
		}
	}
}

/* Compacted down and then horizontal split */
void InsertBlocks1( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	/* Variables */
	int x, y;  /* Where we are from */
	int i, j;  /* Where we are going */
	int bb_x, bb_y, bb_x2, bb_y2;
	int num_free;

	/* Initialization */
	bb_x = area->averageX; 
	bb_y = p_opts->grid_y1 + 1;  /* TOP section */
	bb_x2 = 0;
	bb_y2 = 0;
	num_free = 0;

	/* Try to create the smallest bounding box 
	possible to fit the blocks around the average X */
	do
	{
		/* Increase our search size by 2 each time*/
		bb_x2 +=2;
		bb_y2 +=2;
		bb_x --;
		/* Do a recount, bounds check!!! */
		num_free = 0;
		for( i=bb_x; i<bb_x+bb_x2; ++i )
			for( j=bb_y; j<bb_y+bb_y2; ++j )
			{
				/* OOB checking*/
				if( i<= p_opts->grid_x1 || i> p_opts->grid_x2 ||
					j<= p_opts->grid_y1 || j> p_opts->grid_y2 )
				{
					continue;
				}
				
				if( clb[i][j].occ == 0 )
					num_free ++;
			}
		
	} while( num_free < area->num_blocks );
	
	/* Fit the blocks of that type into that Bounding box */
	for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
		for( y=1; y<= p_opts->grid_y1; ++y )
		{
			/* If it was occupied and the type was correct */
			if( clb[x][y].occ == 1  )
			{
				if( block[clb[x][y].u.block].lookup == area->type_num )
				{
					FindHome( bb_x, bb_x2, bb_y, bb_y2, x, y, p_opts );
				}
			}
		}
}

/* Compacted left, then down */
void InsertBlocks2( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	int x,y;  /* X and Y of where we are from*/
	int i,j;  /* I. J is where we are going */
	
	/* Set up where we want to look */
	i=p_opts->grid_x2;
	j=p_opts->grid_y1+1;

	/* Check top right hand corner */
	for( y=0; y<=p_opts->grid_y1; ++y )
	{
		for( x=nx; x>p_opts->grid_x2; --x )
		{
			/* If the CLB was occupied, then we find it a new home */
			if( clb[x][y].occ == 1 )
			{
				while( clb[i][j].occ == 1 )
				{
					--i;
					if( i <= p_opts->grid_x1 )
					{   /* Splill over to next row */
						j++;
						i = p_opts->grid_x2;
					}
				}
#ifdef _DEBUG
				if( clb[i][j].occ != 0 )
					printf("Overwriting a NON-empty clb\n");
#endif 
				
				/* Now perform the move */
				clb[i][j].occ = 1;
				clb[i][j].u.block = clb[x][y].u.block;
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
				block[clb[i][j].u.block].x = i;
				block[clb[i][j].u.block].y = j;
				i--;
				if( i <= p_opts->grid_x1 )
					{   /* Splill over to next row */
						j++;
						i = p_opts->grid_x2;
					}
			}
		}
	}
}

void InsertBlocks3( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	/* Variables */
	int x, y;  /* Where we are from */
	int i, j;  /* Where we are going */
	int bb_x, bb_y, bb_x2, bb_y2;
	int num_free;

	/* Initialization */
	bb_x = p_opts->grid_x2; /* right hand side */ 
	bb_y = area->averageY;  
	bb_x2 = 0;
	bb_y2 = 0;
	num_free = 0;

	/* Try to create the smallest bounding box 
	possible to fit the blocks around the average X */
	do
	{
		/* Increase our search size by 2 each time*/
		bb_x2 +=2;
		bb_y2 +=2;
		bb_y --;
		/* Do a recount, bounds check!!! */
		num_free = 0;
		for( i=bb_x; i>bb_x-bb_x2; --i )
			for( j=bb_y; j<bb_y+bb_y2; ++j )
			{
				/* OOB checking*/
				if( i<= p_opts->grid_x1 || i> p_opts->grid_x2 ||
					j<= p_opts->grid_y1 || j> p_opts->grid_y2 )
				{
					continue;
				}
				
				if( clb[i][j].occ == 0 )
					num_free ++;
			}
		
	} while( num_free < area->num_blocks );
	
	/* Fit the blocks of that type into that Bounding box */
	for( x=p_opts->grid_x2+1; x<=nx; ++x )
		for( y=p_opts->grid_y1+1; y<=p_opts->grid_y2; ++y )
		{
			/* If it was occupied and the type was correct */
			if( clb[x][y].occ == 1  )
			{
				if( block[clb[x][y].u.block].lookup == area->type_num )
				{
					FindHome( bb_x-bb_x2+1, bb_x2, bb_y, bb_y2, x, y, p_opts );
				}
			}
		}
}

/* Bottom right hand corner
*  Compacted Up then to the left 
*/
void InsertBlocks4( struct s_placer_opts* p_opts, Compaction_Data* area)
{
	int x,y;  /* X and Y of where we are from*/
	int i,j;  /* I. J is where we are going */
	
	/* Set up where we want to look */
	i=p_opts->grid_x2;
	j=p_opts->grid_y2;

	/* Check bottom right hand corner */
	for( y=ny; y>p_opts->grid_y2; --y )
	{
		for( x=nx; x>p_opts->grid_x2; --x )	
		{
			/* If the CLB was occupied, then we find it a new home */
			if( clb[x][y].occ == 1 )
			{
				while( clb[i][j].occ == 1 )
				{
					--i;
					if( i <= p_opts->grid_x1 )
					{   /* Splill over to next row */
						j--;
						i = p_opts->grid_x2;
					}
				}
#ifdef _DEBUG
				if( clb[i][j].occ != 0 )
					printf("Overwriting a NON-empty clb\n");
#endif 
				
				/* Now perform the move */
				clb[i][j].occ = 1;
				clb[i][j].u.block = clb[x][y].u.block;
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
				block[clb[i][j].u.block].x = i;
				block[clb[i][j].u.block].y = j;
				i--;
				if( i <= p_opts->grid_x1 )
					{   /* Splill over to next row */
						j--;
						i = p_opts->grid_x2;
					}
			}
		}
	}
}

/* Bottom center coordinates*/
void InsertBlocks5( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	/* Variables */
	int x, y;  /* Where we are from */
	int i, j;  /* Where we are going */
	int bb_x, bb_y, bb_x2, bb_y2;
	int num_free;

	/* Initialization */
	bb_x = area->averageX; 
	bb_y = p_opts->grid_y2;  /* BOTTOM section */
	bb_x2 = 0;
	bb_y2 = 0;
	num_free = 0;

	/* Try to create the smallest bounding box 
	possible to fit the blocks around the average X */
	do
	{
		/* Increase our search size by 2 each time*/
		bb_x2 +=2;
		bb_y2 +=2;
		bb_x --;
		/* Do a recount, bounds check!!! */
		num_free = 0;
		for( i=bb_x; i<bb_x+bb_x2; ++i )
			for( j=bb_y; j>bb_y-bb_y2; --j )
			{
				/* OOB checking*/
				if( i<= p_opts->grid_x1 || i> p_opts->grid_x2 ||
					j<= p_opts->grid_y1 || j> p_opts->grid_y2 )
				{
					continue;
				}
				
				if( clb[i][j].occ == 0 )
					num_free ++;
			}
		
	} while( num_free < area->num_blocks );
	
	/* Fit the blocks of that type into that Bounding box */
	for( x=p_opts->grid_x1+1; x<=p_opts->grid_x2; ++x )
		for( y=p_opts->grid_y2+1; y<=ny; ++y )
		{
			/* If it was occupied and the type was correct */
			if( clb[x][y].occ == 1  )
			{
				if( block[clb[x][y].u.block].lookup == area->type_num )
				{
					FindHome( bb_x, bb_x2, bb_y-bb_y2+1, bb_y2, x, y, p_opts );
				}
			}
		}
}

/* Bottom left hand corner 
*  Compacted right, then up 
*/
void InsertBlocks6( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	int x,y;  /* X and Y of where we are from*/
	int i,j;  /* I. J is where we are going */
	
	/* Set up where we want to look */
	i=p_opts->grid_x1+1;
	j=p_opts->grid_y2;

	/* Check bottom left hand corner */
	for( y=ny; y>p_opts->grid_y2; --y )		
	{
		for( x=1; x<=p_opts->grid_x1; ++x )
		{
			/* If the CLB was occupied, then we find it a new home */
			if( clb[x][y].occ == 1 )
			{
				while( clb[i][j].occ == 1 )
				{
					++i;
					if( i > p_opts->grid_x2 )
					{   /* Splill over to next row */
						j--;
						i = p_opts->grid_x1 + 1;
					}
				}
#ifdef _DEBUG
				if( clb[i][j].occ != 0 )
					printf("Overwriting a NON-empty clb\n");
#endif 
				
				/* Now perform the move */
				clb[i][j].occ = 1;
				clb[i][j].u.block = clb[x][y].u.block;
				clb[x][y].occ = 0;
				clb[x][y].u.block = -1;
				block[clb[i][j].u.block].x = i;
				block[clb[i][j].u.block].y = j;
				i++;
				if( i > p_opts->grid_x2 )
					{   /* Splill over to next row */
						j--;
						i = p_opts->grid_x1 + 1;
					}
			}
		}
	}
}

void InsertBlocks7( struct s_placer_opts* p_opts, Compaction_Data* area )
{
	/* Variables */
	int x, y;  /* Where we are from */
	int i, j;  /* Where we are going */
	int bb_x, bb_y, bb_x2, bb_y2;
	int num_free;

	/* Initialization */
	bb_x = p_opts->grid_x1+1; 
	bb_y = area->averageY;  /* Left section */
	bb_x2 = 0;
	bb_y2 = 0;
	num_free = 0;

	/* Try to create the smallest bounding box 
	possible to fit the blocks around the average X */
	do
	{
		/* Increase our search size by 2 each time*/
		bb_x2 +=2;
		bb_y2 +=2;
		bb_y --;
		/* Do a recount, bounds check!!! */
		num_free = 0;
		for( i=bb_x; i<bb_x+bb_x2; ++i )
			for( j=bb_y; j<bb_y+bb_y2; ++j )
			{
				/* OOB checking*/
				if( i<= p_opts->grid_x1 || i> p_opts->grid_x2 ||
					j<= p_opts->grid_y1 || j> p_opts->grid_y2 )
				{
					continue;
				}
				
				if( clb[i][j].occ == 0 )
					num_free ++;
			}
		
	} while( num_free < area->num_blocks );
	
	/* Fit the blocks of that type into that Bounding box */
	for( x=1; x<=p_opts->grid_x1; ++x )
		for( y=p_opts->grid_y1+1; y<=p_opts->grid_y2; ++y )
		{
			/* If it was occupied and the type was correct */
			if( clb[x][y].occ == 1  )
			{
				if( block[clb[x][y].u.block].lookup == area->type_num )
				{
					FindHome( bb_x, bb_x2, bb_y, bb_y2, x, y, p_opts );
				}
			}
		}
}

/*  Find a home for a block located at clb[x_from][y_from]
*   bb_x <= x < bb_x+bb_x2
*   bb_y <= y < bb_y+bb_y2
*/
void FindHome( int bb_x, int bb_x2, int bb_y, int bb_y2, 
			   int x_from, int y_from, struct s_placer_opts* p_opts  )
{
	/* Variables */
	int found;
	int x,y;

	/* Initialization*/
	found = 0;
	
	while( !found )
	{
		x = my_irand( bb_x2-1 );
		y = my_irand( bb_y2-1 );
		
		/* OOB checking*/
		if( bb_x+x<= p_opts->grid_x1 || bb_x+x> p_opts->grid_x2 ||
			bb_y+y<= p_opts->grid_y1 || bb_y+y> p_opts->grid_y2 )
		{
			found = 0;
		}
		if( clb[bb_x+x][bb_y+y].occ == 0 )
		{
			found = 1;

			/* Do the move */
			clb[bb_x+x][bb_y+y].u.block = clb[x_from][y_from].u.block;
			clb[bb_x+x][bb_y+y].occ = 1;
			clb[x_from][y_from].u.block = -1;
			clb[x_from][y_from].occ = 0;
			block[clb[bb_x+x][bb_y+y].u.block].x = bb_x+x;
			block[clb[bb_x+x][bb_y+y].u.block].y = bb_y+y;
		}
	}	
}

