/*
*   Functions related to Supergrid are located here
*   The majority of the functions are in place.c
*   Should be moved to this file when I have time
*
*   Written by: David Leong davel@shaw.ca
*
*/

/**
*  <<INCLUDES section>> 
*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "draw.h"
#include "place.h"
#include "SuperGrid.h"

/*
* Function Implementations 
*/

/*  Function implementation for dumping a placement from super grid to normal location
*
*/
void DumpSuperGridPlacement( char* place_file, struct s_placer_opts* p_opts )
{
	/* Variables */
	int i,x,y;
	struct s_block* cur;
	FILE* fp;
	fp = fopen( place_file, "w+" );

	/* Print out NX and NY
	*/
	fprintf(fp, ".SIZE %i %i\n", p_opts->norm_X, p_opts->norm_Y); 

	/* Loop through all CLBs and print their locations
	*/
	for( i=0; i<num_blocks; ++i )
	{
		cur = &block[i];
		if( clb[cur->x][cur->y].type == IO )
		{
			/* Convert IO */
			if( cur->x == 0 )
			{
				x= 0;
				y= cur->y - p_opts->grid_y1;
			}
			else if( cur->x == nx+1 )
			{
				x = p_opts->norm_X + 1;
				y = cur->y - p_opts->grid_y1;
			}
			else if ( cur->y == 0 )
			{
				x = cur->x - p_opts->grid_x1;
				y = 0;
			} 
			else /* cur->y == ny+1 */
			{
				x = cur->x - p_opts->grid_x1;
				y = p_opts->norm_Y + 1;
			}
			fprintf(fp, "%i %i\n", x, y ); 
		}
		else
		{	/* Convert CLB */
			x = cur->x - p_opts->grid_x1;
			y = cur->y - p_opts->grid_y1;
			fprintf(fp, "%i %i\n", x, y );

			if( x<=0 || x>=nx+1 || y<=0 || y>nx+1 )
			{
				printf("Out of bounds\n");
			}
		}
	}
	fclose( fp );
}

/* Function implementation for reading a super grid placement to normal location
*/
void ReadSuperGridPlacement( char* place_file )
{
	/* Variables */
	int i,j,tempInt, tempInt2;
	char buf[BUFSIZE];
	char* token;
	char seps[]   = " ,\t\n";
	int found;
	FILE* fp;
	fp = fopen( place_file, "r" );

	/* Set all initial occupancies to 0 first */
	for (i=0;i<=nx+1;i++) 
		for (j=0;j<=ny+1;j++) 
			{
			clb[i][j].occ = 0;
			}
	
	fgets( buf, BUFSIZE, fp );
	token = strtok( buf, seps );  /* .SIZE token */ 
	token = strtok( NULL, seps ); /* NX */
	tempInt = atoi( token ); 
	token = strtok( NULL, seps ); /* NY */
	tempInt2 = atoi( token );

	/* Size check */
	if( tempInt2 != ny || tempInt != nx )
	{
		printf("Wrong Input Size!\n");
	}

	/* Go through and repopulate all the blocks */
	for( i=0; i<num_blocks; ++i )
		{
		fgets( buf, BUFSIZE, fp );
		token = strtok( buf, seps );
		tempInt = atoi( token );   /* X */
		token = strtok ( NULL, seps );
		tempInt2 = atoi( token );  /* Y */

		/* Set CLB data structures */
		if( tempInt == 0 || tempInt == nx + 1 ||
			tempInt2 == 0 || tempInt2 == ny +1 )
			{
			/* IO block */
			clb[tempInt][tempInt2].u.io_blocks[clb[tempInt][tempInt2].occ] = i;
			clb[tempInt][tempInt2].occ++;
			}
		else  /* CLB */
			{
			clb[tempInt][tempInt2].occ = 1;
			clb[tempInt][tempInt2].u.block = i;
			}

		/* Set BLOCK data structures */
		block[i].x = tempInt;
		block[i].y = tempInt2;
		block[i].lookup = -1;

		if( tempInt < 0 || tempInt > nx+1 || tempInt2 < 0 || tempInt2 > ny+1 )
			{
			printf("Out of bounds in reading placement file from super grid\n");
			}
		}

	/* Do some error checking */
	for( i=0; i<num_blocks; ++i)
	{
		if( block[i].type == CLB )
		{
			if(clb[block[i].x][block[i].y].u.block != i )
				printf("BLOCK READ ERROR\n");
		}
		else 
		{	
			found = 0;
			for( j=0; j<clb[block[i].x][block[i].y].occ; ++j)
			{
				if( clb[block[i].x][block[i].y].u.io_blocks[j] == i )
					found = 1;
			}
			if( found == 0 )
				printf("IO block not set!\n");
		}
	}
	/* Read the block locations back and set the CLB locations accordingly */
	fclose( fp );
}

/* Super grid idea, have to find a home for this block */
int Find_Home(int blocknum, struct s_placer_opts* s_opts)
{
	/* Naive implememntation right now. Just use the first block we encounter */
	int x,y;

	/* Look for the first available block and make the switch*/
	for(x=s_opts->grid_x1+1; x<=s_opts->grid_x2; ++x)
		for(y=s_opts->grid_y1+1; y<=s_opts->grid_y2; ++y)
		{
			/* block not occupied */
			if( clb[x][y].occ == 0 )
			{
				/* Reset CLB of where the block is from */
				clb[block[blocknum].x][block[blocknum].y].occ = 0;
				clb[block[blocknum].x][block[blocknum].y].u.block = -1;

				/* Set the to CLB to be occupied by the block */
				clb[x][y].occ = 1;
				clb[x][y].u.block = blocknum;
				
				/* Update the block data */
				block[blocknum].x = x;
				block[blocknum].y = y;
				return 1;
			}
		}
	return 0;  
} /* Find_Home */

/* Super grid idea, have to clean up outliers at the end */
void Cleanup_Outliers( struct s_placer_opts* p_opts )
{
	int i;
	/* Go through every block */
	for( i=0; i<num_blocks; ++i )
	{
		/* ONLY CLBs have a floorplan */ 
		if( block[i].type == CLB )
		{
			/* Are we out of the target area? */
			if( block[i].x <= p_opts->grid_x1 ||
				block[i].x > p_opts->grid_x2 ||
				block[i].y <= p_opts->grid_y1 ||
				block[i].y > p_opts->grid_y2 )
				{
				if( !Find_Home( i, p_opts ) )
					{
					printf("Unable to find a home for the outlier, ERROR!!\n");
					}
				} /* if oob*/
		}  /* if clb*/
	} /* for all blocks*/

} /* Cleanup_Outliers */

/* Find the index of a component prefixed with type name, returns -1 if not found
*/
int Find_Element( char* name, fp* floorplan, int num_elements )
{
	int num = -1;
	int j;
	char* rtn;
	for( j=0; j< num_elements; j++ )
	{
		// Do a search for the keyword
		rtn = strstr( name, floorplan[j].name );
		if (rtn != NULL)   //if found, return j;
			return j;
	}
	return num;
}

/* Recount the number of free CLBs per floorplan
*
*/
void RecountFreeClbsPerFloorplan( fp* floorplan, int num_elements )
{
	int i, x, y, x_to, y_to;
	for( i=0; i< num_elements; ++i )
	{
		floorplan[i].numFreeCLB = 0;
		x_to = floorplan[i].xFrom + floorplan[i].xMax - 1;  /* Find_to function limit*/
		y_to = floorplan[i].yFrom + floorplan[i].yMax - 1;
		for( x=floorplan[i].xFrom; x<=x_to; ++x )
			for( y=floorplan[i].yFrom; y<=y_to; ++y )
			{
				if( clb[x][y].occ == 0 )
				{
					floorplan[i].numFreeCLB++;
				}
			}
	}
}

/* Dump incremental floorplan */
void DumpIncrementalFloorplan( char* dest_file )
{
	FILE* fp;
	int i;

	printf("In DumpIncrementalFloorplan\n");
	fflush(stdout);
	fp = fopen( dest_file, "w+" );
	printf("file opened %s\n",dest_file);
	fflush(stdout);
	


	/*print out number of blocks we had */
	fprintf( fp, "%i %i %i\n", num_blocks, nx, ny );
	/* Go through every CLB and dump the previous placement */
	for( i=0; i<num_blocks; ++i )
		{
		fprintf( fp, "%s %i %i\n", block[i].name, block[i].x, block[i].y );
		}

	fclose( fp );

	printf("Finished DIF\n");
	fflush(stdout);
}  /* DumpIncrementalFloorplan */

/* Find the index for the previous location */
int FindPrevLocation( char* name, clb_loc* clbs, int num_elems )
{
	int i;
	for( i=0; i<num_elems; ++i )
	{
		if( strcmp( name, clbs[i].name ) == 0 )
			return i;
	}
	return -1;
} /* Find Prev Location */

/* Read Incremental Floorplan*/
void ReadIncrementalFloorplan(  char* src_file, fp* floorplan, int num_fp_elems, struct s_placer_opts* p_opts )
{
	/* Variables */
	char buf[BUFSIZE];
	char seps[]   = " ,\t\n";
	char* token;
	FILE* fp;
	int i,j, prev_num_blocks;
	int x_to, y_to;
	int oldnx, oldny;
	int which_type;
	clb_loc* clbs;
	
	/* Initialization */
	fp = fopen( src_file, "r" );

	/* Get the data from the file */
	fgets( buf, BUFSIZE, fp );
	token = strtok( buf, seps );       /* OLD NUM CLBS */
	prev_num_blocks = atoi(token);
	token = strtok( NULL, seps );      /* NX */
	oldnx = atoi( token );        
	token = strtok( NULL, seps );      /* NY */
	oldny = atoi( token );       
	clbs = (clb_loc*) malloc( prev_num_blocks*sizeof(clb_loc) );
	for( j=0; j<prev_num_blocks; ++j )
	{
		fgets( buf, BUFSIZE,fp );
		/* Name token */
		token = strtok( buf, seps );
		strcpy( clbs[j].name, token );
		/* X loc token */
		token = strtok( NULL, seps );
		clbs[j].x = atoi( token );
		/* Y loc token */
		token = strtok( NULL, seps );
		clbs[j].y = atoi( token );
	}

	/* Go through each block and try to place.*/
	for( i=0; i<num_blocks; ++i )
		{
			/* If this was part of the non-delta block, we place it in the previous location */
			/* Now look for where it was placed previously */
			which_type = FindPrevLocation( block[i].name, clbs, prev_num_blocks );
				
			if( which_type == -1 )
			{
				/* New block added to the circuit, don't place it yet. */
				continue;
			}
			/* Place the CLB now */
			if( block[i].type == CLB )
				{
				x_to = clbs[which_type].x + p_opts->grid_x1 + (p_opts->norm_X-oldnx)/2;
				y_to = clbs[which_type].y + p_opts->grid_y1 + (p_opts->norm_Y-oldny)/2;
				clb[x_to][y_to].occ = 1;
				clb[x_to][y_to].u.block = i;
				block[i].x = x_to;
				block[i].y = y_to;
				}
			else  /* IO Block */
				{
				if( clbs[which_type].x == 0 )
					{
					x_to = 0;
					y_to = clbs[which_type].y + p_opts->grid_y1 + (p_opts->norm_Y-oldny)/2;
					}
				else if ( clbs[which_type].x == oldnx +1 )
					{
					x_to = nx + 1;
					y_to = clbs[which_type].y + p_opts->grid_y1 + (p_opts->norm_Y-oldny)/2;
					}
				else if ( clbs[which_type].y == 0 )
					{
					x_to = clbs[which_type].x + p_opts->grid_x1 + (p_opts->norm_X-oldnx)/2;
					y_to = 0;
					}
				else if ( clbs[which_type].y == oldny + 1 )
					{
					x_to = clbs[which_type].x + p_opts->grid_x1 + (p_opts->norm_X-oldnx)/2;
					y_to = ny + 1;
					}
				else
					{
					printf( "ERROR in incremental place\n");
					}
				clb[x_to][y_to].u.io_blocks[clb[x_to][y_to].occ] = i;
				clb[x_to][y_to].occ++;
				block[i].x = x_to;
				block[i].y = y_to;
				} /* else IO */
		} /* for */

	/* Cleanup */
	fclose( fp );
	free(clbs);
} /* Read Incremental Floorplan */

void DebugPrintFloorplan( fp* floorplan, int num_elem )
{
    int i;
    for( i=0; i< num_elem; ++i )
    {
        printf("Floorplan %i: %s num free: %i : %i %i %i %i\n", i, floorplan[i].name, floorplan[i].numFreeCLB,
                floorplan[i].xFrom, floorplan[i].xMax, floorplan[i].yFrom, floorplan[i].yMax);
    }
}

