#include <stdio.h> 
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "graphics.h"
#include "path_delay.h"
#include "draw.h"
#include "conge_est.h"

t_subblock_data *luts_in_clbs;
double **rent_lut_conge_map, **rent_clb_conge_map, **net_win_conge_map, **net_lut_conge_map, 
**net_clb_conge_map, **bounding_box_conge_map,  **bounding_box_area_conge_map, **pins_clb_conge_map, 
**pins_lut_conge_map, **iterative_route_conge_map, **channel_conge_map;
double **rent_lut_conge_image, **rent_clb_conge_image, **net_win_conge_image, **net_lut_conge_image,
**net_clb_conge_image, **bounding_box_conge_image,  **bounding_box_area_conge_image, 
**pins_clb_conge_image, **pins_lut_conge_image,  **iterative_route_conge_image, **channel_conge_image,
**diff_real_est_conge_image;
double *cutsPerArea;
double *samplesPerArea;  
enum congestion_type congestion;
vboolean *netVisited;
int ***nets_cartesian; 
int window, blend_time, update_conge;
double peak_temp, bottom_temp, peak_sat, bottom_sat, blend_rate;
char *window_size_string, *peak_temp_string, *blend_rate_string, *blend_time_string,
*bottom_temp_string, *peak_sat_string, *bottom_sat_string;

vboolean calced_rent_lut_conge, calced_rent_clb_conge, calced_net_win_conge, calced_net_lut_conge, calced_net_clb_conge, 
calced_bounding_box_conge, calced_bounding_box_area_conge, calced_channel_conge, calced_pins_clb_conge, 
calced_pins_lut_conge, calced_iterative_route_conge, iterative_route_mode, ever_calced_iterative_route, compare_real;
t_subblock_data *subblock_data_ptr_g; 
t_timing_inf timing_inf_g;        
struct s_router_opts router_opts_g;
struct s_det_routing_arch det_routing_arch_g;
t_segment_inf *segment_inf_g; 
t_chan_width_dist chan_width_dist_g;
vboolean placed, first_display, command_line;
float ***net_slack_g, ***net_delay_g;
struct s_placer_opts *placer_opts_g;

float *x_clb_left, *y_clb_bottom, clb_width;

char conge_file_name[BUFSIZE];

unsigned int display_width, display_height;  /* screen size */
float xmult, ymult;                  /* Transformation factors */

enum pic_type pic_on_screen;

double minSatStart, minSatEnd, maxSatStart, maxSatEnd, total_est_conge, total_real_conge, minSatReal, maxSatReal;

char congestion_file[BUFSIZE];
float **place_conge_map;
 
static void calcCutsIterative(int globalxmin, int globalymin, int globalxmax, int globalymax);
static void calcCutsSquare(int xmin, int ymin, int xmax, int ymax);
static void calcCut(int xmin, int xmax, int ymin, int ymax);
static double calcCutPerimeter(int xmin, int ymin, int xmax, int ymax);
static void print_conge_map(double **conge_map);
static void calcImage(double **conge_map, double **conge_image);
static void iterative_route();

void init_conge_est(t_subblock_data *subblock_data)
	{		
	sprintf(congestion_file, "congestion_file");
	place_conge_map = NULL;
	command_line = FALSE;
	placed = FALSE;
	first_display = TRUE;
	pic_on_screen = NO_PICTURE;
	update_conge = 1;
	calced_rent_lut_conge = FALSE;
	calced_net_win_conge = FALSE;
	calced_net_lut_conge = FALSE;
	calced_rent_clb_conge = FALSE;
	calced_bounding_box_conge = FALSE;
	calced_channel_conge = FALSE;
 	calced_bounding_box_area_conge = FALSE;
	calced_pins_clb_conge = FALSE;
	calced_pins_lut_conge = FALSE;
	calced_iterative_route_conge = FALSE;
	iterative_route_mode = FALSE;
	ever_calced_iterative_route = FALSE;
	compare_real = FALSE;
	congestion = NONE;
	luts_in_clbs = subblock_data; 
	window = INIT_WIN_SIZE;		
	peak_temp = NUM_COLOR-12;
 	bottom_temp = 0;
	peak_sat = NUM_COLOR-12;
 	bottom_sat = 0;
 	blend_time = 0; 	
 	blend_rate = 0; 	
	window_size_string = malloc(30*sizeof(char));
	peak_temp_string = malloc(30*sizeof(char));
	bottom_temp_string = malloc(30*sizeof(char));
	bottom_sat_string = malloc(30*sizeof(char));
	peak_sat_string = malloc(30*sizeof(char));
	blend_time_string = malloc(30*sizeof(char));
	blend_rate_string = malloc(30*sizeof(char));
	sprintf(window_size_string, "Window Size: %d", window);
	sprintf(peak_temp_string, "Peak Temp: %d\%", (int)(100*(peak_temp)/(NUM_COLOR-12)));
	sprintf(bottom_temp_string, "Bottom Temp: %d\%", (int)(100*(bottom_temp)/(NUM_COLOR-12)));
	sprintf(bottom_sat_string, "Bottom Saturation: %d\%", (int)(100*(bottom_sat)/(NUM_COLOR-12)));
	sprintf(peak_sat_string, "Peak Saturation: %d\%", (int)(100*(peak_sat)/(NUM_COLOR-12)));
	sprintf(blend_time_string, "Blend Iterations: %d\%", blend_time);
	sprintf(blend_rate_string, "Blend Rate: %g\%", blend_rate);	
	}

float calc_place_conge(struct s_bb *bb_coords)
	{	
	int x, y, currentNet, maxX, maxY, minX, minY, node;
	float total_cost, bb_width, bb_height, bb_area, bb_inc, bb_extra_pin_factor;
	
	if (place_conge_map == NULL)	
		place_conge_map = (float**)alloc_matrix(0, nx, 0, ny, sizeof(float));
		
	for (x = 0; x <= nx; x++)
		{
		for (y = 0; y <= ny; y++)
			{
			place_conge_map[x][y] = 0.;
			}
		}	
		
	for (currentNet = 0; currentNet < num_nets; currentNet++)
		{
		maxX = 1;
		maxY = 1;
		minX = nx;
		minY = ny;
		
		if (bb_coords == NULL)
			{
			for(node = 0; node < net[currentNet].num_pins; node++)
				{
				maxX = max(block[net[currentNet].blocks[node]].x, maxX);	
				maxY = max(block[net[currentNet].blocks[node]].y, maxY);
				minX = min(block[net[currentNet].blocks[node]].x, minX);	
				minY = min(block[net[currentNet].blocks[node]].y, minY);
				}	
			}
		else
			{
			minX = max(0, bb_coords[currentNet].xmin);
			maxX = min(nx, bb_coords[currentNet].xmax);
			minY = max(0, bb_coords[currentNet].ymin);
			maxY = min(ny, bb_coords[currentNet].ymax);
			}
			
		bb_width = (float)(bb_coords[currentNet].xmax - bb_coords[currentNet].xmin + 1); 
		bb_height =  (float)(bb_coords[currentNet].ymax - bb_coords[currentNet].ymin + 1);	
		bb_extra_pin_factor = min(bb_width, bb_height)*((float)(max(0, net[currentNet].num_pins-3)));
		bb_area = bb_width*bb_height;
		bb_inc = (bb_width + bb_height + bb_extra_pin_factor)/bb_area;
				
		for (x = max(0,minX); x <= min(nx,maxX); x++)
			{
			for (y = max(0,minY); y <= min(nx,maxY); y++)
				{				
				place_conge_map[x][y] += bb_inc;	
				}
			}		
		}	
	
	total_cost = 0;
		
	for (x = 0; x <= nx; x++)
		{
		for (y = 0; y <= ny; y++)
			{					
			total_cost += place_conge_map[x][y] * place_conge_map[x][y];	
			}
		}	
	
	return (float)total_cost;
	}

void calc_conge()	
	{	
 	double **conge_map, **conge_image; 
	int x, y, i, xmin, xmax, ymin, ymax, partitionCount, maxArea, currentNet, node, maxX, minX, maxY, minY, pin;
	double fitProd, fitSumX, fitSumY, fitSumX2, saturateMaxRent, saturateMinRent;
	double half_perimeter, bounding_box_inc, bounding_box_area,
	bounding_box_extra_pin_factor;
	double minSat, maxSat;
	float total_cost;
	
	total_cost = 0;
		
	if (congestion == BOUNDING_BOX || congestion == BOUNDING_BOX_AREA)
		{
		conge_map = (double**)alloc_matrix(0, nx, 0, ny, sizeof(double));
		
		for (x = 1; x <= nx; x++)
			{
			for (y = 1; y <= ny; y++)
				{
				conge_map[x][y] = 0;
				}
			}	
			
		for (currentNet = 0; currentNet < num_nets; currentNet++)
			{
			maxX = 1;
			maxY = 1;
			minX = nx;
			minY = ny;
			
			for(node = 0; node < net[currentNet].num_pins; node++)
				{
				maxX = max(block[net[currentNet].blocks[node]].x, maxX);	
				maxY = max(block[net[currentNet].blocks[node]].y, maxY);
				minX = min(block[net[currentNet].blocks[node]].x, minX);	
				minY = min(block[net[currentNet].blocks[node]].y, minY);
				}	
				
			half_perimeter = (double)((maxX-minX) + (maxY-minY) + 2);	
			bounding_box_extra_pin_factor = ((double)(min(maxX-minX+1, maxY-minY+1)*(max(0, net[currentNet].num_pins-3))*window))/20;
			bounding_box_area = (double)(maxX-minX+1)*(maxY-minY+1);
			bounding_box_inc = (half_perimeter + bounding_box_extra_pin_factor)/bounding_box_area;
					
			for (x = max(1, minX-1); x <= min(nx, maxX+1); x++)
				{
				for (y = max(1, minY-1); y <= min(ny, maxY+1); y++)
					{
					if (congestion == BOUNDING_BOX_AREA)
						conge_map[x][y] += bounding_box_inc;	
					else
						conge_map[x][y]++;
					}
				}	
			}						
		}	
	else if (congestion == NET_PER_WIN || congestion == NET_PER_CLB || congestion == NET_PER_LUT 
		 		|| congestion == RENT_PER_LUT || congestion == RENT_PER_CLB || congestion == PINS_PER_CLB
		 		|| congestion == PINS_PER_LUT)
		{
		if (congestion == RENT_PER_LUT)
			maxArea = luts_in_clbs->max_subblocks_per_block * window * window;
		else if (congestion == RENT_PER_CLB)
			maxArea = window * window;
			
		conge_map = (double**)alloc_matrix(0, nx, 0, ny, sizeof(double));
		nets_cartesian = (int***)alloc_matrix(0, nx, 0, ny, sizeof(int*));
		
		if (congestion == RENT_PER_LUT || congestion == RENT_PER_CLB) 
			{
			cutsPerArea = my_malloc((maxArea+1)*sizeof(double));			
			samplesPerArea = my_malloc((maxArea+1)*sizeof(double));
			partitionCount = 0;
			
			for (i = 0; i <= maxArea; i++)
				{
				samplesPerArea[i] = 0;
				cutsPerArea[i] = 0;
				}
			}
				
		netVisited = my_malloc(num_nets*sizeof(vboolean));	
			
		for (i=0; i<num_blocks; i++)
			{		
			if (block[i].type == CLB)
				{
				nets_cartesian[block[i].x][block[i].y] = block[i].nets;
				}
			}
					
		for (x=1; x<=nx; x++)
			{		
			for (y=1; y<=ny; y++)
				{
				xmin = x - window/2;
				xmax = xmin + window - 1; 				
				ymin = y - window/2;
				ymax = ymin + window - 1;	
				
				if ((congestion == RENT_PER_LUT || congestion == RENT_PER_CLB || congestion == NET_PER_WIN || 
					congestion == NET_PER_LUT || congestion == NET_PER_CLB) && ((clb[x][y].occ == 0) || 
					(clb[x][y].type == CLB) && (luts_in_clbs->num_subblocks_per_block[clb[x][y].u.block] == 0)))
					{
					conge_map[x][y] = -1;					
					}
				else
					{				
					if (congestion == NET_PER_WIN || congestion == NET_PER_LUT || congestion == NET_PER_CLB)
						{
						conge_map[x][y] = (double)(calcCutPerimeter(xmin, ymin, xmax, ymax));
						}
					else if (congestion == PINS_PER_CLB || congestion == PINS_PER_LUT)
						{
						conge_map[x][y] = 0;
						
						if (clb[x][y].type == CLB)
							{
							if (clb[x][y].occ == 1) 
								{
								for (pin = 0; pin < pins_per_clb; pin++)
									{
									if (nets_cartesian[x][y][pin] != OPEN)
										conge_map[x][y]++;
									}
									
								if (congestion == PINS_PER_LUT)
									conge_map[x][y] /= luts_in_clbs->num_subblocks_per_block[clb[x][y].u.block];
								}
							else
								{
								conge_map[x][y] = -1;
								}
							}
						}
					else
						{
						//calcCutsSquare(xmin, ymin, xmax, ymax);	
						calcCutsIterative(xmin, ymin, xmax, ymax);	 
						
						fitProd = 0;
						fitSumX = 0;
						fitSumY = 0;
						fitSumX2 = 0;
						partitionCount = 0; 
			
						for (i = 1; i <= maxArea; i++)
							{
							if (samplesPerArea[i] > 0 && cutsPerArea[i] > 0)
								{
								partitionCount++;
								cutsPerArea[i] = (double)log10(cutsPerArea[i]/samplesPerArea[i]);										
								fitProd += log10((double)i) * cutsPerArea[i];
								fitSumX += log10((double)i);
								fitSumX2 += log10((double)i)*log10((double)i);
								fitSumY += cutsPerArea[i];
								}	
							}	
						
						partitionCount = max(1, partitionCount); //prevent divide by zero
						fitSumX2 = max(1,fitSumX2);	//prevent divide by zero
						conge_map[x][y] = (fitProd - fitSumX*fitSumY/partitionCount)/fitSumX2;
						
						for (i = 0; i <= maxArea; i++)
							{
							samplesPerArea[i] = 0;
							cutsPerArea[i] = 0;
							}
						}
					}
				}
			}
		
		free_matrix(nets_cartesian, 0, nx, 0, sizeof(int*));
		
		if (congestion == RENT_PER_LUT || congestion == RENT_PER_CLB)
			{
			free(cutsPerArea);
			free(samplesPerArea);	
			}
			
		free(netVisited);	
		}
	else if (congestion == CHAN_WIDTH)
		{
		conge_map = channel_conge_map;
		}
	else if (congestion == ITERATIVE_ROUTE)
		{
		iterative_route();
		conge_map = iterative_route_conge_map;
		}
	
	conge_image = (double**)alloc_matrix(0, nx, 0, ny, sizeof(double));
	calcImage(conge_map, conge_image); 		
	}
	
static void calcImage(double **conge_map, double **conge_image_0)
	{
	double sides, sides_count;
	double **conge_image_1, **conge_image_swapper;
	int x, y, blends, node, pin, netX, netY, num_pins, balance, balanced_node;
	minSatStart = 999999999;
	maxSatStart = 0;
	blends = 0;
	balance = 0;
				
	if (blend_time > 0 && blend_rate > 0)
		{	
		conge_image_1 = (double**)alloc_matrix(0, nx, 0, ny, sizeof(double));
			
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{
				if (conge_map[x][y] < minSatStart && conge_map[x][y] != -1)
					minSatStart = conge_map[x][y];
				}
			}	
				
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{		
				conge_image_0[x][y] = max(minSatStart,conge_map[x][y]);
				}
			}		
		}						

	for (blends = 0; blends < blend_time*10 && blend_rate > 0; blends++)	
		{
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{		
				sides = 0;
				sides_count = 0;
				
				if (congestion == PINS_PER_CLB || congestion == PINS_PER_LUT)
					{			
					if ((clb[x][y].occ == 1) && (clb[x][y].type == CLB))
						{
						for (pin = 0; pin < pins_per_clb; pin++)
							{
							if (block[clb[x][y].u.block].nets[pin] != OPEN)
								{
								balance++;
								num_pins = net[block[clb[x][y].u.block].nets[pin]].num_pins;
								
								for (node = 0; node < num_pins && node < 8; node++)
									{
									balanced_node = (node + balance) % num_pins;
									netX = block[net[block[clb[x][y].u.block].nets[pin]].blocks[balanced_node]].x;
									netY = block[net[block[clb[x][y].u.block].nets[pin]].blocks[balanced_node]].y;
									
									if (netX != x && netY != y && netX <= nx && netX >= 1 && netY <= ny && netY >=1)
										{
										sides += conge_image_0[netX][netY];
										sides_count++;
										}
									}
								}	
							}
						sides /= sides_count;
						}
					}
				else
					{			
					if (x > 1)
						{
						sides += conge_image_0[x-1][y];
						sides_count++;
						}
						
					if (x < nx)
						{
						sides += conge_image_0[x+1][y];
						sides_count++;
						}
					
					if (y > 1)
						{
						sides += conge_image_0[x][y-1];
						sides_count++;
						}
					
					if (y < ny)
						{
						sides += conge_image_0[x][y+1];	
						sides_count++;	
						}			
						
					if (x > 1 && y > 1)
						{
						sides += conge_image_0[x-1][y-1];
						sides_count++;
						}
						
					if (x > 1 && y < ny)
						{
						sides += conge_image_0[x-1][y+1];
						sides_count++;
						}
						
					if (x < nx && y < ny)
						{
						sides += conge_image_0[x+1][y+1];
						sides_count++;
						}
						
					if (x < nx && y > 1)
						{
						sides += conge_image_0[x+1][y-1];
						sides_count++;
						}
						
					sides /= sides_count;
					}
				conge_image_1[x][y] = (1-blend_rate/1000)*conge_image_0[x][y] + sides*blend_rate/1000;	
				}
			}
		conge_image_swapper = conge_image_0;
		conge_image_0 = conge_image_1;
		conge_image_1 = conge_image_swapper;
		}
	
	if (blend_time > 0 && blend_rate > 0)
		{
		free(conge_image_1);
		conge_image_swapper = conge_image_0;
		}	
	else
		conge_image_swapper = conge_map;

	maxSatStart = 0;	
	minSatStart = 999999999;

	if (congestion == PINS_PER_CLB || congestion == PINS_PER_LUT)
		{
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{
				if ((clb[x][y].occ == 1 ) && (conge_image_swapper[x][y] < minSatStart))
					minSatStart = conge_image_swapper[x][y];
				}
			}
			
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{
				if (clb[x][y].occ == 0)
					conge_image_swapper[x][y] = minSatStart;
				}
			}
		}	

	minSatStart = 999999999;

	if (congestion == PINS_PER_CLB || congestion == PINS_PER_LUT)
		{
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{
				if ((clb[x][y].occ == 1 ) && (conge_image_0[x][y] < minSatStart))
					minSatStart = conge_image_0[x][y];
				}
			}
			
		for (x=1; x<=nx; x++)
			{
			for (y=1; y<=ny; y++)
				{
				if (clb[x][y].occ == 0)
					conge_image_0[x][y] = minSatStart;
				}
			}
		}

	minSatStart = 999999999;

	for (x=1; x<=nx; x++)
		{
		for (y=1; y<=ny; y++)
			{
			if (conge_image_swapper[x][y] < minSatStart && 
			!(conge_map[x][y] == -1 && (blend_time < 1 || blend_rate < 1 )))
				minSatStart = conge_image_swapper[x][y];

			if (conge_image_swapper[x][y] > maxSatStart)
				maxSatStart = conge_image_swapper[x][y];
			}
		}	
	
	maxSatEnd = maxSatStart - (maxSatStart-minSatStart)*(1 - peak_sat/(double)(NUM_COLOR-12));
	minSatEnd = minSatStart +(maxSatStart-minSatStart)*bottom_sat/(double)(NUM_COLOR-12);
	total_est_conge = 0;
		
	for (x=1; x<=nx; x++)
		{
		for (y=1; y<=ny; y++)
			{		
			conge_image_0[x][y] = max(min(maxSatEnd, conge_image_swapper[x][y]), minSatEnd); 
			total_est_conge += conge_image_0[x][y];
				
			if (conge_map[x][y] == -1 && (blend_time < 1 || blend_rate < 1))
				conge_image_0[x][y] = bottom_temp;	
			else
				{
				conge_image_0[x][y] = conge_image_0[x][y] - minSatEnd;											
				conge_image_0[x][y] *= (double)(peak_temp - bottom_temp)/(double)(maxSatEnd-minSatEnd);	
				conge_image_0[x][y] += bottom_temp;						
				conge_image_0[x][y] = max (0, conge_image_0[x][y]);
				conge_image_0[x][y] = min (129, conge_image_0[x][y]); 		
				}
			}
		}

	if (congestion == CHAN_WIDTH)
		{
		total_real_conge = total_est_conge;
		maxSatReal = maxSatStart;
		minSatReal = minSatStart;
		}
				
	if (congestion == RENT_PER_LUT)
		{
		rent_lut_conge_map = conge_map;
		rent_lut_conge_image = conge_image_0;
		}
	else if (congestion == RENT_PER_CLB)
		{
		rent_clb_conge_map = conge_map;
		rent_clb_conge_image = conge_image_0;		
		}
	else if (congestion == NET_PER_WIN)
		{
		net_win_conge_map = conge_map;
		net_win_conge_image = conge_image_0;		
		}
	else if (congestion == NET_PER_LUT)
		{
		net_lut_conge_map = conge_map;
		net_lut_conge_image= conge_image_0;		
		}
	else if (congestion == NET_PER_CLB)
		{
		net_clb_conge_map = conge_map;
		net_clb_conge_image = conge_image_0;		
		}
	else if (congestion == BOUNDING_BOX)
		{
		bounding_box_conge_map = conge_map;	
		bounding_box_conge_image = conge_image_0;			
		}
	else if (congestion == BOUNDING_BOX_AREA)
		{
		bounding_box_area_conge_map = conge_map;
		bounding_box_area_conge_image = conge_image_0;		
		}		
	else if (congestion == PINS_PER_CLB)
		{
		pins_clb_conge_map = conge_map;		
		pins_clb_conge_image = conge_image_0;				
		}	
	else if (congestion == PINS_PER_LUT)
		{
		pins_lut_conge_map = conge_map;		
		pins_lut_conge_image = conge_image_0;				
		}	
	else if (congestion == ITERATIVE_ROUTE)
		{
		iterative_route_conge_map = conge_map;		
		iterative_route_conge_image = conge_image_0;				
		}	
	else if (congestion == CHAN_WIDTH)
		{
		channel_conge_image = conge_image_0;					
		}		
		
	if (compare_real)	
		calc_diff_real_est(conge_image_0);
	}
	
void calc_diff_real_est(double **conge_est_image)	
	{
	int x, y;
		
	for (x = 1; x <= nx; x++ )
		{
		for (y = 1; y <= ny; y++)
			{
			diff_real_est_conge_image[x][y] = abs(channel_conge_image[x][y] - conge_est_image[x][y]);
			}
		}	
		
	}
	
void callCalcImage()
	{
	if (congestion == RENT_PER_LUT)
		{
		calcImage(rent_lut_conge_map, rent_lut_conge_image);
		}
	else if (congestion == RENT_PER_CLB)
		{
		calcImage(rent_clb_conge_map, rent_clb_conge_image);	
		}
	else if (congestion == NET_PER_WIN)
		{
		calcImage(net_win_conge_map, net_win_conge_image);	
		}
	else if (congestion == NET_PER_LUT)
		{
		calcImage(net_lut_conge_map, net_lut_conge_image);	
		}
	else if (congestion == NET_PER_CLB)
		{
		calcImage(net_clb_conge_map, net_clb_conge_image);	
		}
	else if (congestion == BOUNDING_BOX)
		{
		calcImage(bounding_box_conge_map, bounding_box_conge_image);		
		}
	else if (congestion == BOUNDING_BOX_AREA)
		{
		calcImage(bounding_box_area_conge_map, bounding_box_area_conge_image);		
		}		
	else if (congestion == PINS_PER_CLB)
		{
		calcImage(pins_clb_conge_map, pins_clb_conge_image);			
		}	
	else if (congestion == PINS_PER_LUT)
		{
		calcImage(pins_lut_conge_map, pins_lut_conge_image);			
		}	
	else if (congestion == ITERATIVE_ROUTE)
		{
		calcImage(iterative_route_conge_map, iterative_route_conge_image);			
		}	
	else if (congestion == CHAN_WIDTH)
		{
		calcImage(channel_conge_image, channel_conge_map);				
		}
	}

static void iterative_route()
	{
	int i,j;
    int **chanx_occ;   /* [1..nx][0..ny] */
    int **chany_occ;   /* [0..nx][1..ny] */
	t_ivec **clb_opins_used_locally;  	
	float **net_slack, **net_delay;
	struct s_linked_vptr *net_delay_chunk_list_head;
		
	iterative_route_mode = TRUE;
	ever_calced_iterative_route = TRUE;
	
// 	if (placer_opts_g->place_freq != PLACE_NEVER && !ever_calced_iterative_route)
//     	{
//     	ever_calced_iterative_route = TRUE;
//   	 	free_lookups_and_criticalities(net_delay_g, net_slack_g);   	 	
//     	}
//     else if(placer_opts_g->place_freq != PLACE_NEVER && pic_on_screen != PLACEMENT)
//     	{
     	free_lookups_and_criticalities(net_delay_g, net_slack_g);
//     	}
	
	chanx_occ = (int **) alloc_matrix (1, nx, 0, ny, sizeof(int));
    chany_occ = (int **) alloc_matrix (0, nx, 1, ny, sizeof(int));
    iterative_route_conge_map = (double **) alloc_matrix (0, nx+1, 0, ny+1, sizeof(double));	
	clb_opins_used_locally = alloc_route_structs (*subblock_data_ptr_g); 	
	
	if (timing_inf_g.timing_analysis_enabled) 
		{
       	net_slack = alloc_and_load_timing_graph (timing_inf_g, *subblock_data_ptr_g);      
       	net_delay = alloc_net_delay (&net_delay_chunk_list_head);
    	}
    else 
    	{
       	net_delay = NULL;    /* Defensive coding. */
       	net_slack = NULL;
    	}
        
 	try_route (window*5, router_opts_g, det_routing_arch_g, segment_inf_g, timing_inf_g, 
	net_slack, net_delay, chan_width_dist_g, clb_opins_used_locally);	
    load_channel_occupancies (chanx_occ, chany_occ);

    // Initialize channel_conge_map and marked_clbs
    for (i=0; i<=nx+1; i++) 
    	{
        for (j=0; j<=ny+1; j++) 
            iterative_route_conge_map[i][j] = 0;
    	}

    // generate clb map to depopulate
    // X-Directed
    for (j=0;j<=ny;j++) 
    	{
        for (i=1;i<=nx;i++) 
        	{
            // mark all adjacent blocks namely blocks [i][j] and [i][j+1]
            if (i>0 && i<nx+1 && j>0 && j<ny+1) 
            	{
                if (iterative_route_conge_map[i][j] < chanx_occ[i][j] ) 
                    iterative_route_conge_map[i][j] = chanx_occ[i][j];				
            	}
            if (i>0 && i<nx+1 && j+1>0 && j+1<ny+1) 
            	{
                if (iterative_route_conge_map[i][j+1] < chanx_occ[i][j] ) 
                    iterative_route_conge_map[i][j+1] = chanx_occ[i][j];
            	}  // if
        	}  // for
    	}  // for
    // Y-Directed
    for (j=1;j<=ny;j++) 
    	{
        for (i=0;i<=nx;i++) 
        	{
            // mark all adjacent blocks namely [i][j] and [i+1][j]
            if (i>0 && i<nx+1 && j>0 && j<ny+1) 
            	{
                if (iterative_route_conge_map[i][j] < chany_occ[i][j] ) 
                    iterative_route_conge_map[i][j] = chany_occ[i][j];				
            	}
            if (i+1>0 && i+1<nx+1 && j>0 && j<ny+1) 
            	{
                if (iterative_route_conge_map[i+1][j] < chany_occ[i][j] )
                    iterative_route_conge_map[i+1][j] = chany_occ[i][j];
            	} // if
        	} // for
    	} // for
		
	iterative_route_mode = FALSE;
	free_matrix (chanx_occ, 1, nx, 0, sizeof(int));
    free_matrix (chany_occ, 0, nx, 1, sizeof(int));    
	free_route_structs (clb_opins_used_locally);
//	free_timing_graph (net_slack);
	//free_subblock_data (subblock_data_ptr_g);
    free_net_delay (net_delay, &net_delay_chunk_list_head);
	fflush (stdout);	
	}
 	
static void calcCutsIterative(int globalxmin, int globalymin, int globalxmax, int globalymax)
	{
	int window, xmin, xmax, ymin, ymax, section; 
	
	for (window = globalxmax - globalxmin + 1; window > 0; window--)
		{
		xmin = globalxmin;
		
		for(xmax = globalxmin + window - 1; xmax < globalxmax + window; xmax += window)
			{
			ymin = globalymin;	
			
			for(ymax = globalymin + window - 1; ymax < globalymax + window; ymax += window)
				{
				calcCut(xmin, min(xmax, globalxmax), ymin, min(ymax, globalymax));				
				ymin += window;	
				}		
			xmin += window;
			} 		
		} 	
	}
	
static void calcCutsSquare(int xmin, int ymin, int xmax, int ymax)
	{
	int x, y, i, node, window, currentNet, area;
	vboolean cut;	
	 
	window = xmax - xmin + 1;
 
	if (window == 0)
		return;
	 
	calcCut(xmin, xmax, ymin, ymax);	
	window = window/2;	
 	calcCutsSquare(xmin, ymax-window+1, xmin+window-1, ymax); 	//Calc top left square
	calcCutsSquare(xmax-window+1, ymax-window+1, xmax, ymax);  	//Calc top right square
	calcCutsSquare(xmin, ymin, xmin+window-1, ymin+window-1);  	//Calc bottom left square
	calcCutsSquare(xmax-window+1, ymin, xmax, ymin+window-1); 	//Calc bottom right square	
	}
	
static double calcCutPerimeter(int xmin, int ymin, int xmax, int ymax)
	{
 	int i, x, y, j, currentNet, node, cuts, area;
 	vboolean cut;	
	cuts = 0;
	
	if (congestion != NET_PER_WIN)
		{
		area = 0;
		
		for(x = max(xmin,1); x <= min(xmax,nx); x++)
			{			
			for(y = max(ymin,1); y <= min(ymax,ny); y++)
				{				
				if ((clb[x][y].occ == 1) && (clb[x][y].type == CLB))
					{
					if (congestion == NET_PER_LUT)
						area += luts_in_clbs->num_subblocks_per_block[clb[x][y].u.block]; 
					else 
						area ++;
					}
				}	
			}
		}
	
	for (i=0; i<num_nets; i++)
		{
		netVisited[i] = FALSE;
		}		
					
	for(x = max(xmin,1); x <= min(xmax,nx); x++)
		{			
		for(y = max(ymin,1); y <= min(ymax,ny); y++)
			{				
			if ((clb[x][y].occ == 1) && (clb[x][y].type == CLB))
				{
				for(i = 0; i < pins_per_clb; i++)
					{ 
					currentNet = nets_cartesian[x][y][i];	
								
					if ((currentNet != OPEN) && !netVisited[currentNet])
						{
						netVisited[currentNet] = TRUE;
						cut = FALSE;
																								
						for(node = 0; (node < net[currentNet].num_pins) && !cut; node++)
							{										
							if ((block[net[currentNet].blocks[node]].x < max(xmin,1))  || 
							    (block[net[currentNet].blocks[node]].x > min(xmax,nx)) ||
								(block[net[currentNet].blocks[node]].y < max(ymin,1))  || 
								(block[net[currentNet].blocks[node]].y > min(ymax,ny))) 
								{
								cut = TRUE;
								cuts++;
								}
							} 					
						}	 	 
					}
				}
			}	
		}
	 
	if (congestion == NET_PER_WIN)
		return (double)cuts; 
	else
		{
		if (area == 0)
			return (double)0;
		else
			return (double)cuts/(double)area;
		}
	} 	
	
static void calcCut(int xmin, int xmax, int ymin, int ymax)
 	{
 	int area, i, x, y, j, currentNet, node; 
 	vboolean cut;
	area = 0;	
	
	for (i=0; i<num_nets; i++)
		{
		netVisited[i] = FALSE;
		}							
		
	for(x = max(xmin,1); x <= min(xmax,nx); x++)
		{			
		for(y = max(ymin,1); y <= min(ymax,ny); y++)
			{				
			if ((clb[x][y].occ == 1) && (clb[x][y].type == CLB))
				{
				if (congestion == RENT_PER_LUT)
					area += luts_in_clbs->num_subblocks_per_block[clb[x][y].u.block]; 
				else 
					area ++;
				}
			}	
		}
	
	samplesPerArea[area]++;	
	
	for(x = max(xmin,1); x <= min(xmax,nx); x++)
		{			
		for(y = max(ymin,1); y <= min(ymax,ny); y++)
			{				
			if ((clb[x][y].occ == 1) && (clb[x][y].type == CLB))
				{
				for(i = 0; i < pins_per_clb; i++)
					{ 
					currentNet = nets_cartesian[x][y][i];	
								
					if ((currentNet != OPEN) && !netVisited[currentNet])
						{
						netVisited[currentNet] = TRUE;
						cut = FALSE;
																								
						for(node = 0; (node < net[currentNet].num_pins) && !cut; node++)
							{										
							if ((block[net[currentNet].blocks[node]].x < max(xmin,1))  || 
							    (block[net[currentNet].blocks[node]].x > min(xmax,nx)) ||
								(block[net[currentNet].blocks[node]].y < max(ymin,1))  || 
								(block[net[currentNet].blocks[node]].y > min(ymax,ny))) 
								{
								cut = TRUE;
								cutsPerArea[area]++;
								}
							} 					
						}	 	 
					}
				}
			}	
		}
	} 
	
void load_congestion()
	{
	char buf[BUFSIZE], tempChar;
 	FILE *fp;
	int x, y, tempInt;
	double conge;
 	
 	if ((fp = fopen(conge_file_name, "r")) == NULL)
        {
        printf("Error opening file %s.\n", conge_file_name);
        exit(1);
        }
     else
     	printf("Successfully opened congestion file %s.\n", conge_file_name);
     
     channel_conge_map = (double**)alloc_matrix(0, nx+1, 0, ny+1, sizeof(double));
     diff_real_est_conge_image = (double**)alloc_matrix(0, nx + 1, 0, ny + 1, sizeof(double));
     calced_channel_conge = TRUE;
     
     for (y=ny; y > 0 ;y--) 
     	{
        for (x=1; x <= nx; x++) 
        	{
            fscanf(fp,"%d%c", &tempInt, &tempChar);
            channel_conge_map[x][y] = (double)tempInt;
        	}
        	
        fscanf(fp,"\n", &tempChar);
    	}
 	
    congestion = CHAN_WIDTH;
    calcImage(channel_conge_map, (double**)alloc_matrix(0, nx, 0, ny, sizeof(double))); 
    congestion = NONE;
	}
 
void free_rent_lut_conge()
	{
	if(calced_rent_lut_conge)
		{
		free_matrix(rent_lut_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(rent_lut_conge_image, 0, nx, 0, sizeof(double));
		calced_rent_lut_conge = FALSE;
		}
	}

void free_rent_clb_conge()
	{
	if(calced_rent_clb_conge)
		{
		free_matrix(rent_clb_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(rent_clb_conge_image, 0, nx, 0, sizeof(double));
		calced_rent_clb_conge = FALSE;
		}
	}
	
void free_net_win_conge()
	{
	if(calced_net_win_conge)
		{
		free_matrix(net_win_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(net_win_conge_image, 0, nx, 0, sizeof(double));
		calced_net_win_conge = FALSE;
		}
	}
	
void free_net_lut_conge()
	{
	if(calced_net_lut_conge)
		{
		free_matrix(net_lut_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(net_lut_conge_image, 0, nx, 0, sizeof(double));
		calced_net_lut_conge = FALSE;
		}
	}
	
void free_net_clb_conge()
	{
	if(calced_net_clb_conge)
		{
		free_matrix(net_clb_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(net_clb_conge_image, 0, nx, 0, sizeof(double));
		calced_net_clb_conge = FALSE;
		}
	}

void free_bounding_box_conge()
	{
	if(calced_bounding_box_conge)
		{
		free_matrix(bounding_box_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(bounding_box_conge_image, 0, nx, 0, sizeof(double));
		calced_bounding_box_conge = FALSE;
		}
	}

void free_bounding_box_area_conge()
	{
	if(calced_bounding_box_area_conge)
		{
		free_matrix(bounding_box_area_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(bounding_box_area_conge_image, 0, nx, 0, sizeof(double));
		calced_bounding_box_area_conge = FALSE;
		}
	}
	
void free_channel_conge()
	{
	if(calced_channel_conge)
		{
		free_matrix(channel_conge_map, 0, nx + 1, 0, sizeof(double));
		free_matrix(channel_conge_image, 0, nx, 0, sizeof(double));
		free_matrix(diff_real_est_conge_image, 0, nx, 0, sizeof(double));
		calced_channel_conge = FALSE;
		}
	}

void free_pins_clb_conge()
	{
	if(calced_pins_clb_conge)
		{
		free_matrix(pins_clb_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(pins_clb_conge_image, 0, nx, 0, sizeof(double));
		calced_pins_clb_conge = FALSE;
		}
	}
	
void free_pins_lut_conge()
	{
	if(calced_pins_lut_conge)
		{
		free_matrix(pins_lut_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(pins_lut_conge_image, 0, nx, 0, sizeof(double));
		calced_pins_lut_conge = FALSE;
		}
	}
	
void free_iterative_route_conge()
	{
	if(calced_iterative_route_conge)
		{
		free_matrix(iterative_route_conge_map, 0, nx, 0, sizeof(double));
		free_matrix(iterative_route_conge_image, 0, nx, 0, sizeof(double));
		calced_iterative_route_conge = FALSE;
		}
	}
	
static void print_conge_map(double **conge_map)
	{
	int x, y;
	printf("\n\n\n\n");
	for (y = ny; y >=  1; y--)
		{
		for (x = 1; x <= nx; x++)
			{
			printf("%g ", conge_map[x][y]);
			}
			
		printf("\n");
		}
	fflush (stdout);
	}
	
void free_int_linked_list(struct int_linked_list **list_head)
	{
	struct int_linked_list **list_next, **list_current;	
	list_next = list_head;
	
	for(list_current = list_head; *list_next != NULL; list_current = list_next)
		{
		*list_next = (*list_current)->next;
		free(*list_current);	
		*list_current = NULL;
		}
	}

void int_list_add(struct int_linked_list **list, struct int_linked_list **list_head, int value)
 		{
 		struct int_linked_list *new;
 		new = (struct int_linked_list *) my_malloc(sizeof(struct int_linked_list));
 		new->value = value;
 		new->next = NULL;
 		
 		if (*list != NULL)
 			(*list)->next = new;
 		else
 			*list_head = new;
 			
 		*list = new;
 		}
	

