#include <stdio.h>
#include <math.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "stats.h"
#include "congestion.h"
#include "conge_est.h"
#include "graphics.h"

/********************** Subroutines local to this module *********************/

void dump_clbs_local (char* file_name);

/************************* Subroutine definitions ****************************/

void dump_clbs_local (char* file_name )
{

    /* Output routine for debugging. */
    int i, j, index;
    int bnum;
    FILE* fp; 
    int len;
    char name[200];
    strcpy(name, " ");
    len = strlen( file_name );
    strncpy( name, file_name, len-4 );
    name[len-4] = '\0';
    strcat( name, "clb");
    fp = fopen( name, "w" );
 
    for (i=0;i<=nx+1;i++) 
    {
        for (j=0;j<=ny+1;j++)
            {
            if (clb[i][j].type == CLB && clb[i][j].occ == 1)
            {
                bnum = clb[i][j].u.block;
                fprintf( fp, " %2d ", block[bnum].num_subblocks );
            }
            else if( clb[i][j].type == IO  || clb[i][j].type == INPAD || clb[i][j].type == OUTPAD )
            {
                fprintf( fp, "%2d ", clb[i][j].occ );
            }
            else 
            {
                fprintf( fp, " %2d ", 0 );
            }
        }
        fprintf( fp, "\n");
    }
    fclose( fp );
}

static print_channel_occ_stdout(int **chanx_occ, int **chany_occ) 
{	
    int i,j;

    printf("X-directed channel capacities\n");
    for (j=ny;j>=0;j--) {
        for (i=1;i<=nx;i++) {
            printf("%d ",chanx_occ[i][j]);
        }
        printf("\n");
    }
    printf("Y-directed channel capacities\n");
    for (j=ny;j>=1;j--) {
        for (i=0;i<=nx;i++) {
            printf("%d ",chany_occ[i][j]);
        }
        printf("\n");
    }
}

void print_channel_occ_file(int **chanx_occ, int **chany_occ, char* file_name) 
{	
    int i,j;
    FILE *xfp, *yfp;
    int len;
    char name2[200];
    char name3[200];

    strcpy(name2, " ");
    len = strlen( file_name );
    strncpy( name2, file_name, len-4 );
    name2[len-4] = '\0';
    strcat( name2, "xChan");
    
    xfp = fopen (name2, "w");
    for (j=ny;j>=0;j--) {
        for (i=1;i<=nx;i++) {
            fprintf(xfp,"%3d ",chanx_occ[i][j]);
        }
        fprintf(xfp,"\n");
    }

    strcpy(name3, " ");
    len = strlen( file_name );
    strncpy( name3, file_name, len-4 );
    name3[len-4] = '\0';
    strcat( name3, "yChan");

    yfp = fopen (name3, "w");
    for (j=ny;j>=1;j--) {
        for (i=0;i<=nx;i++) {
            fprintf(yfp,"%3d ",chany_occ[i][j]);
        }
        fprintf(yfp,"\n");
    }
    fclose(xfp);
    fclose(yfp);

}

void print_channel_conge_file(char* file_name) 
{	
    int i,j;
    FILE *real;
    int len;
    char name4[200];
	
    strcpy(name4, " ");
    len = strlen( file_name );
    strncpy( name4, file_name, len-4 );
    name4[len-4] = '\0';
    strcat( name4, "Real");
    
    real = fopen (file_name, "w");
    for (j=ny;j>=0;j--) {
        for (i=1;i<=nx;i++) {     	
            fprintf(real,"%d ",(int)(channel_conge_map[i][j]));
        }
        fprintf(real,"\n");
    }

    fclose(real);

}


void generate_congestion_map(char *congestion_file, int cw_constraint, int congestion_radius, int route_success, vboolean show_graphics )
{
    int i,j,bug;
    int bnum;
    int **chanx_occ;   /* [1..nx][0..ny] */
    int **chany_occ;   /* [0..nx][1..ny] */
 //   int **channel_conge_map;	   /* [0..nx+1][[0..ny+1] */

    struct s_coord {int x; int y;};

    struct s_coord *peaks;
    int num_peaks;
    struct s_coord congestion_center;
    int max_manhattan_dist;
    int max_chan_width;
    int num_total_subblocks;
    int num_total_clbs;
    int man_dist;

    int num_empty_clbs;
    vboolean done;
    int iteration;
    int num_region_subblocks;
    int num_region_clbs;
    int dpop_val;
    int num_new_clbs;
    int dist;
    enum congestion_type old_conge;

    float avg_cluster_size;
    //	float old_dimension;
    
    /* David addition for multiple peaks depopulation */
    int** marked_clbs;          // [0..nx+1][0..ny+1]
    int*  clb_dpop_amount;  // [0 .. num_blocks]
    int   multiloop_done;       // Are we done depopulating multiple regions?
    int   total_clbs_chosen;  // The total number of clbs we have depopulated
    int   total_subblocks_chosen; // Total number of subblocks currently chosen
    int   total_num_new_clbs;       // Total number of new clbs going to be created
    /* End multi peak depopulation */

    /*depopulation control scheme*/
    float cwc_over_amount;   // How much larger than our target are we?
    int    dpop_growth_amount; // How much larger are we allowing?
    /* End new depopulation control scheme*/
    FILE *cg_fp;
	
    chanx_occ = (int **) alloc_matrix (1, nx, 0, ny, sizeof(int));
    chany_occ = (int **) alloc_matrix (0, nx, 1, ny, sizeof(int));
    peaks = (struct s_coord *) my_malloc (ny*ny*sizeof(struct s_coord));
    marked_clbs = (int**) alloc_matrix( 0, nx+1, 0, ny+1, sizeof(int));
    clb_dpop_amount = (int*) malloc( sizeof( int ) * num_blocks );
    
    if (!calced_channel_conge)
    	{
    	channel_conge_map = (double**)alloc_matrix(0, nx + 1, 0, ny + 1, sizeof(double));
    	diff_real_est_conge_image = (double**)alloc_matrix(0, nx + 1, 0, ny + 1, sizeof(double));
    	}
    	
    calced_channel_conge = TRUE;
    multiloop_done = 0;
    total_clbs_chosen = 0;
    total_subblocks_chosen = 0;
    total_num_new_clbs = 0;
    cg_fp = my_fopen (congestion_file, "w", 0);
    compare_real = TRUE;
    
    // printf("\n\n ----- %i ------ \n\n",show_graphics);
    
    if(show_graphics) {
      init_world (0., (y_clb_bottom[ny+1] + clb_width)/10,
        x_clb_left[nx+1]+ clb_width + x_clb_left[nx+1]*2, 0.);
	  }
    
    //	printf("congestion_radius=%d, nx=%d\n",congestion_radius,nx);
    //	congestion_radius = (int) (((float)nx / (float)congestion_radius) + 0.5);
    //	printf("congestion_radius=%d, nx=%d\n",congestion_radius,nx);

    // Load the channel occupancies
    load_channel_occupancies (chanx_occ, chany_occ);

//    dump_clbs_local( congestion_file );
//    print_channel_occ_file(chanx_occ, chany_occ, congestion_file );

    // Initialize channel_conge_map and marked_clbs
    for (i=0; i<=nx+1; i++) {
        for (j=0; j<=ny+1; j++) {
            channel_conge_map[i][j] = 0;
            marked_clbs[i][j] = 0;
        }
    }
    // Reset the depopulation amout that each CLB is going to be depopulated by
    for( i=0; i<num_blocks; ++i )
    {
        clb_dpop_amount[i] = 0;
    }
	
    // generate clb map to depopulate
    // X-Directed
         for (j=0;j<=ny;j++) {
        for (i=1;i<=nx;i++) {
        	printf("%d ", chanx_occ[i][j]);
        	printf("%d ", chany_occ[i][j]);
        }
     }
    for (j=0;j<=ny;j++) {
        for (i=1;i<=nx;i++) {
            // mark all adjacent blocks namely blocks [i][j] and [i][j+1]
            if (i>0 && i<nx+1 && j>0 && j<ny+1) {
                if ((int)(channel_conge_map[i][j]) < chanx_occ[i][j] ) {
                    channel_conge_map[i][j] = chanx_occ[i][j];
                }					
            }
            if (i>0 && i<nx+1 && j+1>0 && j+1<ny+1) {
                if ((int)(channel_conge_map[i][j]) < chanx_occ[i][j] ) {
                    channel_conge_map[i][j+1] = chanx_occ[i][j];
                } // if	
            }  // if
        }  // for
    }  // for
    // Y-Directed
    for (j=1;j<=ny;j++) {
        for (i=0;i<=nx;i++) {
            // mark all adjacent blocks namely [i][j] and [i+1][j]
            if (i>0 && i<nx+1 && j>0 && j<ny+1) {
                if ((int)(channel_conge_map[i][j]) < chany_occ[i][j] ) {
                    channel_conge_map[i][j] = chany_occ[i][j];
                }					
            }
            if (i+1>0 && i+1<nx+1 && j>0 && j<ny+1) {
                if ((int)(channel_conge_map[i][j]) < chany_occ[i][j] ) {
                    channel_conge_map[i+1][j] = chany_occ[i][j];
                }	// if
            } // if
        } // for
    } // for



	print_channel_conge_file(congestion_file);
    // Output the MAX CW
    max_chan_width = 0;
   for (j=ny+1; j>=0; j--) {	
            for (i=0; i<=nx+1; i++) {		
                
                // We only care about clbs which have not been chosen yet.
                //
                //printf("%2d ",channel_conge_map[i][j]);
                    if (channel_conge_map[i][j] > max_chan_width) {
                        max_chan_width = channel_conge_map[i][j];
                    }
            } // for
            //printf("\n");
        } // for
    
    // Fix for an infinite loop bug where unroutable but max cw is less than or = to constraint
    if( route_success == FALSE && max_chan_width <= cw_constraint )
    {
        cw_constraint -=5;
    }
    else if ( route_success == FALSE && ( ((float)max_chan_width /(float)cw_constraint) > 1.1f ) )
    {
        cw_constraint -= 2;  // overconstraint by 10 more CW
    }
    
    printf("-------------------------------------------\n");
    printf("The maximum channel width is %d\n",max_chan_width);
    printf("The channel width constraint is %d\n",cw_constraint);
    printf("-------------------------------------------\n");
    
    // Added a multiloop done 
    //

    while ( multiloop_done == 0 )
    //for( bug=0; bug<1; ++bug )
    {
        // Output the channel_conge_map and calculate the maximum channel width
        max_chan_width = 0;
        num_total_clbs = 0;
        num_total_subblocks = 0;
        printf("clb_map: nx = %d, ny = %d\n",nx,ny);
        for (j=ny+1; j>=0; j--) {	
            for (i=0; i<=nx+1; i++) {		
                
                // We only care about clbs which have not been chosen yet.
                //
                //printf("%2d ",channel_conge_map[i][j]);
                if( marked_clbs[i][j] == 0 )
                    {
                    if (channel_conge_map[i][j] > max_chan_width) {
                        max_chan_width = channel_conge_map[i][j];
                    }
                    if (channel_conge_map[i][j] != 0) {
                        num_total_clbs++;
                        bnum = clb[i][j].u.block;
                        num_total_subblocks+= block[bnum].num_subblocks;
                    }  // if 
                } // if !marked
            } // for
            //printf("\n");
        } // for

        printf("-------------------------------------------\n");
        printf("The current channel width is %d\n",max_chan_width);
        
        // For the multi peak version, we stop when we have chosen to depopulate all CLBs w/ cw > constraint
        if( max_chan_width <= cw_constraint )
            {
            multiloop_done = 1;
            continue;
            }
        // Find all the peaks now.  
        num_peaks = 0;
        for (j=ny+1; j>=0; j--) {	
            for (i=0; i<=nx+1; i++) {	
                if (channel_conge_map[i][j] == max_chan_width && marked_clbs[i][j] == 0 ) {
                    peaks[num_peaks].x = i;
                    peaks[num_peaks].y = j;
                    num_peaks++;
                }
            }
        }
        printf("There are %i number of peaks in this placement\n", num_peaks);
        if (num_peaks == 0) {
            printf("ERROR: congestion.c, max peak could not be found again\n");
            exit(1);
        }

        // Calculate manhattan distance from center and take that as the peak.
        // Find the peak closes to the center
        max_manhattan_dist = nx + ny;
        congestion_center.x = -1;
        congestion_center.y = -1;
        for (i=0; i<num_peaks; i++) {
            man_dist = abs(nx/2-peaks[i].x)+abs(ny/2-peaks[i].y);
            if (man_dist < max_manhattan_dist) {
                congestion_center.x = peaks[i].x;
                congestion_center.y = peaks[i].y;
                max_manhattan_dist = abs(nx/2-peaks[i].x)+abs(ny/2-peaks[i].y);
            }
        }

        printf("Peak is at %d %d\n",congestion_center.x, congestion_center.y);

        num_empty_clbs = ((nx+2)*(ny+2)) - 4 - num_blocks + nx + ny + 1;  // expand by 1 row or column?? need to be modified
        if(num_empty_clbs<=0) num_empty_clbs = 1;
        done = FALSE;
        iteration=1;
        while (!done && iteration < nx/congestion_radius) {
            // Calculate the total number of subblocks in the total number of clbs
            num_region_subblocks = 0;
            num_region_clbs = 0;
            for (i=congestion_center.x-iteration; i<=congestion_center.x+iteration; i++) {
                for (j=congestion_center.y-iteration; j<=congestion_center.y+iteration; j++) {

                    // Force to be inside radius of a circle, but the count is a square.
                    dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
                        abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );

                    if (i>0 && i<nx+1 && j>0 && j<ny+1 && dist<iteration) {
                        if (clb[i][j].occ != 0 && marked_clbs[i][j] == 0 ) {	   // only add clbs which have not been marked for other regions
                            bnum = clb[i][j].u.block;
                            num_region_subblocks += block[bnum].num_subblocks;
                            num_region_clbs++;					
                        }  // if
                    } // if
                } // for i
            } // for j

            // Determine the new number of clbs needed after depopulation
            dpop_val = floor ( ((float)num_region_subblocks / (float)num_region_clbs) - 0.2 );
            num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;

            if (dpop_val == 0) {
                printf("No more depopulation is possible\n");
                //multiloop_done = 1;
                dpop_val = 1;
                done = 1;
                continue;
            }

            // Multiple loop should have a smaller congestion radius.
            //
            /*
            if (num_new_clbs > num_empty_clbs) {  // VERY BAD to have this
                done = TRUE;
                iteration--;
            } else */ if (num_region_clbs == num_total_clbs) {
                done = TRUE;
            } else {
                iteration++;
            }

        }  // while !done or at radius

        // Calculate dpop value for iteration value 
        // Why do we recalculate the value? If the loop exited, the values would have already been there!!
        // Ask marvin?
        num_region_subblocks = 0;
        num_region_clbs = 0;
        for (i=congestion_center.x-iteration; i<=congestion_center.x+iteration; i++) {
            for (j=congestion_center.y-iteration; j<=congestion_center.y+iteration; j++) {
                dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
                    abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );
                if (i>0 && i<nx+1 && j>0 && j<ny+1 && dist<iteration) {
                    if (clb[i][j].occ != 0 && marked_clbs[i][j] == 0 ) {
                        bnum = clb[i][j].u.block;
                        num_region_subblocks += block[bnum].num_subblocks;
                        num_region_clbs++;
                    }
                }
            }
        }
        dpop_val = floor ( ((float)num_region_subblocks / (float)num_region_clbs) - 0.2 );
        if( dpop_val == 0 ) dpop_val = 1;
        num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;

        // decrease dpop_val until it can be decreased no more
        // do not use this for multiple loop as this will depopulate
        // Need a new formula for multi-block selection
        done = FALSE;
        /*  // Old depopulation growth control scheme
        while (!done) {
            dpop_val--;
            if (dpop_val != 0) {
                num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;
                if (num_new_clbs > num_empty_clbs) {
                    dpop_val++;
                    num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;
                    done = TRUE;
                }
            } else {
                dpop_val++;
                num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;
                done = TRUE;
            }
        }
        */
       
        // New depopulation control scheme
        cwc_over_amount = (float) max_chan_width / (float)cw_constraint;
        if( cwc_over_amount > 1.5f ) cwc_over_amount = 1.5f;                                // Cap @50% max depop
        if( cwc_over_amount < 1.025f ) cwc_over_amount = 1.025f;                                //  Cap @2.5% min depop 
        dpop_growth_amount = 0.25* congestion_radius * ceil(num_region_clbs * ( cwc_over_amount -1 ));  // Depopulate at least 10%
        done = FALSE;
        while( !done )
        {
            // Can't depopulate if we only have 1 clb
            if( dpop_val == 1 || dpop_val == 0 )
            {
                done = TRUE;  
            }
            // If we can still depopulate some more
            else if( num_new_clbs< dpop_growth_amount )
            {
            dpop_val--;  // depopulate 1 more LE from each CLB
            num_new_clbs = ceil ((float)num_region_subblocks / (float)dpop_val) - num_region_clbs;
            }
            // Otherwise we're over the depop_growth_amount factor
            else
            {
            done = TRUE;;
            }
        }
        // End new depopulation control scheme
        total_num_new_clbs += num_new_clbs;

        avg_cluster_size = (float)num_total_subblocks / (float)num_total_clbs;
        printf("avg_cluster_size %f, num_total_clbs %d, num_total_subblocks %d\n",avg_cluster_size,num_total_clbs, num_total_subblocks);
        printf("radius %d, dpop_val %d, num_region_subblocks %d, num_region_clbs %d, num_new_clbs %d, num_empty_clbs %d\n",iteration,dpop_val,num_region_subblocks,num_region_clbs,num_new_clbs,num_empty_clbs);
        printf("-------------------------------------------\n");
       

        //	// dpop_val is the maximum value to increase depop region by 1 column and 1 row.
        //	num_region_subblocks = 0;
        //	num_region_clbs = 0;
        //	for (i=congestion_center.x-congestion_radius; i<=congestion_center.x+congestion_radius; i++) {
        //		for (j=congestion_center.y-congestion_radius; j<=congestion_center.y+congestion_radius; j++) {
        //			dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
        //							     abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );
        //			if (i>0 && i<nx+1 && j>0 && j<ny+1 && dist<congestion_radius) {
        //				if (clb[i][j].occ != 0) {
        //					bnum = clb[i][j].u.block;
        //					num_region_subblocks += block[bnum].num_subblocks;
        //					num_region_clbs++;
        //				}
        //			}
        //		}
        //	}
        //	avg_cluster_size = (float)num_region_subblocks / (float)num_region_clbs;
        //	old_dimension = sqrt(num_region_clbs);
        //	dpop_val = (int) (((float) num_region_subblocks / ((old_dimension+1.0)*(old_dimension+1.0))) + 0.5);
        //
        //	printf("congestion_radius=%d, num_region_subblocks=%d, num_region_clbs=%d, avg_cluster_size=%f, old_dimension=%f, dpop_val=%d\n",congestion_radius,num_region_subblocks, num_region_clbs, avg_cluster_size, old_dimension, dpop_val);
        //
        //	// Output region to the congestion file
        //	cg_fp = my_fopen (congestion_file, "w", 0);
        //	fprintf(cg_fp,".CLUSTER_SIZE %d\n",dpop_val);
        //	for (i=congestion_center.x-congestion_radius; i<=congestion_center.x+congestion_radius; i++) {
        //		for (j=congestion_center.y-congestion_radius; j<=congestion_center.y+congestion_radius; j++) {
        //			dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
        //							     abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );
        //			if (i>0 && i<nx+1 && j>0 && j<ny+1 && dist<congestion_radius) {
        //				if (clb[i][j].occ != 0) {
        //					bnum = clb[i][j].u.block;
        //					fprintf(cg_fp,".CLB %s\n", block[bnum].name);
        //				}
        //			}
        //		}
        //	}
        //	fclose(cg_fp);

        // Output region to the congestion file
        fprintf(cg_fp,".CLUSTER_SIZE %d\n",dpop_val);
        for (i=congestion_center.x-iteration; i<=congestion_center.x+iteration; i++) {
            for (j=congestion_center.y-iteration; j<=congestion_center.y+iteration; j++) {
                dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
                    abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );
                if (i>0 && i<nx+1 && j>0 && j<ny+1 && dist<iteration) {
                    if (clb[i][j].occ != 0 && marked_clbs[i][j] == 0 ) {
                        bnum = clb[i][j].u.block;
                        fprintf(cg_fp,".CLB %s\n", block[bnum].name);

                        // Set up the clb marked lists
                        marked_clbs[i][j] = 1;                             // this CLB has been marked for a particular region
                        if(  clb_dpop_amount[bnum] != 0 )
                        {
                            printf("Trying to depopulate a CLB twice!\n");
                            exit( 0 );
                        }
                        clb_dpop_amount[bnum] = dpop_val; // The value that this CLB will be depopulated by
                        total_clbs_chosen ++;
                        total_subblocks_chosen += block[bnum].num_subblocks;
                    }
                }
            }
        }

        // Exit if we've chosen > 20% of the clbs or subblocks, limit the amount we are depopulating
        // Do we keep here?
        printf("The current total number of CLBs marked: %i\n", total_clbs_chosen);
        printf("The current total of new CLBs created %i\n", total_num_new_clbs );
        if( total_clbs_chosen > (num_blocks*0.25)  )
        {
            multiloop_done = 1;
        }

        // Print out the congestion map  for debug purposes
#ifdef _PRINT
        printf("congestion_map: nx = %d, ny = %d\n",nx,ny);
        for (j=ny+1; j>=0; j--) {	
            for (i=0; i<=nx+1; i++) {		
                dist = floor (sqrt ( abs(congestion_center.x-i)*abs(congestion_center.x-i) +
                    abs(congestion_center.y-j)*abs(congestion_center.y-j) ) );
                if (dist<iteration) {
                    printf("%2d ",channel_conge_map[i][j]);
                } else {
                    printf(" 0 ");
                }
            }
            printf("\n");
        }
#endif 
    }

     // Print out a depopulation map at the end to see how we are depopulating
     // overall
    printf("----------------------------------------------------------\n");
    printf("Depopulation Map\n");
    printf("----------------------------------------------------------\n");
    for (j=ny; j>0; j--) 
    {	
        for (i=1; i<=nx; i++)		
        {
        if( clb[i][j].occ == 1 ) 
            {
            bnum = clb[i][j].u.block;
            printf("%2d ", clb_dpop_amount[bnum] );
            }
        else
            {
            printf(" n ");
            }
        }
        printf("\n");
    }
    printf( "Number of CLBs going to be depopulated %i num subblocks %i , total number of CLBs: %i\n", total_clbs_chosen, total_subblocks_chosen, num_blocks );
    
     // Print out the congestion map  for debug purposes
#ifdef _DEBUG
        printf("congestion_map: nx = %d, ny = %d\n",nx,ny);
        for (j=ny+1; j>=0; j--) {	
            for (i=0; i<=nx+1; i++) {		
                if( clb[i][j].occ ==1 && clb[i][j].type == CLB)
                {
                    bnum = clb[i][j].u.block;
                    if (clb_dpop_amount[bnum] != 0 ) {
                        printf("%3d ",channel_conge_map[i][j]);
                    } else {
                        printf(" 0  ");
                    }
                }
                else 
                {
                    printf(" 0  ");
                }
            }
            printf("\n");
        }
#endif 

    old_conge = congestion;
    congestion = CHAN_WIDTH;
    calc_conge();
    congestion = old_conge;
    // Cleanup
    free_matrix (chanx_occ, 1, nx, 0, sizeof(int));
    free_matrix (chany_occ, 0, nx, 1, sizeof(int));    
    free_matrix(marked_clbs, 0, nx+1, 0, sizeof(int));
    free( clb_dpop_amount );
    free(peaks);
    fclose(cg_fp);  

}


// *******************************************************************************************
// generate_congestion_map (auto6)
//
// This version decreases max_dop_val 1 by 1 until the number of new CLBs increases the array
// size by a factor of 1. 
// 
// depop_val = floor (average_channel_width - 0.2);
//
// *******************************************************************************************
//void generate_congestion_map(char *congestion_file, int cw_constraint)
//{
//	int i,j,k;
//	int x,y,bnum;
//	int **chanx_occ;   /* [1..nx][0..ny] */
//	int **chany_occ;   /* [0..nx][1..ny] */
//	int **channel_conge_map;	   /* [0..nx+1][[0..ny+1] */
//	int **marked_clbs; /* [0..nx+1][[0..ny+1] */
//
//	// Queue variables
//	int *Qx, *Qy;
//	int QHead, QTail;
//
//	int num_total_clbs;
//	int num_dpop_clbs;
//
//	int max_chan_width;
//	int num_region_subblocks;
//	int max_dpop_chan_width;
//	int num_new_clbs;
//	int dpop_val;
//	boolean marked_clbs_done;
//
//	int *temp;
//
//	FILE *cg_fp;
//
//	chanx_occ = (int **) alloc_matrix (1, nx, 0, ny, sizeof(int));
//	chany_occ = (int **) alloc_matrix (0, nx, 1, ny, sizeof(int));
//	channel_conge_map = (int **) alloc_matrix (0, nx+1, 0, ny+1, sizeof(int));
//	marked_clbs = (int **) alloc_matrix (0, nx+1, 0, ny+1, sizeof(int));
//	
//	Qx = (int *) my_malloc((nx+2)*(ny+2)*sizeof(int));
//	Qy = (int *) my_malloc((nx+2)*(ny+2)*sizeof(int));
//	QHead = 0;
//	QTail = 0;
//
//	// Load the channel occupancies
//	load_channel_occupancies (chanx_occ, chany_occ);
//	print_channel_occ_file(chanx_occ, chany_occ);
//
//	printf("The channel width constraint is %d\n",cw_constraint);
//	
//	// Initialize channel_conge_map and marked_clbs
//	for (i=0; i<=nx+1; i++) {
//		for (j=0; j<=ny+1; j++) {
//			channel_conge_map[i][j] = 0;
//			marked_clbs[i][j] = 0;
//		}
//	}
//
//	// generate clb map to depopulate
//	// X-Directed
//	for (j=0;j<=ny;j++) {
//		for (i=1;i<=nx;i++) {
//			// mark all adjacent blocks namely blocks [i][j] and [i][j+1]
//			if (i>0 && i<nx+1 && j>0 && j<ny+1) {
//				if (channel_conge_map[i][j] < chanx_occ[i][j] && clb[i][j].occ != 0) {
//					channel_conge_map[i][j] = chanx_occ[i][j];
//				}					
//			}
//			if (i>0 && i<nx+1 && j+1>0 && j+1<ny+1) {
//				if (channel_conge_map[i][j+1] < chanx_occ[i][j] && clb[i][j+1].occ != 0) {
//					channel_conge_map[i][j+1] = chanx_occ[i][j];
//				}	
//			}
//
//		}
//	}
//	// Y-Directed
//	for (j=1;j<=ny;j++) {
//		for (i=0;i<=nx;i++) {
//			// mark all adjacent blocks namely [i][j] and [i+1][j]
//			if (i>0 && i<nx+1 && j>0 && j<ny+1) {
//				if (channel_conge_map[i][j] < chany_occ[i][j] && clb[i][j].occ != 0) {
//					channel_conge_map[i][j] = chany_occ[i][j];
//				}					
//			}
//			if (i+1>0 && i+1<nx+1 && j>0 && j<ny+1) {
//				if (channel_conge_map[i+1][j] < chany_occ[i][j] && clb[i+1][j].occ != 0) {
//					channel_conge_map[i+1][j] = chany_occ[i][j];
//				}	
//			}
//		}
//	}
//
//	// Output the channel_conge_map and calculate the maximum channel width
//	max_chan_width = 0;
//	num_total_clbs = 0;
//	printf("channel_conge_map:\n");
//	for (j=ny+1; j>=0; j--) {	
//		for (i=0; i<=nx+1; i++) {		
//			printf("%2d ",channel_conge_map[i][j]);
//			if (channel_conge_map[i][j] > max_chan_width) {
//				max_chan_width = channel_conge_map[i][j];
//			}
//			if (channel_conge_map[i][j] != 0) {
//				num_total_clbs++;
//			}
//		}
//		printf("\n");
//	}
//
//	printf("The maximum channel width is %d\n",max_chan_width);
//
//	// channel_conge_map contains the cluster map.  Generate marked_clbs now so that the array size increases by only 1
//	marked_clbs_done = FALSE;
//	max_dpop_chan_width = max_chan_width;
//	while (!marked_clbs_done) {
//		// Count how many blocks in the channel_conge_map have values >= max_dpop_chan_width
//		// Also count how many subblocks they have.
//
//		// Generate marked_clb map
//		for (i=1; i<=nx; i++) {
//			for (j=1; j<=ny; j++) {
//				if (channel_conge_map[i][j] >= max_dpop_chan_width) {
//					marked_clbs[i][j] = channel_conge_map[i][j];
//				} else {
//					marked_clbs[i][j] = 0;
//				}
//			}
//		}
//
////		printf("marked_clbs:\n");
////		for (j=ny+1; j>=0; j--) {	
////			for (i=0; i<=nx+1; i++) {		
////				printf("%2d ",marked_clbs[i][j]);
////			}
////			printf("\n");
////		}
//
//		num_new_clbs = 0;
//		num_dpop_clbs = 0;
//		// calculate how much our array size will increase here
//		for (i=0; i<=nx+1; i++) {
//			for (j=0; j<=ny+1; j++) {
//				if (marked_clbs[i][j] > 0) {
//					// Mark the first found element and put it in the Q
//					QHead = 0; QTail = 0;
//					Qx[QTail] = i; Qy[QTail++] = j;
//					bnum = clb[i][j].u.block;
//					num_region_subblocks = block[bnum].num_subblocks;
//					marked_clbs[i][j] = 0;
////					printf("found root block %d, %d\n",i,j);	
//
//					while (QHead != QTail) {
//						// Consider all neighbours of the head elements.
//						x = Qx[QHead];
//						y = Qy[QHead];
//						if (x-1>0) {
//							if (marked_clbs[x-1][y] > 0) {
//								// add [x-1][u] to the queue
//								Qx[QTail] = x-1; Qy[QTail++] = y;
//								bnum = clb[x-1][y].u.block;
//								num_region_subblocks += block[bnum].num_subblocks;
//								marked_clbs[x-1][y] = 0;
////								printf("\t add block %d, %d\n",x-1,y);
//							}
//						}
//						if (x+1<nx+1) {
//							if (marked_clbs[x+1][y] > 0) {
//								// add [x+1][u] to the queue
//								Qx[QTail] = x+1; Qy[QTail++] = y;
//								bnum = clb[x+1][y].u.block;
//								num_region_subblocks += block[bnum].num_subblocks;
//								marked_clbs[x+1][y] = 0;
////								printf("\t add block %d, %d\n",x+1,y);
//							}
//						}
//						if (y-1>0) {
//							if (marked_clbs[x][y-1] > 0) {
//								// add [x][y-1] to the queue
//								Qx[QTail] = x; Qy[QTail++] = y-1;
//								bnum = clb[x][y-1].u.block;
//								num_region_subblocks += block[bnum].num_subblocks;
//								marked_clbs[x][y-1] = 0;
////								printf("\t add block %d, %d\n",x,y-1);
//							}
//						}
//						if (y+1<ny+1) {
//							if (marked_clbs[x][y+1] > 0) {
//								// add [x][y+] to the queue
//								Qx[QTail] = x; Qy[QTail++] = y+1;
//								bnum = clb[x][y+1].u.block;
//								num_region_subblocks += block[bnum].num_subblocks;
//								marked_clbs[x][y+1] = 0;
////								printf("\t add block %d, %d\n",x,y+1);
//							}
//						}
//						QHead++;
//					}
//
////					printf("num_region_subblocks = %d, QTail = %d\n",num_region_subblocks,QTail);
//					dpop_val = floor ( ((float)num_region_subblocks / (float)QTail) - 0.2 );
////					printf("dpop_val = %d\n",dpop_val);
//					num_new_clbs += ceil ((float)num_region_subblocks / (float)dpop_val) - QTail;
////					printf("num_new_clbs = %d\n",num_new_clbs);
//					num_dpop_clbs += QTail;
//				}
//			}
//		}
//
////		printf("total number of new_clbs = %d\n",num_new_clbs);
////		printf("num_dpop_clbs = %d,num_total_clbs = %d\n",num_dpop_clbs,num_total_clbs);
////		printf("******************\nIteration done!\n*************************\n");
//
//		// Check if num_new_clbs exceeds nx + 1 + ny;
//		if (num_new_clbs > nx + 1 + ny) {
//			max_dpop_chan_width++;
//			marked_clbs_done = TRUE;
//		} else if (num_dpop_clbs == num_total_clbs) {
//			marked_clbs_done = TRUE;
//		} else {
//			max_dpop_chan_width--;
//		}
//
//	}
//
//	if (max_dpop_chan_width > max_chan_width) {
//		max_dpop_chan_width = max_chan_width;
//	}
//
//	printf("max_dpop_chan_width = %d\n",max_dpop_chan_width);
//
//	// Now output congestion map to file.
//	cg_fp = my_fopen (congestion_file, "w", 0);
//
//	// Generate marked_clb map
//	for (i=1; i<=nx; i++) {
//		for (j=1; j<=ny; j++) {
//			if (channel_conge_map[i][j] >= max_dpop_chan_width) {
//				marked_clbs[i][j] = channel_conge_map[i][j];
//			} else {
//				marked_clbs[i][j] = 0;
//			}
//		}
//	}
//
//	// Find subregions
//	for (i=0; i<=nx+1; i++) {
//		for (j=0; j<=ny+1; j++) {
//			if (marked_clbs[i][j] > 0) {
//				// Mark the first found element and put it in the Q
//				QHead = 0; QTail = 0;
//				Qx[QTail] = i; Qy[QTail++] = j;
//				bnum = clb[i][j].u.block;
//				num_region_subblocks = block[bnum].num_subblocks;
//				marked_clbs[i][j] = 0;
//
//				while (QHead != QTail) {
//					// Consider all neighbours of the head elements.
//					x = Qx[QHead];
//					y = Qy[QHead];
//					if (x-1>0) {
//						if (marked_clbs[x-1][y] > 0) {
//							// add [x-1][u] to the queue
//							Qx[QTail] = x-1; Qy[QTail++] = y;
//							bnum = clb[x-1][y].u.block;
//							num_region_subblocks += block[bnum].num_subblocks;
//							marked_clbs[x-1][y] = 0;
//						}
//					}
//					if (x+1<nx+1) {
//						if (marked_clbs[x+1][y] > 0) {
//							// add [x+1][u] to the queue
//							Qx[QTail] = x+1; Qy[QTail++] = y;
//							bnum = clb[x+1][y].u.block;
//							num_region_subblocks += block[bnum].num_subblocks;
//							marked_clbs[x+1][y] = 0;
//						}
//					}
//					if (y-1>0) {
//						if (marked_clbs[x][y-1] > 0) {
//							// add [x][y-1] to the queue
//							Qx[QTail] = x; Qy[QTail++] = y-1;
//							bnum = clb[x][y-1].u.block;
//							num_region_subblocks += block[bnum].num_subblocks;
//							marked_clbs[x][y-1] = 0;
//						}
//					}
//					if (y+1<ny+1) {
//						if (marked_clbs[x][y+1] > 0) {
//							// add [x][y+] to the queue
//							Qx[QTail] = x; Qy[QTail++] = y+1;
//							bnum = clb[x][y+1].u.block;
//							num_region_subblocks += block[bnum].num_subblocks;
//							marked_clbs[x][y+1] = 0;
//						}
//					}
//					QHead++;
//				}
//
//				dpop_val = floor ( ((float)num_region_subblocks / (float)QTail) - 0.2 );
//				fprintf(cg_fp,".CLUSTER_SIZE %d\n",dpop_val);
//				for (k=0; k<QTail; k++) {
//					x = Qx[k];
//					y = Qy[k];
//					bnum = clb[x][y].u.block;
//					fprintf(cg_fp,".CLB %s\n", block[bnum].name);
//				}	
//			}
//		}
//	}
//
//	fclose(cg_fp);
//
//	free_matrix (chanx_occ, 1, nx, 0, sizeof(int));
//	free_matrix (chany_occ, 0, nx, 1, sizeof(int));
//	free_matrix (channel_conge_map, 0, nx+1, 0, sizeof(int));
//	free_matrix (marked_clbs, 0, nx+1, 0, sizeof(int));
//	free(Qx);
//	free(Qy);
//}




// *******************************************************************************************
// generate_congestion_map (auto5)
//
// This version marks all clbs that are greater than the channel width constraint and 
// depopulates all of those blocks.
// 
// depop_val = floor (average_channel_width - 0.2);
//
// Max partition size can be nx or nx/4
//
// *******************************************************************************************
//void generate_congestion_map(char *congestion_file, int cw_constraint)
//{
//	int i,j,k,l,m;
//	int x,y,bnum;
//	int **chanx_occ;   /* [1..nx][0..ny] */
//	int **chany_occ;   /* [0..nx][1..ny] */
//	int **marked_clbs; /* [0..nx+1][[0..ny+1] */
//	int *Qx, *Qy;
//	int QHead, QTail;
//	int largest_cw;
//	int largest_csize;
//	int depop_val;
//	int csize_sum;
//	float csize_avg;
//	int max_chan_width;
//
//	int xmin, xmax, ymin, ymax, xsize, ysize;
//	int partition_cnt;
//	int xpartition,ypartition;
//	int max_partition;
//
//	FILE *cg_fp;
//
//	chanx_occ = (int **) alloc_matrix (1, nx, 0, ny, sizeof(int));
//	chany_occ = (int **) alloc_matrix (0, nx, 1, ny, sizeof(int));
//	marked_clbs = (int **) alloc_matrix (0, nx+1, 0, ny+1, sizeof(int));
//	
//	Qx = (int *) my_malloc((nx+2)*(ny+2)*sizeof(int));
//	Qy = (int *) my_malloc((nx+2)*(ny+2)*sizeof(int));
//	QHead = 0;
//	QTail = 0;
//
//	// Load the channel occupancies
//	load_channel_occupancies (chanx_occ, chany_occ);
//	print_channel_occ_file(chanx_occ, chany_occ);
//
//	printf("The channel width constraint is %d\n",cw_constraint);
//
//	// Clear all channels that have widths smaller than the constraint
//	// X-directed
//	for (j=ny;j>=0;j--) {
//		for (i=1;i<=nx;i++) {
//			if (chanx_occ[i][j] <= cw_constraint) {
//				chanx_occ[i][j] = 0;
//			}
//		}
//	}
//	// Y-directed
//	for (j=ny;j>=1;j--) {
//		for (i=0;i<=nx;i++) {
//			if (chany_occ[i][j] <= cw_constraint) {
//				chany_occ[i][j] = 0;
//			}
//		}
//	}
//	
//	// Initialize marked_clbs
//	for (i=0; i<=nx+1; i++) {
//		for (j=0; j<=ny+1; j++) {
//			marked_clbs[i][j] = 0;
//		}
//	}
//
//	// generate clb map to depopulate
//	// X-Directed
//	for (j=0;j<=ny;j++) {
//		for (i=1;i<=nx;i++) {
//			if (chanx_occ[i][j] > cw_constraint) {
//				// mark all adjacent blocks namely blocks [i][j] and [i][j+1]
//				if (i>0 && i<nx+1 && j>0 && j<ny+1) {
//					if (marked_clbs[i][j] < chanx_occ[i][j] && clb[i][j].occ != 0) {
//						marked_clbs[i][j] = chanx_occ[i][j];
//					}					
//				}
//				if (i>0 && i<nx+1 && j+1>0 && j+1<ny+1) {
//					if (marked_clbs[i][j+1] < chanx_occ[i][j] && clb[i][j+1].occ != 0) {
//						marked_clbs[i][j+1] = chanx_occ[i][j];
//					}	
//				}
//			}
//		}
//	}
//	// Y-Directed
//	for (j=1;j<=ny;j++) {
//		for (i=0;i<=nx;i++) {
//			if (chany_occ[i][j] > cw_constraint) {
//				// mark all adjacent blocks namely [i][j] and [i+1][j]
//				if (i>0 && i<nx+1 && j>0 && j<ny+1) {
//					if (marked_clbs[i][j] < chany_occ[i][j] && clb[i][j].occ != 0) {
//						marked_clbs[i][j] = chany_occ[i][j];
//					}					
//				}
//				if (i+1>0 && i+1<nx+1 && j>0 && j<ny+1) {
//					if (marked_clbs[i+1][j] < chany_occ[i][j] && clb[i+1][j].occ != 0) {
//						marked_clbs[i+1][j] = chany_occ[i][j];
//					}	
//				}
//			}
//		}
//	}
//
//	// Output all the marked clbs
//	max_chan_width = 0;
//	printf("marked clbs:\n");
//	for (j=ny+1; j>=0; j--) {	
//		for (i=0; i<=nx+1; i++) {		
//			printf("%2d ",marked_clbs[i][j]);
//			if (marked_clbs[i][j] > max_chan_width) {
//				max_chan_width = marked_clbs[i][j];
//			}
//		}
//		printf("\n");
//	}
//
//	printf("The maximum channel width is %d\n",max_chan_width);
//
//	// I should have a list of all the clbs that need to be de-populated.
//	// go though all adjacent elements and put the coordinates in Qx and Qy
//	// when we are
//	cg_fp = my_fopen (congestion_file, "w", 0);
//	for (i=0; i<=nx+1; i++) {
//		for (j=0; j<=ny+1; j++) {
//			if (marked_clbs[i][j] > 0) {
//
//				// Mark the first found element and put it in the Q
//				QHead = 0; QTail = 0;
//				Qx[QTail] = i; Qy[QTail++] = j;
//
//				largest_cw = marked_clbs[i][j];
//				marked_clbs[i][j] = 0;
//
//				bnum = clb[i][j].u.block;
//				largest_csize = block[bnum].num_subblocks;
////				csize_sum = block[bnum].num_subblocks;
//
//				while (QHead != QTail) {
//
//					// Consider all neighbours of the head elements.
//					x = Qx[QHead];
//					y = Qy[QHead];
//					// check [x-1][y]
//					if (x-1>0) {
//						if (marked_clbs[x-1][y] > 0) {
//							// put clb in the Q
//							Qx[QTail] = x-1; Qy[QTail++] = y;						
//							if (marked_clbs[x-1][y] > largest_cw) {
//								largest_cw = marked_clbs[x-1][y];
//							}	
//							marked_clbs[x-1][y] = 0;
//							bnum = clb[x-1][y].u.block;
////							csize_sum += block[bnum].num_subblocks;
//							if (block[bnum].num_subblocks > largest_csize) {
//								largest_csize = block[bnum].num_subblocks;
//							}
//						}
//					}
//					// check [x+1][y]
//					if (x+1<nx+1) {
//						if (marked_clbs[x+1][y] > 0) {
//							Qx[QTail] = x+1; Qy[QTail++] = y;						
//							if (marked_clbs[x+1][y] > largest_cw) {
//								largest_cw = marked_clbs[x+1][y];
//							}	
//							marked_clbs[x+1][y] = 0;
//							bnum = clb[x+1][y].u.block;
////							csize_sum += block[bnum].num_subblocks;
//							if (block[bnum].num_subblocks > largest_csize) {
//								largest_csize = block[bnum].num_subblocks;
//							}
//						}
//					}
//					// check [x][y-1]
//					if (y-1>0) {
//						if (marked_clbs[x][y-1] > 0) {
//							Qx[QTail] = x; Qy[QTail++] = y-1;						
//							if (marked_clbs[x][y-1] > largest_cw) {
//								largest_cw = marked_clbs[x][y-1];
//							}	
//							marked_clbs[x][y-1] = 0;
//							bnum = clb[x][y-1].u.block;
////							csize_sum += block[bnum].num_subblocks;
//							if (block[bnum].num_subblocks > largest_csize) {
//								largest_csize = block[bnum].num_subblocks;
//							}
//						}
//					}
//					// check [x][y+1]
//					if (y+1<ny+1) {
//						if (marked_clbs[x][y+1] > 0) {
//							Qx[QTail] = x; Qy[QTail++] = y+1;							
//							if (marked_clbs[x][y+1] > largest_cw) {
//								largest_cw = marked_clbs[x][y+1];
//							}	
//							marked_clbs[x][y+1] = 0;
//							bnum = clb[x][y+1].u.block;
////							csize_sum += block[bnum].num_subblocks;
//							if (block[bnum].num_subblocks > largest_csize) {
//								largest_csize = block[bnum].num_subblocks;
//							}
//						}
//					}
//					QHead++;
//				} // end while (QHead != QTail)
//
//				// Determine the bounding box of the constrained region
//				xmin = nx;
//				xmax = 0;
//				ymin = ny;
//				ymax = 0;
//				for (k=0; k<QTail; k++) {
//					x = Qx[k];
//					y = Qy[k];
//					if (x<xmin)
//						xmin = x;
//					if (x>xmax)
//						xmax = x;
//					if (y<ymin)
//						ymin = y;				
//					if (y>ymax)
//						ymax = y;					
//				}
//
//				xsize = xmax - xmin + 1;
//				ysize = ymax - ymin + 1;
//
////				fflush(stdout);
////				printf("xmin=%d, xmax = %d, ymin=%d, ymax = %d, xsize = %d, ysize = %d\n",xmin,xmax,ymin,ymax,xsize,ysize);
////				printf("nx=%d, ny=%d\n",nx,ny);
////				fflush(stdout);
//
//				max_partition = ceil((float)nx/4);
////				max_partition = nx;
//
////				printf ("max partition = %d\n",max_partition);
//				if (xsize > max_partition) {
//					xpartition = floor ((float) xsize / (float)(ceil ((float)xsize / (float)max_partition)));
//				} else {
//					xpartition = xsize;
//				}
//				if (ysize > max_partition) {
//					ypartition = floor ((float) ysize / (float)(ceil ((float)ysize / (float)max_partition)));
//				} else {
//					ypartition = ysize;
//				}
////				xpartition = xsize;
////				ypartition = ysize;
//	
////				printf("xpartition=%d, ypartition=%d\n",xpartition,ypartition);
////				fflush(stdout);
//
//				if (xpartition > max_partition || ypartition > max_partition) {
//					printf("ERROR: Congestion.c, xpartition or ypartition too large\n");
//					fflush(stdout);
//					exit(1);
//				}
//
//				for (l=xmin; l<xmax; l+=xpartition ) {
//					for (m=ymin; m<ymax; m+=ypartition ) {
////						printf("l=%d, m=%d\n",l,m);
//						csize_sum = 0;
//						partition_cnt = 0;
//						for (k=0; k<QTail; k++) {
//							x = Qx[k];
//							y = Qy[k];
//							bnum = clb[x][y].u.block;
//							if (x>=l && x<l+xpartition && y>=m && y<m+ypartition) {
//								csize_sum += block[bnum].num_subblocks;
//								partition_cnt++;
////								printf("\tfound block x=%d, y=%d\n",x,y);
//							}
//
//						}
////						printf("partition_cnt = %d\n",partition_cnt);
////						fflush(stdout);
//						if (partition_cnt != 0) {						
//							csize_avg = csize_sum / partition_cnt;
//							depop_val = floor (csize_avg - 0.2);
//
//							if (depop_val == 0) {
//								depop_val = 1;
//							}
//							fprintf(cg_fp,".CLUSTER_SIZE %d\n",depop_val);
////							printf("depop val is %d\n",depop_val);
//							for (k=0; k<QTail; k++) {
//								x = Qx[k];
//								y = Qy[k];
//								bnum = clb[x][y].u.block;
//								if (x>=l && x< l+xpartition && y>=m && y<m+ypartition) {
//									fprintf(cg_fp,".CLB %s\n", block[bnum].name);
////									printf("\tfound block x=%d, y=%d, name %s\n",x,y,block[bnum].name);
//								}
//							}
//						}
//					}
//				}
//
//				// Output results to file
//				// Figure out what the depop value should be
////				csize_avg = csize_sum / QTail;
////				//depop_val =  floor ((float) cw_constraint / (float) largest_cw * largest_csize);
////				depop_val = floor (csize_avg - 0.2);
////				printf("csize_avg is %f, QTail is %d, depop_val is %d\n",csize_avg, QTail, depop_val);			
////				if (depop_val == largest_csize) {
////					depop_val--;
////				}
////				if (depop_val == 0) {
////					depop_val = 1;
////				}
////
////				fprintf(cg_fp,".CLUSTER_SIZE %d\n",depop_val);
////				for (k=0; k<QTail; k++) {
////					x = Qx[k];
////					y = Qy[k];
////					bnum = clb[x][y].u.block;
////					fprintf(cg_fp,".CLB %s\n", block[bnum].name);
////				}		
//
//			} // if (marke... first block
//		} // for (j...)
//	} // for (i...)
//
//	fclose(cg_fp);
//
//	printf("done\n");
//	fflush(stdout);
//
////	dump_clbs_local();
//
//	free_matrix (chanx_occ, 1, nx, 0, sizeof(int));
//	free_matrix (chany_occ, 0, nx, 1, sizeof(int));
//	free(Qx);
//	free(Qy);
//}
//
