/*
*  Written by: David Leong
*  Implementation of quartus routines
*
*
*
*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "draw.h"
#include "place.h"
#include "quartus.h"

/* Find a block in the list */
int  Find_block_num( char* name )
{
    int i;
    for( i=0; i<num_blocks; ++i )
    {
        if( strcmp( name, block[i].name ) == 0 )
            return i;
    }
    return -1;
}
/* parse my own placement format */
void parse_quartus_placement( char* in_file )
{
    /* Variables */
	int i,j,tempInt, tempInt2;
	char buf[BUFSIZE];
	char* token;
	char seps[]   = " ,\t\n";
	int found;
	FILE* fp;
	fp = fopen( in_file, "r" );

	/* Set all initial occupancies to 0 first */
	for (i=0;i<=nx+1;i++) 
		for (j=0;j<=ny+1;j++) 
			{
			clb[i][j].occ = 0;
			}
	
	/* Go through and repopulate all the blocks */
    do
    {	
        if( fgets(buf, 500, fp ) == NULL )
        {   // EOF 		
            break;
        }

        token = strtok( buf, seps );  /* Name of the block */
        i = Find_block_num( token );
        token = strtok( NULL, seps );
        tempInt = atoi( token );   /* X */
        token = strtok ( NULL, seps );
        tempInt2 = atoi( token );  /* Y */

        /* Set CLB data structures */
        if( tempInt == 0 || tempInt == nx + 1 ||
            tempInt2 == 0 || tempInt2 == ny +1 )
        {
            /* IO block */
            clb[tempInt][tempInt2].u.io_blocks[clb[tempInt][tempInt2].occ] = i;
            clb[tempInt][tempInt2].occ++;
        }
        else  /* CLB */
        {
            clb[tempInt][tempInt2].occ = 1;
            clb[tempInt][tempInt2].u.block = i;
        }

        /* Set BLOCK data structures */
        block[i].x = tempInt;
        block[i].y = tempInt2;
        block[i].lookup = -1;

        if( tempInt < 0 || tempInt > nx+1 || tempInt2 < 0 || tempInt2 > ny+1 )
        {
            printf("Out of bounds in reading placement file from super grid\n");
        }
    } while(1);

    /* Do some error checking */
	for( i=0; i<num_blocks; ++i)
	{
		if( block[i].type == CLB )
		{
			if(clb[block[i].x][block[i].y].u.block != i )
				printf("BLOCK READ ERROR\n");
		}
		else 
		{	
			found = 0;
			for( j=0; j<clb[block[i].x][block[i].y].occ; ++j)
			{
				if( clb[block[i].x][block[i].y].u.io_blocks[j] == i )
					found = 1;
			}
			if( found == 0 )
				printf("IO block not set!\n");
		}
    }  /* For all blocks */
	fclose( fp );
}

