#!/bin/sh

# Create multiple PBS job scripts and
# submit each one from its own directory.
# Created by Marvin, Modified for David
# Script to test the effect of changing the seed

PWD=`pwd`
VPR="ivpr"                                     # VPR Directory
CCT_DIR="Incremental_FPGA_respin"                  # .net files directory
ARCH="arch.arch"                               # architecture files
SIM_SET="Inc_Baseline"                         # Which Simulation
PLACE_DIR="Placement"         
ROUTE_DIR="Routing"
ROUTE120_DIR="Routing120"
LOGFILE_DIR="Logs"
LOGFILE120_DIR="Logs120"
IMD_DIR="IMD"
OPTS="-nodisp -max_router_iterations 100"
PBS_O_WORKDIR="/global/home/davel"
PBS_O_SIMDIR="/global/home/davel/Incremental_FPGA_respin"

cd $PBS_O_WORKDIR                             #/global/home/davel
for SIMULATIONS in pdc clma ex1010 misex3 pdc spla
do

# Create PBS file here
##################################################################################
PBSNAME="BASELINE_$SIMULATIONS.pbs"
cat > pbsfiles/${PBSNAME} << EOF
#!/bin/sh
#PBS -S /bin/sh

echo "Running on host: " \`hostname\`
cd $PBS_O_SIMDIR

for CCT in ${SIMULATIONS}_p5d ${SIMULATIONS}_p10d 
do
for I in 10 1 0.5 0.25 0.125
do
for S in  1 2 3 4 5 
do

# Create the logfile name
LOGNAME="${SIM_SET}_\${CCT}_i\${I}_s\${S}"

#echo "Started at: " \`date\`
$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE_DIR/\${LOGNAME}.r $OPTS -verify_binary_search -seed \${S} -pres_fac_mult 1.3 -inner_num \${I} -place_only -dump_imd $IMD_DIR/\${LOGNAME}.imd > $LOGFILE_DIR/\${LOGNAME}.log
$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE_DIR/\${LOGNAME}.r $OPTS -verify_binary_search -seed \${S} -pres_fac_mult 1.3 -route_only > $LOGFILE_DIR/\${LOGNAME}_2.log
#echo "Finished at: " \`date\`

# Extract Channel Width and re-run with 20% increase in CW
CW=0
CW=\`grep Best $LOGFILE_DIR/\${LOGNAME}_2.log | sed -e 's/Best routing used a channel width factor of //g' -e 's/\.//g'\`
CW=\`echo "\${CW}*1.2/1" | bc\`

$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE120_DIR/\${LOGNAME}.r2 $OPTS -route_only -route_chan_width \${CW} -pres_fac_mult 1.1 > $LOGFILE120_DIR/\${LOGNAME}_120.log

done # for i
done # for s
done # for files

EOF
##################################################################################

# Submit job to the PBS queue
qsub -l walltime=125:00:00,mem=512mb pbsfiles/${PBSNAME}
done
# EOF
