#!/bin/sh

# Create multiple PBS job scripts and
# submit each one from its own directory.
# Created by Marvin, Modified for David
# Script to test the effect of changing the seed

PWD=`pwd`
VPR="ivpr"                                     # VPR Directory
ARCH="arch.arch"                               # architecture files
SIM_SET="Inc_Bench"                         # Which Simulation
PLACE_DIR="Placement"         
ROUTE_DIR="Routing"
ROUTE120_DIR="Routing120"
LOGFILE_DIR="Logs"
LOGFILE120_DIR="Logs120"
IMD_DIR="IMD"
OPTS="-nodisp -max_router_iterations 100"
PBS_O_WORKDIR="/global/home/davel"
PBS_O_SIMDIR="/global/home/davel/Incremental_FPGA_respin"

cd $PBS_O_WORKDIR                             #/global/home/davel
SIMULATIONS=$1
INITIAL_TEMP=$2

# Create PBS file here
##################################################################################
PBSNAME="BENCHMARK_$SIMULATIONS.pbs"
cat > pbsfiles/${PBSNAME} << EOF
#!/bin/sh
#PBS -S /bin/sh

echo "Running on host: " \`hostname\`
cd $PBS_O_SIMDIR

for CCT in ${SIMULATIONS}_p5d ${SIMULATIONS}_p10d 
do
for A in 7
do
for N in 0
do
for I in 1 1.5 2 2.5 3
do
for S in  1 2 3 4 5 
do

# Create the logfile name
LOGNAME="${SIM_SET}_\${CCT}_a\${A}_n\${N}_i\${I}_s\${S}"
INC_PARAM="-floorplan \${CCT}.fp -incremental ${SIMULATIONS}.imd"

#echo "Started at: " \`date\`
$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE_DIR/\${LOGNAME}.r $OPTS \${INC_PARAM} -place_only -fp_param 0 0 0.125 -seed \${S} -sgrid 50 50 0 > $LOGFILE_DIR/\${LOGNAME}.log
$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE_DIR/\${LOGNAME}.r $OPTS -verify_binary_search -seed \${S} -pres_fac_mult 1.3 -fp_param 0.\${A} \${N} 0.125 -sgrid_refinement ${INITIAL_TEMP} -inner_num \${I} -dump_imd $IMD_DIR/\${LOGNAME}.imd > $LOGFILE_DIR/\${LOGNAME}_2.log
#echo "Finished at: " \`date\`

# Extract Channel Width and re-run with 20% increase in CW
CW=0
CW=\`grep Best $LOGFILE_DIR/\${LOGNAME}_2.log | sed -e 's/Best routing used a channel width factor of //g' -e 's/\.//g'\`
CW=\`echo "\${CW}*1.2/1" | bc\`

$VPR \${CCT}.net $ARCH $PLACE_DIR/\${LOGNAME}.p $ROUTE120_DIR/\${LOGNAME}.r2 $OPTS -route_only -route_chan_width \${CW} -pres_fac_mult 1.1 > $LOGFILE120_DIR/\${LOGNAME}_120.log

done # A
done # N
done # for i
done # for s
done # for files

EOF
##################################################################################

# Submit job to the PBS queue
qsub -l walltime=125:00:00,mem=512mb pbsfiles/${PBSNAME}
# EOF
