#!/usr/local/bin/perl -w
#
# This script will add .global mclk and .input mclk to a .net file
# If the file already contains a clock, then nothing is changed.
#
use strict;
use English;
use Getopt::Std;

# no parameters $#ARGV = -1
if($#ARGV < 7) {
    print "usage: perl undopack_fast.pl <CCT.blif> <Channel_Width_Constraint> <Arch_File> <Cluster_Method> <Test_Case> <Lut_Size> <Cluster_Size> <Inputs_Per_Cluster>\n";
    exit(-1);
}

# Control and Iteration Variables
my $i;
my $prev_i;
my $command;
my @logData;
my $line;
my $done;
my $max_chan_width;

# Executable paths and variable command line parameters
my $UBCPACK;
my $VPR;
my $ARCH;
my $test_case;
my $pack_opts;
my $vpr_opts;

# pack options
my $lut_size;
my $cluster_size;
my $inputs_per_cluster;
my $cluster_method;
my $cluster_seed;

# Log file directories
my $PACKLOGFILEDIR;
my $VPRLOGFILEDIR;
my $PLACEDFILES;
my $ROUTEDFILES;
my $NETFILES;
my $IMDFILES;
my $FPFILES;
my $CONGFILES;

# Local Variables
my $blif_file;
my $channel_width_constraint;
#my $routed_channel_width;
my $CCT;
my $nx;
my $ny;
my $sx;
my $sy;
my $temp44;
my $init_t;
my $max_iterations;

#stats variables
my $start_time;
my $end_time;
my $run_time;
my $iteration_time;
my $total_time;
my $critical_path_delay;

# Get command line parameters
$blif_file = $ARGV[0];
$channel_width_constraint = $ARGV[1];
$ARCH = $ARGV[2];
$cluster_method = $ARGV[3];
$cluster_seed = $ARGV[3];
$test_case = $ARGV[4];
$lut_size = $ARGV[5];
$cluster_size = $ARGV[6];
$inputs_per_cluster = $ARGV[7];

open(fileIN,"$ARCH") or die("ERROR: Cannot open arch file $ARCH: $!\n");
close(fileIN);

# Executable paths
$UBCPACK = "ubcpack/ipack";
$VPR = "rePlace/vpr";

# Local
#$UBCPACK = "~marvint/research/src_code/ubcpack_david/ubcpack";
#$VPR = "~marvint/research/src_code/vpr_david/VPR-Floorplan-Quench-Linux/vprfpq";

$pack_opts = "-lut_size ".$lut_size." -cluster_size ".$cluster_size." -inputs_per_cluster ".$inputs_per_cluster.
				" -cluster_method ".$cluster_method." -cluster_seed ".$cluster_seed;
$vpr_opts = "-max_router_iterations 100 -pres_fac_mult 1.3 -op_delay -nodisp";	

# Log File Directories
$PACKLOGFILEDIR = "packlogfiles/${test_case}/";
$VPRLOGFILEDIR = "vprlogfiles/${test_case}/";
$PLACEDFILES = "placedfiles/${test_case}/";
$ROUTEDFILES = "routedfiles/${test_case}/";
$NETFILES = "netfiles/${test_case}/";
$IMDFILES = "imdfiles/${test_case}/";
$FPFILES = "fpfiles/${test_case}/";
$CONGFILES = "congfiles/${test_case}/";

# eliminate directory structure on file name	
$CCT = $blif_file;
while ($CCT =~ /\/(\S+).blif$/) {
	$CCT = $1 . ".blif";
}
if ($CCT =~ /(\S+).blif$/) {
	$CCT = $1;
}

# Perform initial packing
$total_time = 0;

$iteration_time = 0;
$start_time = time();
printf "Iteration 1: Packing in progress. Please wait...\n";
$command = $UBCPACK." ". $blif_file." ".$NETFILES.$CCT."_".$channel_width_constraint."_1.net ".$pack_opts." > ".$PACKLOGFILEDIR.$CCT."_".$channel_width_constraint."_1.log";
#printf("\n".$command."\n");
system($command);
$end_time = time();
$run_time = $end_time - $start_time;
$iteration_time += $run_time;
printf "Iteration 1: Packing took $run_time seconds\n";

# Perform initial VPR placement
$start_time = time();
printf "Iteration 1: VPR placement in progress. Please wait...\n";
$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_1.net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_1.placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_1.routed ".
			$vpr_opts." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_1.imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_1.cong ".
			"-channel_width_constraint ".$channel_width_constraint." -place_only"." > ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_1.log";
#printf("\n".$command."\n");
system($command);
$end_time = time();
$run_time = $end_time - $start_time;
$iteration_time += $run_time;
printf "Iteration 1: VPR placement took $run_time seconds\n";

# Perform initial VPR routing
printf "Iteration 1: VPR routing in progress. Please wait...\n";
$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_1.net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_1.placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_1.routed ".
			$vpr_opts." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_1.imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_1.cong ".
			"-channel_width_constraint ".$channel_width_constraint." -route_only"." -route_chan_width ".$channel_width_constraint." >> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_1.log";
#printf("\n".$command."\n");
system($command);
$end_time = time();
$run_time = $end_time - $start_time;
$iteration_time += $run_time;
printf "Iteration 1: VPR routing took $run_time seconds\n";

# first iteration is complete
printf "Iteration 1: Iteration time is $iteration_time seconds\n";
$total_time += $iteration_time;

# Check now if the channel width constraint has been met
# Open up the vpr log file and grep for Best
open(fileIN,"${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_1.log") or die("ERROR: Cannot open ${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_1.log: $!\n");
@logData = <fileIN>;
$nx = -1;
$ny = -1;
$temp44 = -1;
$done = "FALSE";
$max_chan_width = -1;
SEARCHLOOP: foreach $line (@logData)
{
	# Search for nx/ny
	if ($line =~ /^The circuit will be mapped into a (\S+) x (\S+) array of clbs./) {
		$nx = $1;
		$ny = $2;
	}
	# Search for 44% temperature
	if ($line =~ /^The 44 percent temperature point is: (\S+)/) {
		$temp44 = $1;
		$init_t = $temp44 * 0.1;

	}
#	# Search for routed channel width
#	if ($line =~ /^Best routing used a channel width factor of (\S+)./) {
#		$routed_channel_width = $1;
#	}

	# Check if routing was successful
	if ($line =~ /^Circuit successfully routed with a channel width factor of (\S+)./) {
		$done = "TRUE";
	}

	if ($line =~ /^The maximum channel width is (\S+)/) {
		$max_chan_width = $1;
	}


}
close(fileIN);

#if ($routed_channel_width == -1 || $nx == -1 || $ny == -1 || $temp44 == -1) {
if ($nx == -1 || $ny == -1 || $temp44 == -1 || $max_chan_width == -1) {
	printf "ERROR: Circuit could not be routed. Max_chan_width or nx or ny or 44% temp could not be extracted\n";
	exit(1);
}

#$done = "FALSE";
#if ($routed_channel_width <= $channel_width_constraint) {
#	$done = "TRUE";
#}

$prev_i = 1;
$i = 2;
$max_iterations=40;
while ($done eq "FALSE" && $i <= $max_iterations) {
		printf "Channel Width Constraint could not be met: Max_Chan_Width = $max_chan_width, Constraint = $channel_width_constraint\n";

	$sx = $nx * 2.5;
	$sy = $ny * 2.5;
	$init_t = $temp44 * 0.1;

	$iteration_time = 0;

	# Repack the circuit
	$start_time = time();
	printf "Iteration $i: Re-Packing in progress. Please wait...\n";
	$command = $UBCPACK." ". $blif_file." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$pack_opts." -incremental ".
				$CONGFILES.$CCT."_".$channel_width_constraint."_".$prev_i.".cong ".$NETFILES.$CCT."_".$channel_width_constraint."_".$prev_i.".net ".$IMDFILES.$CCT."_".$channel_width_constraint."_".$prev_i.".imd ".
				$FPFILES.$CCT."_".$channel_width_constraint."_".$i.".fp "."> ".$PACKLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
	printf($command);
	system($command);
	$end_time = time();
	$run_time = $end_time - $start_time;
	$iteration_time += $run_time;
	printf "Iteration $i: Re-Packing took $run_time seconds\n";

	# ReVPR the circuit (incremental)
#	$start_time = time();
#	printf "Iteration $i: Re-VPR w/ Incremental Placement in progress. Please wait...\n";
#	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_".$i.".routed ".
#				" -incremental ".$IMDFILES.$CCT."_".$channel_width_constraint."_".$prev_i.".imd "."-sgrid ".$sx." ".$sy." ".$init_t.
#				" -floorplan ".$FPFILES.$CCT."_".$channel_width_constraint."_".$i.".fp"." -fp_param 0.5 5 0.125 "."-place_only -nodisp ".
#				"> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
#	system($command);
#	$end_time = time();
#	$run_time = $end_time - $start_time;
#	$iteration_time += $run_time;
#	printf "Iteration $i: VPR Expansion took $run_time seconds\n";
#
#	$start_time = time();
#	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_".$i.".routed ".
#				$vpr_opts." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_".$i.".imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_".$i.".cong ".
#				"-channel_width_constraint ".$channel_width_constraint." -sgrid_refinement ".$init_t." -fp_param 0.5 5 0.125 "."-place_only ".
#				">> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
#	system($command);
#	$end_time = time();
#	$run_time = $end_time - $start_time;
#	$iteration_time += $run_time;
#	printf "Iteration $i: VPR Quench took $run_time seconds\n";
#
#	$start_time = time();
#	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_".$i.".routed ".
#				$vpr_opts." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_".$i.".imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_".$i.".cong ".
#				"-channel_width_constraint ".$channel_width_constraint." -sgrid_refinement ".$init_t." -fp_param 0.5 5 0.125 "."-route_only "."-route_chan_width ".$channel_width_constraint.
#				" >> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
#	system($command);
#	$end_time = time();
#	$run_time = $end_time - $start_time;
#	$iteration_time += $run_time;
#	printf "Iteration $i: VPR routing took $run_time seconds\n";

	# ReVPR the circuit (fast)
	$start_time = time();
	printf "Iteration $i: Re-VPR w/ -fast Placement in progress. Please wait...\n";
	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_".$i.".routed ".
				"-place_only -nodisp -fast".
				"> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
	system($command);
	print $command;
	$end_time = time();
	$run_time = $end_time - $start_time;
	$iteration_time += $run_time;
	printf "Iteration $i: -fast Placement took $run_time seconds\n";

	$start_time = time();
	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_".$i.".routed ".
				$vpr_opts." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_".$i.".imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_".$i.".cong ".
				"-channel_width_constraint ".$channel_width_constraint." -route_only "."-route_chan_width ".$channel_width_constraint.
				" >> ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_".$i.".log";
	system($command);
  print $command;
	$end_time = time();
	$run_time = $end_time - $start_time;
	$iteration_time += $run_time;
	printf "Iteration $i: VPR routing took $run_time seconds\n";


	printf "Iteration $i: Iteration time is $iteration_time seconds\n";
	$total_time += $iteration_time;

#	exit(1);

	# Check now if the channel width constraint has been met
	# Open up the vpr log file and grep for Best
	open(fileIN,"${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_${i}.log") or die("ERROR: Cannot open ${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_${i}.log: $!\n");
	@logData = <fileIN>;
	$nx = -1;
	$ny = -1;
	$max_chan_width = -1;
	SEARCHLOOP: foreach $line (@logData)
	{
		# Search for nx/ny
		if ($line =~ /^The circuit will be mapped into a (\S+) x (\S+) array of clbs./) {
			$nx = $1;
			$ny = $2;
		}
		# Check if routing was successful
		if ($line =~ /^Circuit successfully routed with a channel width factor of (\S+)./) {
			$done = "TRUE";
		}
		if ($line =~ /^The maximum channel width is (\S+)/) {
			$max_chan_width = $1;
		}

		if ($line =~ /^No more depopulation is possible/) {
			$done = "DEPOP_FAIL";
		}


	}
	close(fileIN);

	if ($nx == -1 || $ny == -1 || $max_chan_width == -1) {
		printf "ERROR: Circuit could not be routed. Routed Channel Width or nx or ny could not be extracted\n";
		exit(1);
	}

	$prev_i++;
	$i++;
	
}

# Determine what the last iteration was
$i--;
#if ($routed_channel_width <= $channel_width_constraint) {
if ($done eq "TRUE") {
	printf "Success: Channel Width Constraint Has Been Met: Max Channel Width = $max_chan_width, Constraint = $channel_width_constraint\n";

	# Copy last vprlogfile to "final"
	$command = "cp ${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_${i}.log ${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_final.log";
	system($command);

	# Reroute the final solution again with -max_router_iterations 300 and -pres_fac_mult 1.3
	printf "Performing final route with aggresive router options\n";
	$start_time = time();
	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_".$i.".placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_reroute.routed ".
				"-max_router_iterations 300 -pres_fac_mult 1.1 -op_delay -nodisp"." -dump_imd ".$IMDFILES.$CCT."_".$channel_width_constraint."_reroute.imd "." -congestion_file ".$CONGFILES.$CCT."_".$channel_width_constraint."_reroute.cong ".
				"-channel_width_constraint ".$channel_width_constraint." -sgrid_refinement ".$init_t." -fp_param 0.5 5 0.125 "."-route_only "."-route_chan_width ".$channel_width_constraint.
				" > ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_reroute.log";
	system($command);
	$end_time = time();
	$run_time = $end_time - $start_time;
	$total_time += $run_time;
	printf "Final re-route took $run_time seconds\n";

	# Replace and reroute from scratch
	printf "Replacing known solution from scratch\n";
	$start_time = time();
	$command = $VPR." ".$NETFILES.$CCT."_".$channel_width_constraint."_".$i.".net ".$ARCH." ".$PLACEDFILES.$CCT."_".$channel_width_constraint."_placedscratch.placed ".$ROUTEDFILES.$CCT."_".$channel_width_constraint."_placedscratch.routed ".$vpr_opts." -route_chan_width ".$channel_width_constraint." > ".$VPRLOGFILEDIR.$CCT."_".$channel_width_constraint."_placedscratch.log";
	system($command);
	$end_time = time();
	$run_time = $end_time - $start_time;
	$total_time += $run_time;
	printf "Place and route from scratch took $run_time seconds\n";

	# Check if place from scratch had a solution or not.
	open(fileIN,"${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_placedscratch.log") or die("ERROR: Cannot open ${VPRLOGFILEDIR}${CCT}_${channel_width_constraint}_placedscratch.log: $!\n");
	@logData = <fileIN>;
	$nx = -1;
	$ny = -1;
	$max_chan_width = -1;
	$done = "FALSE";
	SEARCHLOOP: foreach $line (@logData)
	{
		# Search for nx/ny
		if ($line =~ /^The circuit will be mapped into a (\S+) x (\S+) array of clbs./) {
			$nx = $1;
			$ny = $2;
		}
		# Check if routing was successful
		if ($line =~ /^Circuit successfully routed with a channel width factor of (\S+)./) {
			$done = "TRUE";
		}
		if ($line =~ /^The maximum channel width is (\S+)/) {
			$max_chan_width = $1;
		}
	}
	close(fileIN);
	if ($done eq "TRUE") {
		printf "Place from scratch successful with maximum channel width of $max_chan_width, channel_width_constraint = $channel_width_constraint\n";
	} else {
		printf "Place from scratch un-successful with maximum channel width of $max_chan_width, channel_width_constraint = $channel_width_constraint\n";
	}
} elsif ($done eq "DEPOP_FAIL"){ 
	printf "Failure: No solution exists, no more depopulation is possible: Max Channel Width = $max_chan_width, Constraint = $channel_width_constraint\n";
} else {
	printf "Failure: Could not find a solution in $max_iterations iterations: Max Channel Width = $max_chan_width, Constraint = $channel_width_constraint\n";
}

printf "Total run time is $total_time seconds\n";
