#include <stdio.h>
#include <string.h>
#include "util.h"
#include "vpack.h"
#include "globals.h"
#include "depth_optimal.h"
#include "hash.h"

/*------------------------------------------------------------
 *  Copyright 2005 By Marvin Tom and Guy Lemieux
 *
 * Permission to us e, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 * Permission is granted ONLY for non-commercial use.  For
 * commercial use, please contact the authors.
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
 * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
 * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
 * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
 * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *------------------------------------------------------------*/
#define TOKENS " \t\n"

struct s_ilink {int iblk; struct s_ilink *next;};

/* [0..num_marked_nets] and [0..num_marked_blocks] respectively.  List  *
 * the indices of the nets and blocks that have had their num_pins_of_  *
 * net_in_cluster and gain entries altered.                             */
static int *marked_nets, *marked_blocks;
static int num_marked_nets, num_marked_blocks;
static boolean *net_visited, *block_visited;

/* [0..num_nets-1].  How many pins of each net are contained in the *
 * currently open cluster?                 
 */
static int *num_pins_of_net_in_cluster;

static struct s_NDPoint *empty_NDPoint;

// Placement information
static nx, ny;

// **********************************************************************************
// alloc_and_init_clustering
//
//
//
// **********************************************************************************
static void alloc_and_init_clustering (int lut_size, int cluster_size)
{
	int i;

    num_pins_of_net_in_cluster = (int *) my_malloc (num_nets * sizeof(int));

    for (i=0;i<num_nets;i++) {
        num_pins_of_net_in_cluster[i] = 0;
    }

    marked_nets = (int *) my_malloc ((lut_size + 2) * cluster_size * sizeof(int));
    marked_blocks = (int *) my_malloc (num_blocks * sizeof(int));

    num_marked_nets = 0;
    num_marked_blocks = 0;

	net_visited = (boolean *) my_malloc (num_nets * sizeof(boolean));
	block_visited = (boolean *) my_malloc (num_blocks * sizeof(boolean));

	for (i=0; i<num_nets; i++) {
		net_visited[i] = FALSE;
	}
	for (i=0; i<num_blocks; i++) {
		block_visited[i] = FALSE;
	}
}

// **********************************************************************************
// clear_marked_elements
//
//
//
// **********************************************************************************
static void clear_marked_elements ()
{
	int i;

	for (i=0; i<num_marked_nets; i++) {
		net_visited[marked_nets[i]] == FALSE;
	}
	for (i=0; i<num_marked_blocks; i++) {
		block_visited[marked_blocks[i]] = FALSE;
	}
	num_marked_nets = 0;
	num_marked_blocks = 0;
}

// **********************************************************************************
// alloc_and_init_NDPoint
//
//
//
// **********************************************************************************
static void alloc_and_init_NDPoint (struct s_NDPoint **Ptr, int cluster_size, int blocknum)
{

	//printf("In alloc_and_init_NDPoint\n");
	struct s_NDPoint *LocalPtr;

	if (cluster_size < 1) {
		fprintf(stderr,"Cannot created NDPoint with a cluster size smaller than 1\n");
		exit(1);
	}

	LocalPtr = (struct s_NDPoint *) my_malloc (sizeof(struct s_NDPoint));
	LocalPtr->NDList = (int *) my_malloc(cluster_size * sizeof(int));
	LocalPtr->arrival_time = 0;
	LocalPtr->encapsulated_nets = 0;
	LocalPtr->NDList_size = 1;
	LocalPtr->NDList[0] = blocknum;
	LocalPtr->next = NULL;
	//LocalPtr->root = FALSE;
	LocalPtr->roots = (boolean *) my_malloc(cluster_size * sizeof(boolean));;

	*Ptr = LocalPtr;

};

// **********************************************************************************
// free_NDPoint
//
//
//
// **********************************************************************************
static void free_NDPoint (struct s_NDPoint *Ptr)
{
	free(Ptr->NDList);
	free(Ptr->roots);
	free(Ptr);
};

// **********************************************************************************
// combination_generator_get_next
//
//
//
// **********************************************************************************
boolean combination_generator_get_next(int* prev_array, int N, int R)
{

	int index;
	int counter;
	int i;

	if (R > N || N < 1 || R < 1) {
		printf("ERRROR in combination_generator_get_next. N = %d, R=%d.\n",N,R);
		exit(1);
	}

	index = R-1;
	counter = N-1;
	while (prev_array[index] == counter && index >= 0) {
		index--;
		counter--;
	}

	if (index >= 0) {
		prev_array[index]++;
		for (i=index+1; i<R; i++)
			prev_array[i] = prev_array[i-1] + 1;;

		return TRUE;	
	} else {
		return FALSE;
	}


}

// **********************************************************************************
// Copy_NDPoint
//
//
//
// **********************************************************************************
static void Copy_NDPoint(struct s_NDPoint *NDPoint, int arrival_time, int encapsulated_nets, int *Cset, int Cset_size)
{

	int i;

	NDPoint->arrival_time = arrival_time;
	NDPoint->encapsulated_nets = encapsulated_nets;

	for (i=0; i<Cset_size; i++)
	{
		NDPoint->NDList[i] = Cset[i];
		NDPoint->roots[i] = FALSE;
	}

	NDPoint->NDList_size = Cset_size;

}

// **********************************************************************************
// add_NDPoint
//
//
//
// **********************************************************************************
static void add_NDPoint(int root, int* Cset, int Cset_size, int inputs_per_cluster, int cluster_size, int lut_size)
{

	int i, j, bnum, nnum, src_bnum;
	int num_inputs;
	int encapsulated_nets, encapsulated_nets_ctr;
	int arrival_time;
	int Amin, Amax;
	boolean found;

	struct s_NDPoint *local_NDPoint;
	struct s_NDPoint *prev_local_NDPoint;
	struct s_NDPoint *new_NDPoint;

	// Check feasibility of the cluster first
	num_inputs = 0;
	num_marked_nets = 0;
	num_marked_blocks = 0;

	// Undo num_pin_of_net_in_cluster for all nets
//	for (i=0; i<num_nets; i++) {
//		num_pins_of_net_in_cluster[nnum] = 0;
//	}

	// Add the output nets of each block in the cluster to the block_visited array
	for (i=0; i<Cset_size; i++)
	{
		bnum = Cset[i];
		nnum = block[bnum].nets[0];
		num_pins_of_net_in_cluster[nnum]++;
		marked_nets[num_marked_nets] = nnum;
		num_marked_nets++;
		
	}

	// Count the number of inputs to this cluster
	// For each block in the cluster
	for (i=0; i<Cset_size; i++)
	{
		bnum = Cset[i];
		// For each input of a block %%%%%%
		//for (j=1; j<block[bnum].num_nets; j++)
		for (j=1; j<=lut_size; j++)
		{
			// Find the source net and source block
			nnum = block[bnum].nets[j];
			// some nets are empty
			if (nnum != OPEN) {
			
				src_bnum = net[nnum].pins[0];
				
				// Determine if input has already been added.  If not, mark the net.
				if (num_pins_of_net_in_cluster[nnum] == 0)
				{
					num_inputs++;
					marked_blocks[num_marked_blocks] = src_bnum;
					num_marked_blocks++;
				}
				num_pins_of_net_in_cluster[nnum]++;		
			}
		}
	}	

	// Calculate the total number of encapsulated nets
	encapsulated_nets = 0;
	for (i=0; i<num_marked_nets; i++)	{

		nnum = marked_nets[i];			
		if (net[nnum].num_pins == num_pins_of_net_in_cluster[nnum]) {
			encapsulated_nets++;
		}

	}	

	// marked_nets array contains all the nets that we have considered.  
	// num_pins_of_net_in_cluster contains how many of the pins of each net are in the cluster.
	// marked_blocks array contains all blocks that fanin to the current cluster.\

	if (num_inputs <= inputs_per_cluster) {

		// Calculate the Amin and Amax

		// Search through the first marked block
		bnum = marked_blocks[0];
		if (block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH || block[bnum].type == INPAD) {
			local_NDPoint = empty_NDPoint;
		} else {
			local_NDPoint = block[bnum].NDSet;
		}
		if (local_NDPoint == NULL) {
			printf("ERROR: Previously visited node does not contain a NDPoint 1\n");
			exit(1);
		}
		// Initialize Amin and Amax
		Amin = local_NDPoint->arrival_time;
		Amax = local_NDPoint->arrival_time;
		local_NDPoint = local_NDPoint->next;
		while (local_NDPoint != NULL) 
		{
			if (local_NDPoint->arrival_time > Amax)
				Amax = local_NDPoint->arrival_time;		
			local_NDPoint = local_NDPoint->next;
		}

		// Search through the rest of the marked blocks for Amin and Amax
		for (i=1; i<num_marked_blocks; i++)
		{
			bnum = marked_blocks[i];
			if (block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH || block[bnum].type == INPAD) {
				local_NDPoint = empty_NDPoint;
			} else {
				local_NDPoint = block[bnum].NDSet;
			}
			if (local_NDPoint == NULL) {
				printf("ERROR: Previously visited node does not contain a NDPoint 2 %d\n", block[bnum].type);
				exit(1);
			}
			if (local_NDPoint->arrival_time > Amin)
				Amin = local_NDPoint->arrival_time;

			while (local_NDPoint != NULL)
			{
				if (local_NDPoint->arrival_time > Amax)
					Amax = local_NDPoint->arrival_time;		
				local_NDPoint = local_NDPoint->next;
			}
		}

		// We now have Amin and Amax
		// Count the total number of encapsulated nets per arrival_time and then determine whether to add the NDPoint to the root NDSet
		for (i=Amin; i<=Amax; i++)
		{
			arrival_time = i + 1;
			encapsulated_nets_ctr = encapsulated_nets;

			// Count the number of encapsulated nets for each arrival time in the NDSet
			for (j=0; j<num_marked_blocks; j++)
			{				
				bnum = marked_blocks[j];		
				if (block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH || block[bnum].type == INPAD) {
					prev_local_NDPoint = empty_NDPoint;
				} else {
					prev_local_NDPoint = block[bnum].NDSet;
				}
				//prev_local_NDPoint = block[bnum].NDSet;				
				if (prev_local_NDPoint == NULL)
				{
					printf("ERROR in Add_NDPoint. Input block %d has no NDSet 3\n",bnum);
					exit(1);
				}
				local_NDPoint = prev_local_NDPoint->next;
				found = FALSE;
				while (found == FALSE)
				{
					if (local_NDPoint == NULL) {
						found = TRUE;
					} else if (local_NDPoint->arrival_time > arrival_time) {
						found = TRUE;
					} else {
						prev_local_NDPoint = local_NDPoint;
						local_NDPoint = local_NDPoint->next;
					}
				}

				encapsulated_nets_ctr += prev_local_NDPoint->encapsulated_nets;				
			}

			// Determine whether point should be added to NDSet
			prev_local_NDPoint = block[root].NDSet;

			if (prev_local_NDPoint == NULL) {
				// No point has been created yet
				alloc_and_init_NDPoint (&new_NDPoint, cluster_size, root);
				Copy_NDPoint(new_NDPoint, arrival_time, encapsulated_nets_ctr, Cset, Cset_size);
				block[root].NDSet = new_NDPoint;

			} else {
			
				// Previous points exist.  
				found = FALSE;
				while (found == FALSE)
				{
					local_NDPoint = prev_local_NDPoint->next;
					if (prev_local_NDPoint->arrival_time == arrival_time) {
						// Consider replacing it
						if (encapsulated_nets_ctr > prev_local_NDPoint->encapsulated_nets)
						{						
							Copy_NDPoint(prev_local_NDPoint, arrival_time, encapsulated_nets_ctr, Cset, Cset_size);
						} 
						found = TRUE;

					} else if (local_NDPoint == NULL) {
						// Insert new point at the end
						if (prev_local_NDPoint->arrival_time < arrival_time)
						{
							// Insert a point at the end
							alloc_and_init_NDPoint (&new_NDPoint, cluster_size, root);
							Copy_NDPoint(new_NDPoint, arrival_time, encapsulated_nets_ctr, Cset, Cset_size);
							prev_local_NDPoint->next = new_NDPoint;		
							found = TRUE;
						} else if (prev_local_NDPoint->arrival_time > arrival_time)	{
							// Insert a point at the beginning
							alloc_and_init_NDPoint (&new_NDPoint, cluster_size, root);
							Copy_NDPoint(new_NDPoint, arrival_time, encapsulated_nets_ctr, Cset, Cset_size);
							block[root].NDSet = new_NDPoint;
							new_NDPoint->next = prev_local_NDPoint;
							prev_local_NDPoint = new_NDPoint;							
							found = TRUE;
						} else {
							printf("ERROR: Arrival time read error\n");
						}

						// Locate next location
						local_NDPoint = prev_local_NDPoint->next;

					} else if (local_NDPoint->arrival_time > arrival_time) {
						// Insert new point between two others
						alloc_and_init_NDPoint (&new_NDPoint, cluster_size, root);
						Copy_NDPoint(new_NDPoint, arrival_time, encapsulated_nets_ctr, Cset, Cset_size);
						prev_local_NDPoint->next = new_NDPoint;
						new_NDPoint->next = local_NDPoint;
						found = TRUE;

						// Locate next location
						prev_local_NDPoint = new_NDPoint;
						local_NDPoint = new_NDPoint->next;  // Don't need this.

					}  else {
						prev_local_NDPoint = local_NDPoint;
					} // end if_else_else_else

				} // end while
			} // end if_else
			

			// Clean out the rest of the list.
			while (local_NDPoint != NULL)
			{
				if (local_NDPoint->encapsulated_nets <= prev_local_NDPoint->encapsulated_nets) {
					// delete the local point.
					prev_local_NDPoint->next = local_NDPoint->next;
					free_NDPoint(local_NDPoint);
				} else {
					prev_local_NDPoint = local_NDPoint;
				}

				local_NDPoint = prev_local_NDPoint->next;
			}// end while
			
		}// end for

	} else {// end if (num_inputs <= inputs_per_cluster)
	}

	// Undo the num_pins_of_net_in_cluster
	for (i=0; i<num_marked_nets; i++) {
		nnum = marked_nets[i];
		num_pins_of_net_in_cluster[nnum] = 0;
	}
	num_marked_nets = 0;

	// Undo blocks_visited;
	for (i=0; i<Cset_size; i++) {
		bnum = Cset[i];
	}
	for (i=0; i<num_marked_blocks; i++) {
		bnum = marked_blocks[i];
		nnum = block[bnum].nets[0];
		num_pins_of_net_in_cluster[nnum] = 0;
	}
	num_marked_blocks = 0;

}

// **********************************************************************************
// recursive_find_leaf_set
//
//
//
// **********************************************************************************
static void recursive_find_leaf_set(int* Cset, int Cset_size, boolean* Visited, int cluster_size, int num_blocks_prev_level, 
									int root, int inputs_per_cluster, int lut_size)
{
	int i,j,k;
	
    int bnum,nnum,add_bnum;
	int remaining_spaces;
	int max_blocks_this_level, num_blocks_this_level;	

	int *Cstack;
	int Cstack_size;

	int *indices;

	// Cstack keeps track of all the blocks at the current leel
	Cstack = (int *) my_malloc(num_blocks * sizeof(int));
	indices = (int *) my_malloc(cluster_size * sizeof(int));
	Cstack_size = 0;

	// Calculate total number blocks in this level and add each of these blocks to the local stack.
	// for all blocks in the previous level	
	for (i=Cset_size-1; i>=Cset_size-num_blocks_prev_level; i--) {
		bnum = Cset[i];
		// for all input nets from the previous level block
		for (j=1; j<block[bnum].num_nets; j++) {
			nnum = block[bnum].nets[j];
			if (nnum >= 0) {			
				add_bnum = net[nnum].pins[0];
				// Add block net[j].pins[0] to the Cstack only if the block is a LUT
				if (Visited[add_bnum] == FALSE && block[add_bnum].type == LUT) {			
					Cstack[Cstack_size++] = add_bnum;
					Visited[add_bnum] = TRUE;
				}
			}
		}
	}
	
	if (Cstack_size > 0) {

		//printf("Cstack_size is greater than 1!\n");		

		// Calculate how many remaining spaces are left so we know how many to choose on this level.
		remaining_spaces = cluster_size - Cset_size;
		if (remaining_spaces >= Cstack_size) {
			max_blocks_this_level = Cstack_size;
		} else {
			max_blocks_this_level = remaining_spaces;
		}

		// Go through all combinations in Cstack[0] to Cstack[Cstack_size-1]
		for (num_blocks_this_level=max_blocks_this_level; num_blocks_this_level>0; num_blocks_this_level--) {
			// Cstack_size choose num_blocks_this_level
			// Put elements into Cset[Cset_size] to Cset[Cset_size + num_blocks_this_level-1]

			// Generate first combination
			for (i=0; i<num_blocks_this_level; i++) {
				indices[i] = i;
			}

			// enumerate all other possible combinations
			do {
				for (j=0; j<num_blocks_this_level; j++)
					Cset[Cset_size+j] = Cstack[indices[j]];

				if (Cset_size+num_blocks_this_level == cluster_size) {
					//Clustering Solution has been found!
					add_NDPoint(root, Cset, cluster_size, inputs_per_cluster, cluster_size, lut_size);
				} else {
					// Recursively call itself to find a leaf set.
					recursive_find_leaf_set(Cset, Cset_size+num_blocks_this_level, Visited, cluster_size, num_blocks_this_level, root, inputs_per_cluster, lut_size);
				}
				
			} while (combination_generator_get_next(indices,Cstack_size,num_blocks_this_level));
		}
	} else {
		// Non-full clustering solution has been found
		add_NDPoint(root, Cset, Cset_size, inputs_per_cluster, cluster_size, lut_size);
	}

	// Undo all the visited nodes
	for (i=0; i<Cstack_size; i++) {
		add_bnum = Cstack[i];
		Visited[add_bnum] = FALSE;
	}

	free(Cstack);
	free(indices);
	
}

// **********************************************************************************
// create_NDSet
//
//
//
// **********************************************************************************
static void create_NDSet(int root, int cluster_size, int inputs_per_cluster, int lut_size)
{

	int i;
	int* Cset;
	boolean* Visited;
	struct s_NDPoint *local_NDPoint;

	// Consider moving these variables to depth_optimal cluster routine.
	Cset = (int *) my_malloc(cluster_size * sizeof(int));
	Visited = (boolean *) my_malloc(num_blocks * sizeof(boolean));

	// Initialize the Visited array.  This keeps track of all the blocks we have visited during our recursive search
	for (i=0; i<num_blocks; i++)
		Visited[i] = FALSE;

	Cset[0] = root;
	Visited[root] = TRUE;

	if (block[root].type == INPAD) {
		// Create a dummy NDSET with 0 arrival time and 0 num nets and 0 PDList_size
		alloc_and_init_NDPoint(&(block[root].NDSet), cluster_size, root);

	} else if (block[root].type != OUTPAD){	
		// Enumerate all possible clusters for the given root node.
		
		// Undo all the visited nodes here first.
		// Initialize the Visited array.
		for (i=0; i<num_blocks; i++)
			Visited[i] = FALSE;
		block[root].NDSet = NULL;
		while (block[root].NDSet == NULL && cluster_size > 0)
		{
			// Find a leaf set.  If no leaf sets can be created, decrease the cluster size and try again.
			recursive_find_leaf_set(Cset, 1, Visited, cluster_size, 1, root, inputs_per_cluster, lut_size);
			cluster_size--;
		}
	}

	free(Cset);
	free(Visited);
}

// **********************************************************************************
// 
//
//
//
// **********************************************************************************
struct s_NDPoint* find_NDPoint_with_closest_arrival_time(int *atime_target, int bnum)
{
	struct s_NDPoint *local_NDPoint;
	boolean found;

	local_NDPoint = block[bnum].NDSet;
	if (local_NDPoint == NULL) {
		fprintf(stderr,"Error: local_NDPoint is NULL.  Aborting.\n");
		exit (1);
	}
	
	found = FALSE;
	while (found == FALSE) {
		if (local_NDPoint->next == NULL) {
			found = TRUE;
			if (local_NDPoint->arrival_time > *atime_target) {
				fprintf(stderr,"Error: Required arrival time does not exist.  Aborting.\n");
				exit (1);
			}
		} else if (local_NDPoint->next->arrival_time > *atime_target) {
			found = TRUE;
		} else {
			local_NDPoint = local_NDPoint->next;
		}
	}

	*atime_target = local_NDPoint->arrival_time;
	return(local_NDPoint);

}

// **********************************************************************************
// is_block_member_of_cluster
//
//
//
// **********************************************************************************
boolean is_block_member_of_cluster(int bnum, int **cluster_contents, int *cluster_occupancy, int cnum)
{
	int i;

	for (i=0; i<cluster_occupancy[cnum]; i++) {
		if (cluster_contents[cnum][i] == bnum) {
			return TRUE;
		}
	}
	return FALSE;
}

// **********************************************************************************
// prune_cluster
//
//
//
// **********************************************************************************
void prune_cluster(int cnum, int prune_index, int **cluster_contents, int *cluster_occupancy, int *block_rooted_in_cluster, int lut_size)
{

	// Prune the cluster cnum starting at prune_index
	int i;
	int iblk, inet;
	int bnum, nnum, src_blk;

	struct s_ilink *list_head;
	struct s_ilink *list_tail;
	struct s_ilink *create_elem;
	boolean first;
	
	int *new_cluster;
	int new_cluster_size;
//
//	printf("pruning cluster %d, prune_index %d, prune_block %d\n",cnum, prune_index,cluster_contents[cnum][prune_index]);
//	printf("\tCluster pre-prune: ");
//	for (i=0; i<cluster_occupancy[cnum]; i++) {
//		printf("%d ",cluster_contents[cnum][i]);
//	}
//	printf("\n");
	new_cluster = (int *) my_malloc (cluster_occupancy[cnum] * sizeof (int));
	new_cluster_size = 0;

	list_head = NULL;
	list_tail = NULL;

	// Search through all the blocks in the cluster to see if each block is rooted in this cluster
	// If it is, add the block to the queues
	first = TRUE;
	for (iblk=0; iblk<cluster_occupancy[cnum]; iblk++) {
		bnum = cluster_contents[cnum][iblk];
		if (block_rooted_in_cluster[bnum] == cnum && iblk != prune_index) {
			// Add the block to the queue
			create_elem = (struct s_ilink *) my_malloc (sizeof (struct s_ilink));
			create_elem->iblk = bnum;
			create_elem->next = NULL;
			if (first == TRUE) {
				list_head = create_elem;
				list_tail = create_elem;
				first = FALSE;
			} else {
				list_tail->next = create_elem;
				list_tail = create_elem;
			}
		}
	}
	
	// Go through every element in the queue and mark all of the children.  
	// These are the blocks we need to keep
	num_marked_blocks = 0;
	while (list_head != NULL) {
		bnum = list_head->iblk;
		// don't process the block unless it is a member of this cluster
		if (is_block_member_of_cluster(bnum,cluster_contents,cluster_occupancy,cnum)) {
			// mark the block
			block_visited[bnum] = TRUE;
			marked_blocks[num_marked_blocks++] = bnum;
			// Add all the input blocks of bnum to the queue
			for (inet=1; inet<=lut_size; inet++) {
				nnum = block[bnum].nets[inet];
				if (nnum != -1) {
					src_blk = net[nnum].pins[0];
					if (block_visited[src_blk] == FALSE) {
						create_elem = (struct s_ilink *) my_malloc (sizeof (struct s_ilink));
						create_elem->iblk = src_blk;
						create_elem->next = NULL;
						list_tail->next = create_elem;
						list_tail = create_elem;
					}
				}
			}
		} else {
		}
		
		// Delete the head element
		create_elem = list_head;
		list_head = list_head->next;
		free(create_elem);
	}
	
	// prune_index is not longer rooted anywhere
	bnum = cluster_contents[cnum][prune_index];
	if (block_rooted_in_cluster[bnum] == cnum) {
		// Block bnum was rooted in cnum and is no longer rooted anywhere else
		block_rooted_in_cluster[bnum] = -1;
	}

	// Put all of the other root elements (excluding the prune_index) into the new_cluster
	for (iblk=0; iblk<cluster_occupancy[cnum]; iblk++) {
		bnum = cluster_contents[cnum][iblk];
		if (block_rooted_in_cluster[bnum] == cnum && iblk != prune_index) {
			new_cluster[new_cluster_size++] = bnum;
		}
	}

	// Check if original root is still in the cluster
	// This has to be done because if the primary root needs to stay in the cluster, it can't be the first entry anymore.
	bnum = cluster_contents[cnum][prune_index];
	if (block_visited[bnum] == TRUE) {
		new_cluster[new_cluster_size++] = bnum;
	}
	
	// Put all the remaining marked elements into the new_cluster
	for (iblk=0; iblk<cluster_occupancy[cnum]; iblk++) {
		bnum = cluster_contents[cnum][iblk];
		if (block_rooted_in_cluster[bnum] != cnum && block_visited[bnum] == TRUE && bnum != cluster_contents[cnum][prune_index]) {
			new_cluster[new_cluster_size++] = bnum;
		}
	}

	// Copy new cluster into cnum
	for (iblk=0; iblk<new_cluster_size; iblk++) {
		cluster_contents[cnum][iblk] = new_cluster[iblk];
	}
	cluster_occupancy[cnum] = new_cluster_size;

	// Undo all the marked blocks
	for (iblk=0; iblk<num_marked_blocks; iblk++) {
		block_visited[marked_blocks[iblk]] = FALSE;
	}
	num_marked_blocks = 0;

//	printf("\tCluster post-prune: ");
//	for (i=0; i<cluster_occupancy[cnum]; i++) {
//		printf("%d ",cluster_contents[cnum][i]);
//	}
//	printf("\n");

	// Pruning complete
	free(new_cluster);

}

// **********************************************************************************
// rooted_cluster_elimination
//
//
//
// **********************************************************************************
void rooted_cluster_elimination(int **cluster_contents, int *cluster_occupancy, int **cluster_roots, int *num_clusters_ptr, 
								int max_arrival_time, int lut_size)
{
	int i,j;
	int icluster, iblk, inet;
	int bnum, nnum, cnum, src_blk;
	int *cluster_arrival;
	int smallest_arrival_time;
	int num_clusters;

	// temp variables
	int num_roots;
	int atime;

	struct s_ilink *list_head;
	struct s_ilink *list_tail;
	struct s_ilink *create_elem;

	boolean unused_cluster_root_exists;
	boolean first;

	// Block is in a cluster. It's an int because it also tells me what the arrival time is.
	int **block_in_cluster_arrival; // This tells me two things.  
									// 1. the block is in the indexed cluster. 
									// 2. the arrival time of its in the indexed cluster
	int *block_rooted_in_cluster;
	boolean *net_in_use;

	num_clusters = *num_clusters_ptr;

	cluster_arrival = (int *) my_malloc (num_clusters * sizeof (int));
	block_in_cluster_arrival = (int **) alloc_matrix (0,num_blocks-1,0,num_clusters-1,sizeof(int));	
	block_rooted_in_cluster = (int *) my_malloc (num_blocks * sizeof (int));
	net_in_use = (boolean *) my_malloc (num_nets * sizeof (boolean));

	printf("Starting rooted cluster_elimination. Please wait...\n");

//	for (i=0; i<num_clusters; i++) {
//		printf("root_begin - Cluster %d: ",i);
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//		}
//		printf("\n");
//	}

	// Initialize block_in_cluster
	for (i=0; i<num_clusters; i++) {
		atime = -1;
		for (iblk=0; iblk<cluster_occupancy[i]; iblk++) {			
			if (cluster_roots[i][iblk] > 0) {
				atime = cluster_roots[i][iblk];
			}
		}
		if (atime < 0) {
			fprintf(stderr,"Cluster %d arrival time is negative = %d!\n",i,cluster_arrival[i]);
			exit(1);
		}
		cluster_arrival[i] = atime;

	}
	for (i=0; i<num_blocks; i++) {
		for (j=0; j<num_clusters; j++) {
			block_in_cluster_arrival[i][j] = -1;
		}
	}

	// Populate block_in_cluster arrays
	for (icluster=0; icluster<num_clusters; icluster++) {
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			bnum = cluster_contents[icluster][iblk];
			// block bnum is in cluster icluster with arrival cluster_arrival[icluster]
			block_in_cluster_arrival[bnum][icluster] = cluster_arrival[icluster];
		} 
		
	}

	// Choose 1 root per block
	for (iblk=0; iblk<num_blocks; iblk++) {

		smallest_arrival_time = max_arrival_time + 1;
		block_rooted_in_cluster[iblk] = -1;
		// Search for the cluster with the smallest arrival time
		for (icluster=0; icluster<num_clusters; icluster++) {
			if (block_in_cluster_arrival[iblk][icluster] >= 0) {
				// block found in cluster with smaller arrival
				if (block_in_cluster_arrival[iblk][icluster] < smallest_arrival_time) {
					smallest_arrival_time = block_in_cluster_arrival[iblk][icluster];
					block_rooted_in_cluster[iblk] = icluster;
				}
				// block found in cluster with same arrival
				if (block_in_cluster_arrival[iblk][icluster] == smallest_arrival_time) {
					// Choose one or the other

					// For now just check if iblk/icluster is not a primary root.  If it isn't, then make it a root
					// Find a better method for tie breaking here in the future
					if (cluster_contents[icluster][0] != iblk) {
						smallest_arrival_time = block_in_cluster_arrival[iblk][icluster];
						block_rooted_in_cluster[iblk] = icluster;	
					}				
				}
			}
		}
	}

//	for (i=0; i<num_blocks; i++) {
//		printf("\tpre - block %d rooted in cluster %d\n",i,block_rooted_in_cluster[i]);
//	}

	// go and find a primary root block that is not needed.
	do {
	
		unused_cluster_root_exists = FALSE;

		// reset net_in_use 
		for (inet=0; inet<num_nets; inet++) {
			net_in_use[inet] = FALSE;
		}

		// First, mark all of the nets that are currently in use.
		num_marked_nets = 0;
		for (icluster=0; icluster<num_clusters; icluster++) {
			num_marked_blocks = 0;
			num_marked_nets = 0;
			// Mark all the blocks and output nets in the cluster
			for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
				bnum = cluster_contents[icluster][iblk];
				marked_blocks[num_marked_blocks++] = bnum;
				block_visited[bnum] = TRUE;
				nnum = block[bnum].nets[0];
				marked_nets[num_marked_nets++] = nnum;
				net_visited[nnum] = TRUE;
			}
			// Check all the inputs now
			for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
				bnum = cluster_contents[icluster][iblk];
				if (block[bnum].type != INPAD) {				
					for (inet=1; inet<=lut_size; inet++) {
						nnum = block[bnum].nets[inet];
						if (nnum != -1) {
							if (net_visited[nnum] == FALSE && net_in_use[nnum] == FALSE) {
								net_in_use[nnum] = TRUE;
							}
						}
					}
				}
			}
			// Check all primary output nets
			for (iblk=0; iblk<num_blocks; iblk++) {
				if (block[iblk].type == OUTPAD) {
					nnum = block[iblk].nets[0];
					net_in_use[nnum] = TRUE;
				}
			}
			// Undo marked nets and marked blocks
			for (iblk=0; iblk<num_marked_blocks; iblk++) {
				block_visited[marked_blocks[iblk]] = FALSE;
			}
			num_marked_blocks = 0;
			for (inet=0; inet<num_marked_nets; inet++) {
				net_visited[marked_nets[inet]] = FALSE;
			}
			num_marked_nets = 0;
		}

		// Search through all the clusters for a primary root not in use
		for (icluster=0; icluster<num_clusters; icluster++) {
			if (cluster_occupancy[icluster] != 0) {
				bnum = cluster_contents[icluster][0];
				nnum = block[bnum].nets[0];
//				printf("on cluster %d, first block %d, rooted in cluster %d\n",icluster,bnum,block_rooted_in_cluster[bnum]);
				// Check for primary root clusters not in use.
				if (net_in_use[nnum] == FALSE || (net_in_use[nnum] == TRUE && block_rooted_in_cluster[bnum] != icluster)) {
					unused_cluster_root_exists = TRUE;
//					printf("primary root cluster not in use\n");
					prune_cluster(icluster, 0, cluster_contents, cluster_occupancy, block_rooted_in_cluster, lut_size);					
				} else {
					// Check for secondary clusters not in use
					for (iblk=1; iblk<cluster_occupancy[icluster]; iblk++) {
						bnum = cluster_contents[icluster][iblk];
						nnum = block[bnum].nets[0];
						if (net_in_use[nnum] == FALSE && block_rooted_in_cluster[bnum] == icluster ) {
							unused_cluster_root_exists = TRUE;
//							printf("Secondary cluster not in use, index = %d, block = %d, net = %d\n",iblk,bnum,nnum);
							prune_cluster(icluster, iblk, cluster_contents, cluster_occupancy, block_rooted_in_cluster, lut_size);
						}
					}
				}
			}
		}

	} while (unused_cluster_root_exists);

	// Go through all of the clusters and eliminate any dead nodes.
//	printf("Starting dead node elimination\n");
	for (icluster=0; icluster<num_clusters; icluster++) {

//		printf("\tcluster %d: ",icluster);
//		fflush(stdout);
//		for (i=0; i<cluster_occupancy[icluster]; i++) {
//			printf("%d_%d ",cluster_contents[icluster][i], cluster_roots[icluster][i]);
//		}

		list_head = NULL;
		list_tail = NULL;

		// Add all of the output blocks to the queue.
		first = TRUE;
//		printf("Add all the output blocks\n");
//		fflush(stdout);
		num_marked_blocks = 0;
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			bnum = cluster_contents[icluster][iblk];
			if (block_rooted_in_cluster[bnum] == icluster) {
//				printf("a - Added block %d\n",bnum);
				create_elem = (struct s_ilink *) my_malloc (sizeof (struct s_ilink));
				create_elem->iblk = bnum;
				create_elem->next = NULL;
				if (first == TRUE) {
					list_head = create_elem;
					list_tail = create_elem;
					first = FALSE;
				} else {
					list_tail->next = create_elem;
					list_tail = list_tail->next;
				}
				// Mark bnum here
				marked_blocks[num_marked_blocks++] = bnum;
				block_visited[bnum] = TRUE;
			}
		}

//		printf("Process list\n");
//		fflush(stdout);
		while (list_head != NULL) {

			// Mark the block
			bnum = list_head->iblk;
//			printf("got block %d with input blocks: ",bnum);

			// Process all inputs now
			for (inet=1; inet<=lut_size; inet++) {
				nnum = block[bnum].nets[inet];
				if (nnum != -1) {
					src_blk = net[nnum].pins[0];
//					printf("%d",src_blk);
					if (is_block_member_of_cluster(src_blk,cluster_contents,cluster_occupancy,icluster) && block_visited[src_blk] == FALSE) {
						// Add it to the queue
//						printf("_add ");
						create_elem = (struct s_ilink *) my_malloc (sizeof (struct s_ilink));
						create_elem->iblk = src_blk;
						create_elem->next = NULL;
						list_tail->next = create_elem;
						list_tail = list_tail->next;		
						// Mark src_blk here
						marked_blocks[num_marked_blocks++] = src_blk;
						block_visited[src_blk] = TRUE;
					} else {
//						printf("_na ");
					}					
				}
			}
//			printf("\n");

			// Delete head element and move on
			create_elem = list_head;
			list_head = list_head->next;
			free(create_elem);
		}

		// If there are any unmarked blocks, eliminate them!
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			bnum = cluster_contents[icluster][iblk];
			if (block_visited[bnum] == FALSE) {
				// Delete it
//				printf("I found an unused block in cluster %d!\n",icluster);
//				exit(1);
				prune_cluster(icluster, iblk, cluster_contents, cluster_occupancy, block_rooted_in_cluster, lut_size);	
				icluster--;
				break;
			}
		}

		// Undo all the marked blocks
		for (iblk=0; iblk<num_marked_blocks; iblk++) {
			block_visited[marked_blocks[iblk]] = FALSE;
		}
		num_marked_blocks = 0;

//		printf("\n");

	} // end dead_code_elimination

	// If the net is in use, it must be rooted somewhere.
	// If the net is not in use then it can't be rooted somewhere.
	for (inet=0; inet<num_nets; inet++) {
		if (net_in_use[inet] == TRUE) {
			bnum = net[inet].pins[0];
			if (block_rooted_in_cluster[bnum] == -1 && block[bnum].type != INPAD) {
				fprintf(stderr,"ERROR, Net %d, source block %d, reported in use but is not rooted in any cluster %d\n", inet,bnum,block_rooted_in_cluster[bnum]);
				exit(1);
			}
		}
		if (net_in_use[inet] == FALSE) {
			bnum = net[inet].pins[0];
			if (block_rooted_in_cluster[bnum] >= 0) {
				fprintf(stderr,"ERROR, Net %d, source block %d, reported not in use but is rooted in cluster %d\n", inet,bnum,block_rooted_in_cluster[bnum]);
				exit(1);
			}
		}
	}

	// Undo the cluster roots now. I shouldn't have to do this once I've fixed the other function
	fflush(stdout);
	for (icluster=0; icluster<num_clusters; icluster++) {
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			cluster_roots[icluster][iblk] = -1;
		}
	}

	// Set all the cluster roots now
	for (inet=0; inet<num_nets; inet++) {
		if (net_in_use[inet] == TRUE) {
			bnum = net[inet].pins[0];
			if (block[bnum].type != INPAD) {				
				cnum = block_rooted_in_cluster[bnum];
				for (iblk=0; iblk<cluster_occupancy[cnum]; iblk++) {
					if (cluster_contents[cnum][iblk] == bnum) {
						cluster_roots[cnum][iblk] = cluster_arrival[cnum];
					}
				}
			} 
		}
	}


	// This needs to go at the end because it screws up block_rooted_in_cluster
	// Stop moving it.  You are causing yourself so much damn trouble.
	// Compact the rest of the clusters
	for (i=0; i<num_clusters; i++) {
		if (cluster_occupancy[i] == 0) {
			j = i+1;
			while (cluster_occupancy[j] == 0 && j < num_clusters) {
				j++;
			}
			if (j == num_clusters) {
				break;
			} else {
				// copy cluster[j] to cluster[i];
				for (iblk=0; iblk<cluster_occupancy[j]; iblk++) {
					cluster_contents[i][iblk] = cluster_contents[j][iblk];
					cluster_roots[i][iblk] = cluster_roots[j][iblk];
				}
				cluster_occupancy[i] = cluster_occupancy[j];
				cluster_occupancy[j] = 0;
//				printf("Moving cluster %d to %d\n",j,i);
			}

		}
	}
	num_clusters = i;

//	for (i=0; i<num_clusters; i++) {
//		printf("root_end - Cluster %d: ",i);
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//		}
//		printf("\n");
//	}

	free(cluster_arrival);
	free_matrix(block_in_cluster_arrival,0,num_blocks-1,0,sizeof(int));
	free(block_rooted_in_cluster);
	free(net_in_use);

	*num_clusters_ptr = num_clusters;

}

// **********************************************************************************
// alloc_and_load_cluster_info_depth_optimal
//
//
//
// **********************************************************************************
void alloc_and_load_cluster_info_depth_optimal (int ***cluster_contents_ptr,
        int **cluster_occupancy_ptr, int ***cluster_roots_ptr,
		int cluster_size, int *num_clusters_ptr, int ***block_in_cluster_w_arrival_ptr, 
		int **num_clusters_block_in_ptr, int *max_arrival_time_ptr, int inputs_per_cluster, int lut_size) 
{

	// Cluster contents
    int **cluster_contents;  /* [0..num_clusters-1][0..cluster_size-1] */
    int *cluster_occupancy;  /* [0..num_clusters-1] */
	int **cluster_roots;  /* [0..num_clusters-1][0..cluster_size-1] */

	// ********************************
	// This is not really necessary anymore after I implemented rooted cluster elimination but
	// doing it does not impact rooted cluster elmination so I will leave it in.  

	// internal arrays global
	int *multiple_root_arrival; // This keeps track of which blocks have multiple roots and the lowest arrival time that must be used.

	// internal arrays only valid in the current iteration
	int *root_block_arrival;  // this keeps track of which blocks are roots and its arrival times in the current iteration
	// This is a weird array. first index is the block number.  second index is the arrival time.  value stored is
	// which cluster the block with the given arrival is in.  but the cluster is not indexed by cluster number.  it 
	// is indexed by the primary root of the cluster
	int **block_in_cluster_w_arrival;  /* [0..num_blocks-1][0..max_arrival_time-1] */

	// ********************************

	// post iteration arrays
	int *num_clusters_block_in; /* [0..num_blocks-1] */

	// stats
	int single_node_clusters;
	int node_cnt;
	int **cluster_occ_stats;
	int *cluster_num_inputs;
	int *cluster_num_outputs;
	int num_inputs,num_outputs;

	int bnum, nnum, src_bnum, atime_target, iblk, icluster, inet;
	int rooted_bnum;
	boolean first;
	int max_arrival_time;
	struct s_NDPoint *local_NDPoint;
	struct s_NDPoint *local_NDPoint2;
	int num_clusters, cluster_cnt;
	int i,j;
	boolean found;
	int root_cnt;

	boolean* Visited;
	boolean multiple_root_exists;	

	struct s_linked_elem {int bnum; int arrival_time; struct s_linked_elem *next;};

	struct s_linked_elem *linked_head;
	struct s_linked_elem *linked_tail;
	struct s_linked_elem *linked_ptr;

	struct s_NDPoint *NDPoint_ptr; 

	Visited = (boolean*) my_malloc (num_blocks * sizeof (boolean));
	multiple_root_arrival = (int *) my_malloc (num_blocks * sizeof (int));
	root_block_arrival = (int *) my_malloc (num_blocks * sizeof (int));

	printf("In alloc_and_load_cluster_info_depth_optimal\n");
	
	// Initialize the multiple root block
	for (bnum=0; bnum<num_blocks; bnum++) {
		multiple_root_arrival[bnum] = -1;
	}

	// Determine what the maximum arrival time is
	max_arrival_time = 0;
	for (bnum=0; bnum<num_blocks; bnum++) {
		if (block[bnum].type == OUTPAD || block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH) {			
			if (block[bnum].type == OUTPAD) {
				nnum = block[bnum].nets[0];
				src_bnum = net[nnum].pins[0];
			} else if (block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH) {
				src_bnum = bnum;
			} else {
				fprintf(stderr,"ERROR: Block type error in alloc_and_load_cluster_info_depth_optimal\n");
				exit(1);
			}
		
			local_NDPoint = block[src_bnum].NDSet;
			if (local_NDPoint == NULL) {
				printf("ERROR in alloc_and_load_cluster_info_depth_optimal. local_NDPoint == NULL\n");
				exit(1);
			}
			if (local_NDPoint->arrival_time > max_arrival_time) {
				max_arrival_time = local_NDPoint->arrival_time;
			}
		}
	}

	// Create block_in_cluster_w_arrival
	block_in_cluster_w_arrival = (int **) alloc_matrix (0,num_blocks-1,0,max_arrival_time,sizeof(int));  /* [0..num_blocks-1][0..max_arrival_time-1] */

	// Do all this until no multiple root clusters remain
	// *****************************************************************************************

	do {
	
//		printf("Starting cluster creation process\n");

		multiple_root_exists = FALSE;

		// Set all the root block arrival times to -1
		for (bnum=0; bnum<num_blocks; bnum++) {
			root_block_arrival[bnum] = -1;
		}

		// Set all the root flags in all NDPoints to FALSE
		for (bnum=0; bnum<num_blocks; bnum++) {
			NDPoint_ptr = block[bnum].NDSet;
			while (NDPoint_ptr != NULL) {
				for (i=0; i<NDPoint_ptr->NDList_size; i++) {
					NDPoint_ptr->roots[i] = FALSE;
				}
				NDPoint_ptr = NDPoint_ptr->next;
			}
		}
		// Clear block_in_cluster_w_arrival
		for (i=0; i<num_blocks; i++) {
			//printf("block %d: ",i);
			for (j=0; j<=max_arrival_time; j++) {
				//printf("%d ",j);
				block_in_cluster_w_arrival[i][j] = -1;
			}
		}

		// Add all of the outputs and LUTs and LUT_AND_LATCHs to the search Q
//		printf("Adding outputs and LUTs and LUT_AND_LATCHs to the search Q\n");		
		first = TRUE;
		//max_arrival_time = 0;
		for (bnum=0; bnum<num_blocks; bnum++) {
			
			Visited[bnum] = FALSE;

			if (block[bnum].type == OUTPAD || block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH) {			
				if (block[bnum].type == OUTPAD) {
					nnum = block[bnum].nets[0];
					src_bnum = net[nnum].pins[0];
				} else if (block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH) {
					src_bnum = bnum;
				} else {
					fprintf(stderr,"ERROR: Block type error in alloc_and_load_cluster_info_depth_optimal\n");
					exit(1);
				}

				// Add src_bnum to the queue;
				linked_ptr = (struct s_linked_elem *) my_malloc (sizeof(struct s_linked_elem));
				linked_ptr->bnum = src_bnum;
				linked_ptr->arrival_time = -1;
				linked_ptr->next = NULL;
				if (first == TRUE) {
					linked_head = linked_ptr;
					linked_tail = linked_ptr;
					first = FALSE;	
				} else {
					linked_tail->next = linked_ptr;
					linked_tail = linked_ptr;
				}
				
				// Calculate max_arrival
				local_NDPoint = block[src_bnum].NDSet;
				if (local_NDPoint == NULL) {
					fprintf(stderr,"Error: local_NDPoint for block %d is NULL.  Aborting.\n", src_bnum);
					exit (1);
				}

//				if (local_NDPoint->arrival_time > max_arrival_time) {
//					max_arrival_time = local_NDPoint->arrival_time;
//				}
			}
		}

		// Update the max arrival times for all elements already in the search queue
//		printf("Update the max arrival times for all elements already in the search queue\n");
		linked_ptr = linked_head;
		while (linked_ptr != NULL) {
			linked_ptr->arrival_time = max_arrival_time;
			linked_ptr = linked_ptr->next;
		}

		// Go through the search queue
//		printf("Go through the search queue\n");
		num_clusters = 0;
		while (linked_head != NULL) {
			// Only process the element if no multiple root element exists
			if (multiple_root_exists == FALSE) {
				// removing the head element from the search queue				
				bnum = linked_head->bnum;				
				//printf("Got bnum = %d from the queue\n",bnum);
				if (multiple_root_arrival[bnum] >= 0) {
					atime_target = multiple_root_arrival[bnum];
				} else {
					atime_target = linked_head->arrival_time;		
				}
//				printf("Processing node %d, atime_target = %d, root_block_arrival = %d, block_in_cluster_w_arrival = %d\n",bnum,atime_target,root_block_arrival[bnum],block_in_cluster_w_arrival[bnum][atime_target]);
				local_NDPoint = find_NDPoint_with_closest_arrival_time(&atime_target, bnum);

				if (root_block_arrival[bnum] >= 0 && root_block_arrival[bnum] > atime_target) {
//					printf("\tBlock is already rooted with an arrival time that greater than the target\n");
					// Block is already rooted with an arrival time greater than the target
					// Check if the root is a primary root.  If so, start over again
					if (block_in_cluster_w_arrival[bnum][root_block_arrival[bnum]] == bnum) {
						// Double primary root exists. Start again
//						printf("\tDouble primary root exists, Start again\n");
						if (root_block_arrival[bnum] < atime_target) {
							multiple_root_arrival[bnum] = root_block_arrival[bnum];
						} else {
							multiple_root_arrival[bnum] = atime_target;
						}
//						multiple_root_arrival[bnum] = atime_target;
						multiple_root_exists = TRUE;
					} else {
						// Undo the root setting in the other cluster
						//***************************************************************
//						printf("\t Undo secondary root in alternate cluser %d and create new cluster\n",block_in_cluster_w_arrival[bnum][atime_target]);
//						fflush(stdout);
						rooted_bnum = block_in_cluster_w_arrival[bnum][root_block_arrival[bnum]];
						local_NDPoint2 = block[rooted_bnum].NDSet;
						while (local_NDPoint2->arrival_time != root_block_arrival[bnum]) {
							if (local_NDPoint2 == NULL) {
								fprintf(stderr,"ERROR in alloc_and_load_cluster_info_depth_optimal. local_NDPoint is NULL\n");
								exit(1);
							}
							local_NDPoint2 = local_NDPoint2->next;
						}
						// find bnum in local_NDPoint->NDList
						for (i=0; i<local_NDPoint2->NDList_size; i++) {
							if (local_NDPoint2->NDList[i] == bnum) {
								local_NDPoint2->roots[i] = FALSE;								
							}
						}
						//***************************************************************

						// Create a new cluster rooted at bnum
						//***************************************************************
						rooted_bnum = bnum;
						for (i=0; i<local_NDPoint->NDList_size; i++) {
							iblk = local_NDPoint->NDList[i];
							if (block_in_cluster_w_arrival[iblk][atime_target] == -1) {
								block_in_cluster_w_arrival[iblk][atime_target] = rooted_bnum;
								//printf("%d %d = %d, ", iblk, atime_target, rooted_bnum);
							}
						}
						// Set block number bnum as a root
						block_in_cluster_w_arrival[bnum][atime_target] = rooted_bnum;
						root_block_arrival[bnum] = atime_target;
						local_NDPoint->roots[0] = TRUE;

						num_clusters++;
//						printf("primary cluster created at block %d\n",bnum);
						num_marked_blocks = 0;

						// Mark as Visited all the blocks in the cluster
						for (i=0; i<local_NDPoint->NDList_size; i++) {
							bnum = local_NDPoint->NDList[i];
							Visited[bnum] = TRUE;
							marked_blocks[num_marked_blocks++] = bnum;
						}

						// Add all the inputs to the queue
						for (i=0; i<local_NDPoint->NDList_size; i++) {
							bnum = local_NDPoint->NDList[i];
							for (j=1; j<block[bnum].num_nets; j++) {
								
								nnum = block[bnum].nets[j];
								if (nnum >= 0) {					
									src_bnum = net[nnum].pins[0];
									if (Visited[src_bnum] == FALSE && block[src_bnum].type == LUT) {
										Visited[src_bnum] = TRUE;
										marked_blocks[num_marked_blocks++] = src_bnum;
										linked_ptr = (struct s_linked_elem *) my_malloc (sizeof(struct s_linked_elem));
										linked_ptr->bnum = src_bnum;
										linked_ptr->arrival_time = local_NDPoint->arrival_time - 1;
										linked_ptr->next = NULL;
										linked_tail->next = linked_ptr;
										linked_tail = linked_ptr;

									}
								}
							}
						}

						// Undo the Visited array;
						for (i=0; i<num_marked_blocks; i++) {
							Visited[marked_blocks[i]] = FALSE;				
						}
						//***************************************************************
					}
				} else if (root_block_arrival[bnum] >= 0 && root_block_arrival[bnum] <= atime_target) {
//					printf("\tBlock is already rooted with an arrival time that is less than or equal to the target at primary root %d\n",block_in_cluster_w_arrival[bnum][atime_target]);
					// Do nothing
					

				} else if (root_block_arrival[bnum] < 0 && block_in_cluster_w_arrival[bnum][atime_target] >= 0) {	
//					printf("\tBlock is unrooted but exists in another cluster rooted at %d\n",block_in_cluster_w_arrival[bnum][atime_target]);
					// Find the NDPoint where the block already exists and set the root variable
					rooted_bnum = block_in_cluster_w_arrival[bnum][atime_target];
					local_NDPoint2 = block[rooted_bnum].NDSet;
					while (local_NDPoint2->arrival_time != atime_target) {
						if (local_NDPoint2 == NULL) {
							fprintf(stderr,"ERROR in alloc_and_load_cluster_info_depth_optimal. local_NDPoint is NULL\n");
							exit(1);
						}
						local_NDPoint2 = local_NDPoint2->next;
					}
					// find bnum in local_NDPoint->NDList
					for (i=0; i<local_NDPoint2->NDList_size; i++) {
						if (local_NDPoint2->NDList[i] == bnum) {
							local_NDPoint2->roots[i] = TRUE;
						}
					}
					root_block_arrival[bnum] = atime_target;
					//block_in_cluster_w_arrival[bnum][atime_target] = rooted_bnum;


				} else if (root_block_arrival[bnum] < 0 && block_in_cluster_w_arrival[bnum][atime_target] < 0) {			
//					printf("\tBlock is unrooted and is not in another cluster: \n");
					/// **** !!!!!!!!!
					// Add all of the cluster elements to the block_in_cluster_w_arrival array 
					/// **** !!!!!!!!!
					rooted_bnum = bnum;
					for (i=0; i<local_NDPoint->NDList_size; i++) {
						iblk = local_NDPoint->NDList[i];
						if (block_in_cluster_w_arrival[iblk][atime_target] == -1) {
							block_in_cluster_w_arrival[iblk][atime_target] = rooted_bnum;
							//printf("%d %d = %d, ", iblk, atime_target, rooted_bnum);
						}
					}
					// Set block number bnum as a root
					block_in_cluster_w_arrival[bnum][atime_target] = rooted_bnum;
					root_block_arrival[bnum] = atime_target;
					local_NDPoint->roots[0] = TRUE;

					num_clusters++;
//					printf("primary cluster created at block %d\n",bnum);
					num_marked_blocks = 0;

					// Mark as Visited all the blocks in the cluster
					for (i=0; i<local_NDPoint->NDList_size; i++) {
						bnum = local_NDPoint->NDList[i];
						Visited[bnum] = TRUE;
						marked_blocks[num_marked_blocks++] = bnum;
					}

					// Add all the inputs to the queue
					for (i=0; i<local_NDPoint->NDList_size; i++) {
						bnum = local_NDPoint->NDList[i];
						for (j=1; j<block[bnum].num_nets; j++) {
							
							nnum = block[bnum].nets[j];
							if (nnum >= 0) {					
								src_bnum = net[nnum].pins[0];
								if (Visited[src_bnum] == FALSE && block[src_bnum].type == LUT) {
									Visited[src_bnum] = TRUE;
									marked_blocks[num_marked_blocks++] = src_bnum;
									linked_ptr = (struct s_linked_elem *) my_malloc (sizeof(struct s_linked_elem));
									linked_ptr->bnum = src_bnum;
									linked_ptr->arrival_time = local_NDPoint->arrival_time - 1;
									linked_ptr->next = NULL;
									linked_tail->next = linked_ptr;
									linked_tail = linked_ptr;

								}
							}
						}
					}

					// Undo the Visited array;
					for (i=0; i<num_marked_blocks; i++) {
						Visited[marked_blocks[i]] = FALSE;				
					}

				}
			}

			linked_ptr = linked_head;
			linked_head = linked_head->next;
			free(linked_ptr);
		}

	} while (multiple_root_exists);

	// *****************************************************************************************

	// Begin population of cluster_content arrays

	// Allocate the cluster content arrays
    cluster_contents = (int **) alloc_matrix (0,num_clusters-1,0,cluster_size-1,
                       sizeof(int));
    cluster_occupancy = (int *) my_malloc (num_clusters * sizeof (int));
	cluster_roots = (int **) alloc_matrix (0,num_clusters-1,0,cluster_size-1,
                       sizeof(int));


	num_clusters_block_in = (int *) my_malloc (num_blocks * sizeof (int));

	// Local arrays
	cluster_occ_stats = (int **) alloc_matrix (0,max_arrival_time-1,0,cluster_size-1,
                       sizeof(int));
	cluster_num_inputs = (int *) my_malloc ((inputs_per_cluster+1) * sizeof (int));
	cluster_num_outputs = (int *) my_malloc ((cluster_size+1) * sizeof (int));


	// Zero cluster_occ_stats
	for (i=0; i<max_arrival_time; i++) {
		for (j=0; j<cluster_size; j++) {
			cluster_occ_stats[i][j] = 0;
		}
	}

	// Zero cluster_num_inputs
	for (i=0; i<inputs_per_cluster+1; i++) {
		cluster_num_inputs[i] = 0;
	}

	// Zero cluster_num_outputs
	for (i=0; i<cluster_size+1; i++) {
		cluster_num_outputs[i] = 0;
	}

	// Initialize the block_in arrays
	for (i=0; i<num_blocks; i++) {
		num_clusters_block_in[i] = 0;
		for (j=0; j<max_arrival_time; j++) {
			block_in_cluster_w_arrival[i][j] = -1;
		}
	}

	// Initialize cluster_roots array
	for (i=0; i<num_clusters; i++) {
		for (j=0; j<cluster_size; j++) {
			cluster_roots[i][j] = -1;
		}
	}

	// Populate cluster_contents array
	cluster_cnt = 0;
	for (bnum=0; bnum<num_blocks; bnum++) {

		if (block[bnum].type == LUT || block[bnum].type == LATCH || block[bnum].type == LUT_AND_LATCH) {
			local_NDPoint = block[bnum].NDSet;
			if (local_NDPoint == NULL) {
				fprintf(stderr,"Error: local_NDPoint is NULL.  Aborting.\n");
				exit (1);
			}
			root_cnt = 0;

			while (local_NDPoint != NULL) {
				if (local_NDPoint->roots[0] == TRUE) {
					// Add to cluster contents here
					root_cnt++;
					cluster_occupancy[cluster_cnt] = local_NDPoint->NDList_size;
					for (i=0; i<local_NDPoint->NDList_size; i++) {
						cluster_contents[cluster_cnt][i] = local_NDPoint->NDList[i];
						if (local_NDPoint->roots[i] == TRUE) {
							cluster_roots[cluster_cnt][i] = local_NDPoint->arrival_time;
						}
						//block_in_cluster_w_arrival[local_NDPoint->NDList[i]][local_NDPoint->arrival_time] = cluster_cnt;
						num_clusters_block_in[local_NDPoint->NDList[i]]++;
					}
					cluster_cnt++;
					if (cluster_occupancy[cluster_cnt] == 1) {
						single_node_clusters++;
					}
					cluster_occ_stats[local_NDPoint->arrival_time-1][local_NDPoint->NDList_size-1]++;;
				}
				local_NDPoint = local_NDPoint->next;
			}
			if (root_cnt > 1) {
				nnum = block[bnum].nets[0];
				printf("ERROR: net %d has a multiple root\n",nnum);
				exit(1);
			}

		}

	}
	printf("cluster_contents has been populated\n");

	if (cluster_cnt != num_clusters) {
		fprintf(stderr,"Error: cluster_contents population error.  Aborting. cluster_cnt = %d, num_clusters = %d\n", cluster_cnt, num_clusters);
		exit (1);
	}

//	printf("clusters pre-rooting\n");
//	for (icluster=0; icluster<num_clusters; icluster++) {
//		printf("cluster %d: ",icluster);
//		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
//			printf("%d ",cluster_contents[icluster][iblk]);
//		}
//		printf("\n");
//	}

	// Perform roote cluster elimination
	rooted_cluster_elimination(cluster_contents, cluster_occupancy, cluster_roots, &num_clusters, max_arrival_time, lut_size);
	
	// ************************************************************************************
	// Let's do stats collecting
	// ************************************************************************************	

	// Calculate how many inputs each cluster is using and how many outputs each cluster is using.

	// Calculate number of outputs used in each cluster.
	for (icluster=0; icluster<num_clusters; icluster++) {
		num_outputs = 0;		
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			if (cluster_roots[icluster][iblk] >=0) {
				num_outputs++;
			}
		}
		if (num_outputs > cluster_size) {
			printf("Cluster %d has too many outputs %d even though that's impossible\n",icluster,num_outputs);
			exit(1);
		}
		cluster_num_outputs[num_outputs]++;
	}

	// Calculate number of inputs used in each cluster
//	printf("calculating number of inputs used per cluster after rooted cluster elimination\n");
	num_marked_nets = 0;
	for (icluster=0; icluster<num_clusters; icluster++) {
		// mark all the output nets first

		for (i=0; i<num_nets; i++) {
			net_visited[i] = FALSE;
		}

		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			bnum = cluster_contents[icluster][iblk];
			nnum = block[bnum].nets[0];
			net_visited[nnum] = TRUE;
			marked_nets[num_marked_nets++] = nnum;
		}
	
		// check all input nets now
		num_inputs = 0;
		for (iblk=0; iblk<cluster_occupancy[icluster]; iblk++) {
			bnum = cluster_contents[icluster][iblk];
			for (inet=1; inet<=lut_size; inet++) {
				nnum = block[bnum].nets[inet];
				if (nnum != -1) {
					if (net_visited[nnum] == FALSE) {
						net_visited[nnum] = TRUE;
						marked_nets[num_marked_nets++] = nnum;
						num_inputs++;
					}
				}
			}
		}
		if (num_inputs > inputs_per_cluster) {
//			printf("Cluster %d has too many inputs - %d\n",icluster,num_inputs);			
			exit(1);
		}
		cluster_num_inputs[num_inputs]++;
		for (i=0; i<num_marked_nets; i++) {
			nnum = marked_nets[i];
			net_visited[nnum] = FALSE;
		}
		num_marked_nets =0;
	}


	printf("cluster_num_outputs: ");
	for (i=0; i<cluster_size+1; i++) {
		printf("%d ",cluster_num_outputs[i]);
	}
	printf("\n");
	printf("cluster_num_inputs: ");
	for (i=0; i<inputs_per_cluster+1; i++) {
		printf("%d ",cluster_num_inputs[i]);
	}
	printf("\n");



	//printf("finished!\n");

	// output the cluster contents
//	single_node_clusters = 0;
//	node_cnt = 0;
//	for (i=0; i<num_clusters; i++) {
//		printf("Cluster %d, %d: ",i,cluster_occupancy[i]);
//		node_cnt += cluster_occupancy[i];
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%3d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//			if (cluster_occupancy[i] == 1) {
//				single_node_clusters++;
//			}
//		}
//		printf("\n");
//	}

//	printf("There are %d single node clusters\n",single_node_clusters);
//	printf("There are %d total used LUTs\n",node_cnt);
//	printf("There are %d total clusters\n",num_clusters);
//	printf("Average LUT usage per cluster is %f \n", (float)node_cnt/(float)num_clusters);
//	printf("LUT Utilization is %f \n",( (float)node_cnt/(float)num_clusters ) / cluster_size);
	printf("Max circuit depth is %d\n",max_arrival_time);

	// Output the cluster_occ_stats
	printf("Cluster Occupancy vs. arrival_time\n");
	for (i=cluster_size-1; i>=0; i--) {
		printf("cluster_size %d: ",i+1);
		for (j=0; j<max_arrival_time; j++) {
			printf("%d ",cluster_occ_stats[j][i]);
		}
		printf("\n");
	}

	// ************************************************************************************
	// Stats collecting finished
	// ************************************************************************************	

	*num_clusters_ptr = num_clusters;
	*cluster_roots_ptr = cluster_roots;
    *cluster_occupancy_ptr = cluster_occupancy;
    *cluster_contents_ptr = cluster_contents;
	*block_in_cluster_w_arrival_ptr = block_in_cluster_w_arrival;  /* [0..num_blocks-1][0..max_arrival_time-1] */
	*num_clusters_block_in_ptr = num_clusters_block_in; /* [0..num_blocks-1] */
	*max_arrival_time_ptr = max_arrival_time;

//	printf("finished stats collecting\n");
//	fflush(stdout);

	free(Visited);
	free(multiple_root_arrival);
	free(root_block_arrival);
	free(cluster_occ_stats);
	free(cluster_num_inputs);
	free(cluster_num_outputs);
}

// **********************************************************************************
// depth_optimal_cluster
//
//
//
// **********************************************************************************
void depth_optimal_cluster (int cluster_size, int inputs_per_cluster, int lut_size)
{

	int iblk,i,j, inet, sub_blk;
	int* SearchQ;
	boolean* Visited;
	int SQHead, SQTail;

	int OutputBlockNum, CurrNet, InputBlockNum, SrcNet, SrcBlock;
	boolean AddBlock;

	printf("Performing depth optimal clustering. Please wait...\n");
	alloc_and_init_clustering (lut_size, cluster_size);

	// Create an empty point
	alloc_and_init_NDPoint (&empty_NDPoint, cluster_size, -1);

	// First thing to do is to traverse the tree in breadth first manner

	// Create the search Q
    SearchQ = (int*) my_malloc (num_blocks * sizeof (int));
	Visited = (boolean*) my_malloc (num_blocks * sizeof (boolean));
	SQHead = 0;
	SQTail = 0;

	// Initialize the SearchQ 
	for (iblk=0; iblk<num_blocks; iblk++) {
		SearchQ[iblk] = -3;
		Visited[iblk] = FALSE;
	}

	// Insert all of the primary inputs and flip flop outputs into the SearchQ
	for (iblk=0; iblk<num_blocks; iblk++) {

		// Find all ouptuts of INPAD, LATCH and LUT_AND_LATCH
		if (block[iblk].type == INPAD || block[iblk].type == LATCH || block[iblk].type == LUT_AND_LATCH) {
			// Output net			
			inet = block[iblk].nets[0];			
			// Check all pins on the output net
			for (i=1; i<net[inet].num_pins; i++) {

				InputBlockNum = net[inet].pins[i];

				AddBlock = TRUE;
				// Check all the blocks that input to InputBlockNum has been visited.
				for (j=1; j<block[InputBlockNum].num_nets; j++) {
					SrcNet = block[InputBlockNum].nets[j];
					if (SrcNet >= 0) {							
						SrcBlock = net[SrcNet].pins[0];
							
						if (Visited[SrcBlock] == FALSE && block[SrcBlock].type == LUT ) {
							AddBlock = FALSE;
						}
					}
					
				}
				
				if (AddBlock && Visited[InputBlockNum] == FALSE) {
					SearchQ[SQTail] = InputBlockNum;
					Visited[InputBlockNum] = TRUE;
					SQTail++;
				}

			}
		} else {
			// Check to see if the block has no inputs.  If so, add it to the queue.
			AddBlock = TRUE;
			for (inet=1; inet<block[iblk].num_nets; inet++) {
				if (inet >= 0) {
					AddBlock = FALSE;
				}
			}
			if (AddBlock && Visited[iblk] == FALSE) {
				SearchQ[SQTail] = iblk;
				Visited[iblk] = TRUE;
				SQTail++;
			}
		}
	}

	// Visit each node in the circuit in a breadth first manner
	while (SQHead != SQTail)
	{
		if (SQTail > num_blocks) {
			printf("STOP!!!! The search queue has overflowed. SQHead = %d, SQTail = %d, num_blocks =%d\n",SQHead,SQTail,num_blocks);
			exit(1);
		}

		OutputBlockNum = SearchQ[SQHead];
		CurrNet = block[OutputBlockNum].nets[0];
		//printf("\nI am at block %d which is of type %d\n",OutputBlockNum, block[OutputBlockNum].type);

		// Enumerate each clustering solution at this node.
//		create_NDSet(OutputBlockNum, cluster_size/2, inputs_per_cluster, lut_size);
		create_NDSet(OutputBlockNum, cluster_size, inputs_per_cluster, lut_size);

		// This block fans out to a series of other blocks.  Check each of the fanout blocks to see if all its inputs have been visited.
		// If so, add it to the Search Q.		
		if (block[OutputBlockNum].type == LUT) {
			for (i=1; i<net[CurrNet].num_pins; i++) {			
				InputBlockNum = net[CurrNet].pins[i];
				
				// check if all inputs to the net have been visited.
				AddBlock = TRUE;			
				for (j=1; j<block[InputBlockNum].num_nets; j++) {
					SrcNet = block[InputBlockNum].nets[j];
					if (SrcNet >= 0) {					
						SrcBlock = net[SrcNet].pins[0];
						
						if (Visited[SrcBlock] == FALSE && block[SrcBlock].type == LUT) {
							AddBlock = FALSE;					
						}
					}
				}
				if (AddBlock && Visited[InputBlockNum] == FALSE) {
					SearchQ[SQTail] = InputBlockNum;
					Visited[InputBlockNum] = TRUE;
					SQTail++;
				}
			}
		}
		// Increment SQHead
		SQHead ++;
	}

	// Check to see that every block has been visited
//	printf("\nChecking to see that every block has been visited\n");
//	for (i=0; i<num_blocks; i++) {
//		if (Visited[i] == FALSE) {
//			printf("\tWTF!!! Block %i, type %d was never visited\n",i,block[i].type);
//		}
//	}
//	printf("Check complete!\n");

	printf("\nFinished depth optimal clustering!!!\n");

	free(SearchQ);
	free(Visited);

}

// **********************************************************************************
// get_cluster_seed
//
//
//
// **********************************************************************************
int get_cluster_seed(int **cluster_contents, int *cluster_occupancy, struct s_ilink *unclustered_list_head, int cluster_size)
{

	struct s_ilink *list_ptr;
	struct s_ilink *prev_list_ptr;
	struct s_ilink *largest_ptr;
	struct s_ilink *prev_largest_ptr;
	//int max_occupancy;
	int min_occupancy;
	int largest_cluster;
	int icluster;

	list_ptr = NULL;
	prev_list_ptr = NULL;
	largest_ptr = NULL;
	prev_largest_ptr = NULL;

	if (unclustered_list_head == NULL) {
		return NO_CLUSTER;
	}

	// remember to take the seed out of the unclustered_list;

	// For now, just take the first cluster on the list off and make it the root

	prev_list_ptr = unclustered_list_head;
	list_ptr = unclustered_list_head->next;


	if (list_ptr == NULL) {
		return(NO_CLUSTER);
	}
//
//	unclustered_list_head->next = list_ptr->next;
//	return(list_ptr->iblk);

	// Let's find the largest sized cluster.
	//max_occupancy = 0;
	min_occupancy = cluster_size;
	while (list_ptr != NULL) {
		icluster = list_ptr ->iblk;
		//if (cluster_occupancy[icluster] > max_occupancy) {
		if (cluster_occupancy[icluster] < min_occupancy) {
			largest_cluster = list_ptr->iblk;
			//max_occupancy = cluster_occupancy[icluster];
			min_occupancy = cluster_occupancy[icluster];
			largest_ptr = list_ptr;
			prev_largest_ptr = prev_list_ptr;
		}
		
		prev_list_ptr = list_ptr;
		list_ptr = list_ptr->next;
	}

	if (largest_ptr == NULL) {
		fprintf(stderr, "ERROR in get cluster_seed.  List is not empty but could not find a cluster\n");
		exit(1);
	}

	// Remove the largest sized element
	prev_largest_ptr->next = largest_ptr->next;

	return(largest_ptr->iblk);

}

// **********************************************************************************
// merge_clusters
//
//
//
// **********************************************************************************
void merge_clusters(int dest, int src, int **cluster_contents, int *cluster_occupancy, int **cluster_roots, int cluster_size, int *num_clusters_block_in, int lut_size)
{
	
	int i, j, iblk;
	int bnum, nnum;
	//int temp_cnt;
	boolean absorb_net;

	int *num_pins_in_cluster;
	boolean *block_in_cluster;

	num_pins_in_cluster = (int *) my_malloc (num_nets * sizeof(int));
	block_in_cluster = (boolean *) my_malloc (num_blocks * sizeof(boolean));	

	for (i=0; i<num_blocks; i++) {
		block_in_cluster[i] = FALSE;
	}
	for (i=0; i<num_nets; i++) {
		num_pins_in_cluster[i] = 0;
	}

	// Go through each block already in the cluster					
	for (iblk=0; iblk<cluster_occupancy[dest]; iblk++) {

		// Add the block to the block_in_cluster
		bnum = cluster_contents[dest][iblk];
		block_in_cluster[bnum] = TRUE;
		marked_blocks[num_marked_blocks++] = bnum;
		
	}

	//temp_cnt = 0;
	// 
	for (i=0; i<cluster_occupancy[src]; i++) {
		bnum = cluster_contents[src][i];
		if (block_in_cluster[bnum] == FALSE) {
			block_in_cluster[bnum] = TRUE;
			marked_blocks[num_marked_blocks++] = bnum;
			cluster_contents[dest][cluster_occupancy[dest]] = bnum;
			cluster_occupancy[dest]++;
			//temp_cnt++;
		} else {
			num_clusters_block_in[bnum]--;
		}
	}

	// cluster_contents[isearch][0];  find it;

	// **********************************************************
	// ***** Gotta change how I handle the roots here
	// **********************************************************

	// Copy the roots to the new cluster
	for (i=0; i<cluster_occupancy[src]; i++) {
		if (cluster_roots[src][i] >= 0) {		
			for (j=0; j<cluster_occupancy[dest]; j++) {
				if (cluster_contents[dest][j] == cluster_contents[src][i]) {
					if (cluster_roots[src][i] > cluster_roots[dest][j]) {
						cluster_roots[dest][j] = cluster_roots[src][i];
					}					
				}
			}
		}
	}

	cluster_occupancy[src] = 0;
	cluster_roots[src][0] = -1;

	// undo marked_blocks and marked_nets
	for (i=0; i<num_marked_blocks; i++) {
		block_visited[marked_blocks[i]] = FALSE;
	}
	for (i=0; i<num_marked_nets; i++) {
		net_visited[marked_nets[i]] = FALSE;			
	}
	num_marked_blocks = 0;
	num_marked_nets = 0;

	// let's mark all the output nets of new_cluster
	for (i=0; i<cluster_occupancy[dest]; i++) {
		bnum = cluster_contents[dest][i];
		nnum = block[bnum].nets[0];
		if (block_visited[bnum] == FALSE) {
			marked_blocks[num_marked_blocks++] = bnum;
			block_visited[bnum] = TRUE;
		}
		if (net_visited[nnum] == FALSE) {			
			marked_nets[num_marked_nets++] = nnum;
			net_visited[nnum] = TRUE;			
		}
		num_pins_in_cluster[nnum]++;
	}

	// lets' mark all the input nets of new cluster
	for (i=0; i<cluster_occupancy[dest]; i++) {
		bnum = cluster_contents[dest][i];
		for (j=1; j<=lut_size; j++) {
			nnum = block[bnum].nets[j];
			if (nnum != OPEN) {
				if (net_visited[nnum] == FALSE) {
					marked_nets[num_marked_nets++] = nnum;
					net_visited[nnum] = TRUE;
				}
				num_pins_in_cluster[nnum]++;
			}
		}
	}

	// Check all the roots now.
	for (i=0; i<cluster_occupancy[dest]; i++) {
		//printf("\t\t\tblock %d:",cluster_contents[icluster][i]);
		if (cluster_roots[dest][i] >= 0) {
			//printf("root,  ");
			bnum = cluster_contents[dest][i];
			nnum = block[bnum].nets[0];
			//printf("nnum = %d, num_pins = %d, net_num_pins = %d, ",nnum,num_pins_in_cluster[nnum],net[nnum].num_pins);
			if (nnum != OPEN) {			
				if (num_pins_in_cluster[nnum] == net[nnum].num_pins) {
					//printf("block: ");
					absorb_net = TRUE;
					for (j=1; j<net[nnum].num_pins; j++) {
						bnum = net[nnum].pins[j];

						if (num_clusters_block_in[bnum] > 1) {
							absorb_net = FALSE;
						}
					}
					if (absorb_net) {
						cluster_roots[dest][i] = -1;
					}
				}
			}
		}
		//printf("\n");
	}; // end for

	// undo marked_blocks and marked_nets
	for (i=0; i<num_marked_blocks; i++) {
		block_visited[marked_blocks[i]] = FALSE;
	}
	for (i=0; i<num_marked_nets; i++) {
		net_visited[marked_nets[i]] = FALSE;			
	}
	num_marked_blocks = 0;
	num_marked_nets = 0;
	

}

// **********************************************************************************
// read_place_header
//
//
//
// **********************************************************************************
static void read_place_header (FILE *fp, char *net_file, char *arch_file, 
              char *buf) {

/* Reads the header from the placement file.  Used only to check that this *
 * placement file matches the current architecture, netlist, etc.          */

 char *line_one_names[] = {"Netlist", "file:", " ", "Architecture", "file:"};
 char *line_two_names[] = {"Array", "size:", " ", "x", " ", "logic", "blocks"};
 char net_check[BUFSIZE], arch_check[BUFSIZE], *ptr;
 int nx_check, ny_check, i;


 ptr = my_fgets (buf, BUFSIZE, fp);
 
 if (ptr == NULL) {
    printf ("Error:  netlist file and architecture file used not listed.\n");
    exit (1);
 }
 
 ptr = my_strtok (buf, TOKENS, fp, buf);
 while (ptr == NULL) {   /* Skip blank or comment lines. */
    ptr = my_fgets (buf, BUFSIZE, fp);
    if (ptr == NULL) {
       printf ("Error:  netlist file and architecture file used not listed.\n");       exit (1);
    }
    ptr = my_strtok (buf, TOKENS, fp, buf);
 }
 
 for (i=0;i<=5;i++) {
    if (i == 2) {
       strcpy (net_check, ptr);
    }
    else if (i == 5) {
       strcpy (arch_check, ptr);
    }
    else {
       if (strcmp (ptr, line_one_names[i]) != 0) {
          printf ("Error on line %d, word %d:  \n"
                  "Expected keyword %s, got %s.\n", linenum, i+1, 
                  line_one_names[i], ptr);
          exit (1);
       } 
    }
    ptr = my_strtok (NULL, TOKENS, fp, buf);
    if (ptr == NULL && i != 5) {
       printf ("Error:  Unexpected end of line on line %d.\n", linenum);
       exit (1);
    }
 }
 
 if (strcmp(net_check, net_file) != 0) {
    printf("Warning:  Placement generated with netlist file %s:\n", 
            net_check);
    printf("current net file is %s.\n", net_file);
 }
 
 if (strcmp(arch_check, arch_file) != 0) {
    printf("Warning:  Placement generated with architecture file %s:\n", 
           arch_check);
    printf("current architecture file is %s.\n", arch_file);
 }

/* Now check the second line (array size). */

 ptr = my_fgets (buf, BUFSIZE, fp);

 if (ptr == NULL) {
    printf ("Error:  Array size not listed.\n");
    exit (1);
 }
 
 ptr = my_strtok (buf, TOKENS, fp, buf);
 while (ptr == NULL) {   /* Skip blank or comment lines. */
    ptr = my_fgets (buf, BUFSIZE, fp);
    if (ptr == NULL) {
       printf ("Error:  array size not listed.\n");
       exit (1);
    }
    ptr = my_strtok (buf, TOKENS, fp, buf);
 }

 for (i=0;i<=6;i++) {
    if (i == 2) {
       sscanf (ptr, "%d", &nx_check);
    }
    else if (i == 4) {
       sscanf (ptr, "%d", &ny_check);
    }
    else {
       if (strcmp (ptr, line_two_names[i]) != 0) {
          printf ("Error on line %d, word %d:  \n"
                  "Expected keyword %s, got %s.\n", linenum, i+1, 
                  line_two_names[i], ptr);
          exit (1);
       } 
    }
    ptr = my_strtok (NULL, TOKENS, fp, buf);
    if (ptr == NULL && i != 6) {
       printf ("Error:  Unexpected end of line on line %d.\n", linenum);
       exit (1);
    }
 }

 nx = nx_check;
 ny = ny_check;
 printf("nx = %d, ny = %d\n",nx,ny);

// if (nx_check != nx || ny_check != ny) {
//    printf ("Error:  placement file assumes an array size of %d x %d.\n",
//             nx_check, ny_check);
//    printf ("Current size is %d x %d.\n", nx, ny);
//    exit (1);
// }
}

// **********************************************************************************
// generate_and_read_placement
//
//
//
// **********************************************************************************
void generate_and_read_placement(char *out_fname, int **cluster_contents, struct s_coord **cluster_coords_ptr, int num_clusters)
{

	// VPR executable files
	char COMMAND[1000] = "";
	char  PLACE_PATH[1000] = "";
	char  ROUTE_PATH[1000] = "";
	char* VPR = "~marvint/research/src_code/vpr_orig/vpr/vpr_linux";
	char* ARCH = "~marvint/research/work/arch_ieee/k=4/N=8/arch";	
	char* OPTIONS = "-nodisp -place_only";
	char* LOGFILE = "";

	// file input variables
	FILE *fp;
	int linenum;
	char buf[BUFSIZE], bname[BUFSIZE], *ptr;
	int xtmp, ytmp;
	struct s_hash **hash_table, *h_ptr;
	int i, icluster;

	// cluster coordinates array
	struct s_coord *cluster_coords;
	cluster_coords = (struct s_coord *) my_malloc (num_clusters * sizeof(struct s_coord));
	for (i=0; i<num_clusters; i++) {
		cluster_coords[i].x = -1;
		cluster_coords[i].y = -1;
	}


	// Generate placment files
	strcpy(PLACE_PATH, "placedfiles/temp/");
	//strcat(PLACE_PATH, out_fname);
	strcat(PLACE_PATH, "temp.placed");
	//strcat(PLACE_PATH, ".placed");
	strcpy(ROUTE_PATH, "routedfiles/temp/");
	//strcat(ROUTE_PATH, out_fname);
	strcat(ROUTE_PATH, "temp.routed");
	//strcat(ROUTE_PATH, ".routed");

	// Generate vpr executable
	strcpy(COMMAND, VPR);strcat(COMMAND, " ");
	strcat(COMMAND, out_fname);strcat(COMMAND, " ");
	strcat(COMMAND, ARCH);strcat(COMMAND, " ");
	strcat(COMMAND, PLACE_PATH);strcat(COMMAND, " ");
	strcat(COMMAND, ROUTE_PATH);strcat(COMMAND, " ");
	strcat(COMMAND, OPTIONS);strcat(COMMAND, " ");
	strcat(COMMAND, LOGFILE);

	// Execute VPR executable
	printf("Generating placement information. Please wait...\n");
	//printf("%s",COMMAND);
	system(COMMAND);
	printf("Placement created!\n");	

	// Read in placement information
	printf("Reading the placement from file %s.\n", PLACE_PATH);
	fp = my_fopen (PLACE_PATH, "r", 0);
	linenum = 0;

	// Read the placement file header
	read_place_header (fp, out_fname, ARCH, buf);

	hash_table = alloc_hash_table ();
	for (i=0;i<num_clusters;i++) {

		h_ptr = insert_in_hash_table (hash_table, block[cluster_contents[i][0]].name, i);
	}

	// Read each subsequent line in the placement file
	ptr = my_fgets (buf, BUFSIZE, fp);
	while (ptr != NULL) {
		ptr = my_strtok (buf, TOKENS, fp, buf);
		if (ptr == NULL) {
		   ptr = my_fgets (buf, BUFSIZE, fp);
		   continue;      /* Skip blank or comment lines. */
		}

		strcpy (bname, ptr); 

		// Read x coord
		ptr = my_strtok (NULL, TOKENS, fp, buf);
		if (ptr == NULL) {
		   printf ("Error:  line %d is incomplete.\n", linenum);
		   exit (1);
		}
		sscanf (ptr, "%d", &xtmp);
	   
		// Read y coord
		ptr = my_strtok (NULL, TOKENS, fp, buf);
		if (ptr == NULL) {
		   printf ("Error:  line %d is incomplete.\n", linenum);
		   exit (1);
		}
		sscanf (ptr, "%d", &ytmp);
		
		// Read subblk
		ptr = my_strtok (NULL, TOKENS, fp, buf);
		if (ptr == NULL) {
		   printf ("Error:  line %d is incomplete.\n", linenum);
		   exit (1);
		}
		//sscanf (ptr, "%d", &isubblock);

		// Read block number
		ptr = my_strtok (NULL, TOKENS, fp, buf);
		if (ptr != NULL) {
		   printf ("Error:  extra characters at end of line %d.\n", linenum);
		   exit (1);
		}
		
		// check if block is an input or an output
		if (xtmp > 0 && xtmp < nx+1 && ytmp > 0 && ytmp < ny+1) {		
			h_ptr = get_hash_entry (hash_table, bname);
			if (h_ptr == NULL) {
			   printf ("Error:  block %s on line %d does not exist in the netlist.\n",
						bname, linenum);
			   exit (1);
			}
			icluster = h_ptr->index;

			// cluster icluster has coordinates xtmp and ytmp
			cluster_coords[icluster].x = xtmp;
			cluster_coords[icluster].y = ytmp;
			//printf("cluster %d: x = %d, y= %d\n",icluster, cluster_coords[icluster].x,cluster_coords[icluster].y);
		}

		ptr = my_fgets (buf, BUFSIZE, fp);

	}

	// Check that every cluster has a coordinate
	for (i=0; i<num_clusters; i++) {
		if (cluster_coords[i].x == -1 || cluster_coords[i].y == -1) {
			fprintf(stderr,"Incomplete placement file. CLB %d does not have any coordinates\n",i);
			exit(1);
		}
	}

	*cluster_coords_ptr = cluster_coords;

}

// **********************************************************************************
// cluster_consolidate
//
//
//
// **********************************************************************************
void cluster_consolidate(int **cluster_contents, int *cluster_occupancy, int **cluster_roots, struct s_coord *cluster_coords,
							int cluster_size, int *num_clusters_ptr, int lut_size, int inputs_per_cluster, 
							int **block_in_cluster_w_arrival, int *num_clusters_block_in, int max_arrival_time)

{

	int num_clusters;

	// temporary variables
	int i,j;
	int bnum,nnum,iblk;
	int istart, icluster, isearch;
	int node_cnt;

	// Unclustered list data structure
	struct s_ilink *unclustered_list_head;
	struct s_ilink *memory_pool;
    struct s_ilink *list_ptr;
	struct s_ilink *prev_list_ptr;
	int *new_cluster;
	int *new_cluster_roots;

	int new_cluster_size;
	int inputs_used;	
	int remaining_spaces;
	int remaining_inputs;
	int potential_cluster_size;

	// temporary variables
	int highest_gain_cluster;
	float highest_gain_value;

	// duplicate data stuctures
	int *marked_nets2, *marked_blocks2;
	int num_marked_nets2, num_marked_blocks2;
	boolean *net_visited2, *block_visited2;
	int inputs_used2;

	float attraction_factor;
	int manhattan_distance;
	int common_nets;

	printf("Performing cluster consolidation\n");

	// Create duplicate data structures
    marked_nets2 = (int *) my_malloc ((lut_size + 2) * cluster_size * sizeof(int));
    marked_blocks2 = (int *) my_malloc (num_blocks * sizeof(int));

    num_marked_nets2 = 0;
    num_marked_blocks2 = 0;

	net_visited2 = (boolean *) my_malloc (num_nets * sizeof(boolean));
	block_visited2 = (boolean *) my_malloc (num_blocks * sizeof(boolean));

	for (i=0; i<num_nets; i++) {
		net_visited2[i] = FALSE;
	}
	for (i=0; i<num_blocks; i++) {
		block_visited2[i] = FALSE;
	}

	// Create local array to hold cluster info
	//new_cluster = (int *) my_malloc (cluster_size * sizeof (struct s_ilink));
	//new_cluster_size = 0;

//	printf("\n\nStarting Cluster Consolidation\n\n");

	num_clusters = *num_clusters_ptr;

	// Clusters pre-consolidate
//	for (i=0; i<num_clusters; i++) {
//		printf("pre - Cluster %d, %d: ",i,cluster_occupancy[i]);
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%3d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//		}
//
//		printf("\n");
//	}

	// Create the lists of clusters that can be compacted indexed by the number of inputs that it uses.
    unclustered_list_head = (struct s_ilink *) my_malloc (sizeof(struct s_ilink));
    unclustered_list_head->next = NULL;

    memory_pool = (struct s_ilink *) my_malloc (num_clusters * sizeof (struct s_ilink));
    list_ptr = memory_pool;

    for (i=0;i<num_clusters;i++) {
        if (cluster_occupancy[i] < cluster_size) {
            list_ptr->next = unclustered_list_head->next;
            unclustered_list_head->next = list_ptr;
            list_ptr->iblk = i;
            list_ptr++;
        }
    }

	// undo marked_nets and marked_blocks
	for (i=0; i<num_marked_blocks; i++) {
		block_visited[marked_blocks[i]] = FALSE;
	}
	for (i=0; i<num_marked_nets; i++) {
		net_visited[marked_nets[i]] = FALSE;			
	}
	num_marked_blocks = 0;
	num_marked_nets = 0;

//	printf("unclustered list has been created: ");
//
//	list_ptr = unclustered_list_head->next;
//	while (list_ptr != NULL) {
//		printf("%d ",list_ptr->iblk);
//		list_ptr = list_ptr->next;
//	}
//	printf("\n");

	// Unclustered list has been created.
	istart = get_cluster_seed(cluster_contents, cluster_occupancy, unclustered_list_head, cluster_size);

//	printf("Got cluster seed %d\n",istart);
	while (istart != NO_CLUSTER) {

		// Let's reset the cluster
		new_cluster = cluster_contents[istart];
		new_cluster_size = cluster_occupancy[istart];	
	
//		// Add the current cluster to the new_cluster
//		for (i=0; i<cluster_occupancy[istart]; i++) {
//			new_cluster[new_cluster_size] = cluster_contents[istart][i];
//			new_cluster_roots[new_cluster_size] = cluster_roots[istart][i];
//			new_cluster_size++;
//		}

		// let's mark all the output nets of new_cluster
		for (i=0; i<new_cluster_size; i++) {
			bnum = new_cluster[i];
			nnum = block[bnum].nets[0];
			if (block_visited[bnum] == FALSE) {
				marked_blocks[num_marked_blocks++] = bnum;
				block_visited[bnum] = TRUE;
			}
			if (net_visited[nnum] == FALSE) {			
				marked_nets[num_marked_nets++] = nnum;
				net_visited[nnum] = TRUE;
			}
		}

		// lets' mark all the input nets of new cluster
		inputs_used = 0;
		for (i=0; i<new_cluster_size; i++) {
			bnum = new_cluster[i];
			for (j=1; j<=lut_size; j++) {
				nnum = block[bnum].nets[j];
				if (nnum != OPEN) {
					if (net_visited[nnum] == FALSE) {
						marked_nets[num_marked_nets++] = nnum;
						net_visited[nnum] = TRUE;
						inputs_used++;
					}
				}
			}
		}
		
		remaining_spaces = cluster_size - new_cluster_size;
		remaining_inputs = inputs_per_cluster - inputs_used;

//		printf("cluster %d has %d remaining spaces and %d remaining_inputs\n",istart, remaining_spaces, remaining_inputs);		

		// Lets calculate the attraction of all the blocks in the list now.
		list_ptr = unclustered_list_head->next;
		highest_gain_cluster = NO_CLUSTER;
//		highest_gain_value = 0;
		highest_gain_value = nx + ny;
		while (list_ptr != NULL) {			
			isearch = list_ptr->iblk;
//			printf("\tChecking cluster %d with %d blocks: ",isearch, cluster_occupancy[isearch]);

			// Check how many spaces the new cluster would take.
			for (i=0; i<new_cluster_size; i++) {
				bnum = new_cluster[i];					
				if (block_visited2[bnum] == FALSE) {
					marked_blocks2[num_marked_blocks2++] = bnum;
					block_visited2[bnum] = TRUE;
				}				
			}
			for (i=0; i<cluster_occupancy[isearch]; i++) {
				bnum = cluster_contents[isearch][i];
				if (block_visited2[bnum] == FALSE) {
					marked_blocks2[num_marked_blocks2++] = bnum;
					block_visited2[bnum] = TRUE;
				}
			}

			potential_cluster_size = num_marked_blocks2;

			// undo marked_blocks2 and marked_nets2 arrays
			for (i=0; i<num_marked_blocks2; i++) {
				block_visited2[marked_blocks2[i]] = FALSE;
			}
			num_marked_blocks2 = 0;


			if (potential_cluster_size <= cluster_size) {
//				printf("\t\tpotential_cluster_size is %d\n",potential_cluster_size);
				// I have to count all the inputs to make sure that input constraint is satisfied.
				///////////////////////////////////////////////////////////////////////

				for (i=0; i<num_nets; i++) {
					net_visited2[i] = FALSE;
				}
				for (i=0; i<num_blocks; i++) {
					block_visited2[i] = FALSE;
				}
				// let's mark all the output nets of new_cluster
//				printf("\n\t\tblocks: ");
				for (i=0; i<new_cluster_size; i++) {
					bnum = new_cluster[i];					
					nnum = block[bnum].nets[0];
					if (block_visited2[bnum] == FALSE) {
						marked_blocks2[num_marked_blocks2++] = bnum;
						block_visited2[bnum] = TRUE;
//						printf("%d ",bnum);
					}
					if (net_visited2[nnum] == FALSE) {			
						marked_nets2[num_marked_nets2++] = nnum;
						net_visited2[nnum] = TRUE;
//						printf("net%d ",nnum);
					}					
				}
//				printf("\n\t\tblocks: ");
				// Let's mark all the output nets of candidate cluster isearch
				for (i=0; i<cluster_occupancy[isearch]; i++) {
					bnum = cluster_contents[isearch][i];
					nnum = block[bnum].nets[0];
					if (block_visited2[bnum] == FALSE) {
						marked_blocks2[num_marked_blocks2++] = bnum;
						block_visited2[bnum] = TRUE;
//						printf("%d ",bnum);
					}
					if (net_visited2[nnum] == FALSE) {			
						marked_nets2[num_marked_nets2++] = nnum;
						net_visited2[nnum] = TRUE;
//						printf("net%d ",nnum);
					}
				}

				// lets' mark all the input nets of new cluster
				inputs_used2 = 0;
//				printf("\n\t\tblocks: ");
				for (i=0; i<new_cluster_size; i++) {
					bnum = new_cluster[i];
//					printf("%d ",bnum);
					for (j=1; j<=lut_size; j++) {
						nnum = block[bnum].nets[j];
//						printf("net%d",nnum);
						if (nnum != OPEN) {
//							printf("a");
							if (net_visited2[nnum] == FALSE) {
								marked_nets2[num_marked_nets2++] = nnum;
								net_visited2[nnum] = TRUE;
								inputs_used2++;								
//								printf("m");
							}
						}
//						printf(" ");
					}
				}
//				printf("new cluster has %d inputs, ",inputs_used2);

//				printf("\n\t\tblocks: ");
				// Let's add all the inputs nets of candidate cluster isearch
				for (i=0; i<cluster_occupancy[isearch]; i++) {
					bnum = cluster_contents[isearch][i];
//					printf("%d ",bnum);
					for (j=1; j<=lut_size; j++) {
						nnum = block[bnum].nets[j];
//						printf("net%d",nnum);
						if (nnum != OPEN) {
							if (net_visited2[nnum] == FALSE) {
								marked_nets2[num_marked_nets2++] = nnum;
								net_visited2[nnum] = TRUE;
								inputs_used2++;
//								printf("m",nnum);
							} 
						}
//						printf(" ");
					}
				}

				// undo marked_blocks2 and marked_nets2 arrays
				for (i=0; i<num_marked_blocks2; i++) {
					block_visited2[marked_blocks2[i]] = FALSE;
				}
				for (i=0; i<num_marked_nets2; i++) {
					net_visited2[marked_nets2[i]] = FALSE;
				}
				num_marked_blocks2 = 0;
				num_marked_nets2 = 0;

				///////////////////////////////////////////////////////////////////////
//				printf("new cluster has %d inputs, ",inputs_used2);
				// If the input constraint is satisfied then I can calculate the attraction factor of the new cluster
				if (inputs_used2 <= inputs_per_cluster) {
//					printf("\t\t\tNumber of inputs = %d\n",inputs_used2);
					// Calculate the attraction of candidate cluster isearch

					// Let's mark all the output nets of candidate cluster isearch
					for (i=0; i<cluster_occupancy[isearch]; i++) {
						bnum = cluster_contents[isearch][i];
						nnum = block[bnum].nets[0];
						if (block_visited2[bnum] == FALSE) {
							marked_blocks2[num_marked_blocks2++] = bnum;
							block_visited2[bnum] = TRUE;
						}
						if (net_visited2[nnum] == FALSE) {			
							marked_nets2[num_marked_nets2++] = nnum;
							net_visited2[nnum] = TRUE;
						}
					}

					inputs_used2 = 0;
					// Let's add all the inputs nets of candidate cluster isearch
					for (i=0; i<cluster_occupancy[isearch]; i++) {
						bnum = cluster_contents[isearch][i];
						for (j=1; j<=lut_size; j++) {
							nnum = block[bnum].nets[j];
							if (nnum != OPEN) {
								if (net_visited2[nnum] == FALSE) {
									marked_nets2[num_marked_nets2++] = nnum;
									net_visited2[nnum] = TRUE;
									inputs_used2++;
								} 
							}
						}
					}

					// Calculate the attraction factor of the candidate cluster isearch
					common_nets = 0;
					for (i=0; i<num_marked_nets; i++) {
						nnum = marked_nets[i];
						if (net_visited2[nnum] == TRUE) {
							common_nets++;
						}
					}

					// Attraction factor based on just manhattan distance first. 
					// Calculate attraction between block istart and isearch
//					manhattan_distance = abs(cluster_coords[istart].x - cluster_coords[isearch].x) +
//										 abs(cluster_coords[istart].y - cluster_coords[isearch].y);
					manhattan_distance = 1;
//					attraction_factor = (float) manhattan_distance - ( (float) common_nets / (float) inputs_per_cluster );
					attraction_factor = common_nets;
//					printf("number of common nets = %d\n",common_nets);
//					printf("attraction = %f\n",attraction_factor);

//					printf("\t\tistart = %d, %d, isearch = %d, %d\n",cluster_coords[istart].x,cluster_coords[istart].y,cluster_coords[isearch].x,cluster_coords[isearch].y);
//					printf("\t\t\tistart = %d, isearch = %d, Attraction factor is %d\n",istart,isearch,attraction_factor);
//					if (attraction_factor > 0) {
//						exit(1);  
//					}
					// Check if current block has higher attraction 
					if (attraction_factor > highest_gain_value && attraction_factor > lut_size+1) {
//					if (attraction_factor < highest_gain_value && attraction_factor < 0.1 * nx) {
						highest_gain_cluster = isearch;
						highest_gain_value = attraction_factor;
					}

					// undo marked_blocks2 and marked_nets2 arrays
					for (i=0; i<num_marked_blocks2; i++) { 
						block_visited2[marked_blocks2[i]] = FALSE;
					}
					for (i=0; i<num_marked_nets2; i++) {
						net_visited2[marked_nets2[i]] = FALSE;
					}
					num_marked_blocks2 = 0;
					num_marked_nets2 = 0;

				}; // end if (inputs_used2 <= inputs_per_cluster) {

			}; // end if (cluster_occupancy[isearch] <= remaining_spaces)

			//printf("\n");

			list_ptr = list_ptr->next;

		}; // end while (list_ptr != NULL)

//		printf("\thighest gain cluster is %d with gain %d\n",highest_gain_cluster, highest_gain_value);		

		// The highest_gain_cluster is now in highest_gain_cluster
		if (highest_gain_cluster != NO_CLUSTER) {
			// copy highest_gain_cluster into new_cluster and store it in istart
			// copy new_cluster into cluster istart;
//			printf("\tMerging cluster %d and cluster %d\n",istart,highest_gain_cluster);
			merge_clusters(istart, highest_gain_cluster, cluster_contents, cluster_occupancy, cluster_roots, cluster_size, num_clusters_block_in, lut_size);

//			printf("\tnew cluster %d: ", istart);
//			for (i=0; i<cluster_occupancy[istart]; i++) {
//				printf("%d ",cluster_contents[istart][i]);
//			}
//			printf("\n");
//			printf("\told cluster %d: ", highest_gain_cluster);
//			for (i=0; i<cluster_occupancy[highest_gain_cluster]; i++) {
//				printf("%d ",cluster_contents[highest_gain_cluster][i]);
//			}
//			printf("\n");


			// Remove highest_gain_cluster from the unclustered_list;
			prev_list_ptr = unclustered_list_head;
			list_ptr = unclustered_list_head->next;

			while (list_ptr->iblk != highest_gain_cluster) {
				prev_list_ptr = list_ptr;
				list_ptr = list_ptr->next;
			}

			if (list_ptr == NULL) {
				fprintf(stderr,"ERROR in unclustered list head.\n");
				exit(1);
			}

			prev_list_ptr->next = list_ptr->next;
			//free(list_ptr);

			if (cluster_occupancy[istart] == cluster_size) {
				istart = get_cluster_seed(cluster_contents, cluster_occupancy, unclustered_list_head, cluster_size);
//				printf("\tCluster is full. Get new seed %d\n",istart);
			} else {
//				printf("\tCluster %d is not full yet\n",istart);
			}

		} else {
			istart = get_cluster_seed(cluster_contents, cluster_occupancy, unclustered_list_head, cluster_size);
//			printf("\tCannot find any more blocks.  Get new seed %d\n",istart);
		}		

		// Undo marked_blocks and marked_nets
		for (i=0; i<num_marked_blocks; i++) {
			block_visited[marked_blocks[i]] = FALSE;
		}
		for (i=0; i<num_marked_nets; i++) {
			net_visited[marked_nets[i]] = FALSE;
		}
		num_marked_nets = 0;
		num_marked_blocks = 0;

	}; // end while (istart != NO_CLUSTER)

//	printf("finished!!\n");


//	printf("Clusters pre-compact\n");
//	for (i=0; i<num_clusters; i++) {
//		printf("pre-compact Cluster %d, %d: ",i,cluster_occupancy[i]);
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%3d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//		}
//
//		printf("\n");
//	}

	// Compact the rest of the clusters
	for (i=0; i<num_clusters; i++) {
		if (cluster_occupancy[i] == 0) {
			j = i+1;
			while (cluster_occupancy[j] == 0 && j < num_clusters) {
				j++;
			}
			if (j == num_clusters) {
				break;
			} else {
				// copy cluster[j] to cluster[i];
				for (iblk=0; iblk<cluster_size; iblk++) {
					cluster_contents[i][iblk] = cluster_contents[j][iblk];
					cluster_roots[i][iblk] = cluster_roots[j][iblk];
				}
				cluster_occupancy[i] = cluster_occupancy[j];
				cluster_occupancy[j] = 0;
//				printf("\tMoving cluster %d to %d\n",j,i);
			}

		}
	}
	num_clusters = i;


	//exit(1);

	// Print out cluster contents
//	printf("pre-Final clusters\n");
//	for (i=0; i<num_clusters; i++) {
//		printf("pre-final - Cluster %d, %d: ",i,cluster_occupancy[i]);
//		for (j=0; j<cluster_occupancy[i]; j++) {
//			printf("%3d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
//		}
//
//		printf("\n");
//	}

	rooted_cluster_elimination(cluster_contents, cluster_occupancy, cluster_roots, &num_clusters, max_arrival_time, lut_size);

	// Print out cluster contents
	printf("Final clusters\n");
	for (i=0; i<num_clusters; i++) {
		printf("final - Cluster %d, %d: ",i,cluster_occupancy[i]);
		for (j=0; j<cluster_occupancy[i]; j++) {
			printf("%3d_%d ",cluster_contents[i][j],cluster_roots[i][j]);
		}

		printf("\n");
	}

	// Let's calculate some stats
	node_cnt = 0;
	for (i=0; i<num_clusters; i++) {
		node_cnt += cluster_occupancy[i];
	}

	printf("\nThere are %d total used LUTs\n",node_cnt);
	printf("There are %d total clusters\n",num_clusters);
	printf("Average LUT usage per cluster is %f \n", (float)node_cnt/(float)num_clusters);
	printf("LUT Utilization is %f \n",( (float)node_cnt/(float)num_clusters ) / cluster_size);

	*num_clusters_ptr = num_clusters;

	// Free data structures
	free(block_visited2);
	free(net_visited2);
	free(marked_nets2);
	free(marked_blocks2);
	free(unclustered_list_head);
	free(memory_pool);

}
