#include "incremental.h"
#include "util.h"
#include "vpack.h"
#include "globals.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#define TOKENS " \t\n"

struct s_incremental inc_global;

/* This function parses the congestion file and returns the congestion array. Subblocks have not been parsed */
struct s_Congestion* parse_congestion_file( char* filename )
{
	struct s_Congestion* cong;
	struct s_Congestion* current;
	struct s_Congestion* new_elem;
	struct s_clb* current_clb;
	struct s_clb* new_clb;
	FILE* fp;
	char string[100];
	char* token;
	fp = fopen( filename, "r" ); 
	cong = NULL;
	current = NULL;
	new_elem = NULL;
	current_clb = NULL;
	new_clb = NULL;

	/* Setup the global structure*/
	inc_global.num_map_elems = 0;
	inc_global.prev_x_size = 0;
	inc_global.prev_y_size = 0;

	if( fp == NULL )
		{
		printf("Unable to open congestion file, bailing out!\n");
		exit(0);
		}

	/* First thing to do is find the first .CLUSTER_SIZE */
	do{	
		if( fgets(string, 100, fp ) == NULL )
		{   // EOF 		
			break;
		}
		token = strtok( string, TOKENS );
		
		/* Parse the .CLUSTER_SIZE and .CLB */
		
		if( strcmp( token, ".CLUSTER_SIZE" ) == 0 )
		{
			/* Create a new structure to hold it */
			new_elem = (struct s_Congestion*) malloc(sizeof(struct s_Congestion));
			
			new_elem->clbs = NULL;
			new_elem->new_cluster_size = 0;
			new_elem->next = NULL;
			new_elem->num_clbs_in_type = 0;
            new_elem->residue = 0;
            new_elem->num_subblocks = 0;
			/* Get the new cluster size */
			token = strtok(NULL, TOKENS);
			new_elem->new_cluster_size = atoi( token );
			
			/* Setup link list stuff */
			if( cong == NULL )
			{
				cong = new_elem;
			}
			else
			{
				current->next = new_elem;
			}
			current = new_elem;
			current_clb = NULL;  /* Reset pointer to current CLB */
			inc_global.num_map_elems ++;
		}
		else 
		if ( strcmp( token, ".CLB") == 0 )
		{
			new_clb = (struct s_clb*) malloc( sizeof(struct s_clb) );
			new_clb->next = NULL;
			new_clb->num_subblocks = 0;
			new_clb->subblocks = NULL;
			new_clb->x = 0;
			new_clb->y = 0;
			token = strtok(NULL, TOKENS);
			strcpy( new_clb->name, token );
			
			/* Start of a new block */
			if( current_clb == NULL )
			{
				current->clbs = new_clb;  /* Current congestion type */
			}
			else
			{
				current_clb->next = new_clb; /* Link list */
			}
			current->num_clbs_in_type++;
			current_clb = new_clb;
		}
	} while( 1 );

	/* Construct the residue congestion element */
    new_elem = (struct s_Congestion*) malloc(sizeof(struct s_Congestion));
	
	new_elem->clbs = NULL;
	new_elem->new_cluster_size = 0;
	new_elem->next = NULL;
	new_elem->num_clbs_in_type = 0;
    new_elem->residue = 1;
    new_elem->num_subblocks = 0;
	/* Setup link list stuff */
    inc_global.residue = new_elem;
	
    /* One CLB to hold all the subblocks */
    new_clb = (struct s_clb*) malloc( sizeof(struct s_clb) );
	new_clb->next = NULL;
	new_clb->num_subblocks = 0;
	new_clb->subblocks = NULL;
	new_clb->x = 0;
	new_clb->y = 0;
	strcpy( new_clb->name, "residue" );
	
	/* Start of a new block in the residue clb */
	new_elem->clbs = new_clb;  
	new_elem->num_clbs_in_type++;

    /* Cleanup */
    fclose( fp );
	inc_global.congestion_map = cong;
	return cong;
}

/* This function parses the 1st pass .net file to extract which subblocks were in the clusters before */
void parse_net_file( char* filename, struct s_Congestion* cong )
{
	FILE* fp;
	FILE* outfp;
	char string[100];
	char string2[100];
	char* token;
	int done,found;
	struct s_Congestion* cur_congestion;
	struct s_clb* cur_clb;
	struct s_subblock* new_subblock;
	struct s_subblock* cur_subblock;
	
	cur_congestion = cong;

	fp = fopen( filename, "r" );
	outfp = fopen( output_file, "w+");
	if( fp == NULL )
	{
		printf("Unable to open previous .net file for incremental parsing\n");
		exit(0);
	}

	done = 0;
	/* Fast forward until we get to the .CLB declarations */
	while( !done )
	{
		fgets( string, 100, fp );
		strcpy( string2, string );
		token = strtok( string2, TOKENS );
		if( token != NULL )
		{
			if( strcmp( token, ".clb" ) == 0 )
			{   /* We've found the token! */
				done = 1;
			}
			else
			{
				fprintf( outfp, string );   /* Straight copy the line to the output file */
			}
		}
		else
		{
			fprintf( outfp, "\n" ); /* Blank line */
		}
	}
	
	/* Now parse the whole file until the end! */
	done = 0;
	found = 0;
	strcpy( string2, string );
	while( !done )
	{
		token = strtok( string, TOKENS ); /* next token at the start of a line */
		if( token == NULL )
		{  /* NULL token, do nothing*/
		}
		else if( strcmp( token, ".clb" ) == 0 )
		{
			found = 0;
			cur_congestion = cong;  /* search from the start */
			cur_clb = NULL;
			cur_subblock = NULL;
			token = strtok( NULL, TOKENS ); /* Name of this CLB */
			while( cur_congestion != NULL && !found )
			{
				cur_clb = cur_congestion->clbs; /* search each clb */
				while( cur_clb != NULL )
				{
					if( strcmp( token, cur_clb->name ) == 0 )
					{
						found = 1;
						break;  /* exit out of inner while */
					}
					cur_clb = cur_clb->next;
				}
				cur_congestion = cur_congestion->next;  /* Next Congestion map element */
			}
		}
		else if ( strcmp( token, "subblock:" ) == 0 && found )
		{
			token = strtok( NULL, TOKENS ); /* Name of this subblock */
			new_subblock = (struct s_subblock*) malloc( sizeof( struct s_subblock ) );
			strcpy( new_subblock->name, token );
			new_subblock->next = NULL;
			new_subblock->index = -1;

			if( cur_subblock == NULL )
			{
				cur_clb->subblocks = new_subblock;
			}
			else
			{
				cur_subblock->next = new_subblock;
			}
			cur_clb->num_subblocks++;
			cur_subblock = new_subblock;
		}
		else
		{ /* Not a line we're interested in */
		}
		
		if( !found ) /* Something that we are not going to rip out, so copy away */
		{
			fprintf( outfp, string2 );
		}
		if( fgets( string, 100, fp ) == NULL )
		{
			done = 1 ; /* Done when we're EOF */
		}
		strcpy( string2, string );  /* For copying to output file */
		
	} /* while (!done) */
	fclose( fp );
	fclose( outfp );
}

/* This function traverses the congestion map and returns a pointer to the CLB if it was found, NULL
   otherwise 
*/
struct s_clb* find_CLB( char* aName )
{
	struct s_Congestion* cur_cong;
	struct s_clb*        cur_clb;

	/* Traverse the linked list */
	cur_cong = inc_global.congestion_map;
	while( cur_cong != NULL )
	{
		cur_clb = cur_cong->clbs;
		while( cur_clb != NULL )
		{
			if( strcmp( cur_clb->name, aName ) == 0 )
			{
				return cur_clb;
			}
			cur_clb = cur_clb->next;
		}
		cur_cong = cur_cong->next;
	}
	return NULL;
}
/* Parses the imd file to get previous placement locations*/
void parse_imd_file( char* imd_filename, struct s_Congestion* cong )
{
	FILE* fp;
	char* token;
	char string[100];
	struct s_clb* clb;

	fp = fopen( imd_filename, "r" );
	if( fp == NULL )
	{
		printf( "Unable to open IMD file, exiting!\n");
		exit(0);
	}

	/* Parse the keystone block */
	fgets( string, 100, fp );
	token = strtok( string, TOKENS ); /* Number of CLBs, don't care */
	token = strtok( NULL, TOKENS ); /* X size */
	inc_global.prev_x_size = atoi( token );
	token = strtok( NULL, TOKENS ); /* Y size */
	inc_global.prev_y_size = atoi( token );
	
	/* Get to the end */
	while( fgets( string, 100, fp ) )
	{
		token = strtok( string, TOKENS );
		clb = find_CLB( token );
		if( clb != NULL )
		{
			token = strtok( NULL, TOKENS ); // X loc
			clb->x = atoi( token );
			token = strtok( NULL, TOKENS ); // Y-loc
			clb->y = atoi(token );
		}
	}

	fclose( fp );
}

struct s_subblock* Find_Subblock( char* aName )
{
	struct s_subblock* cur_subblock;
	struct s_clb* cur_clb;
	struct s_Congestion* cur_congestion;

	cur_congestion = inc_global.congestion_map;
	while( cur_congestion != NULL )
	{
		cur_clb = cur_congestion->clbs;
		while( cur_clb != NULL )
		{
			cur_subblock = cur_clb->subblocks;
			while( cur_subblock != NULL )
			{
				if( strcmp( cur_subblock->name, aName ) == 0 )
				{
					return cur_subblock;
				}
				cur_subblock = cur_subblock->next;
			}
			cur_clb = cur_clb->next;
		}
		cur_congestion = cur_congestion->next;
	}
	return NULL; // Not found
}
/* Function to find the index to every subblock in TVPack data structure */
void set_subblock_index_for_congestion_map()
{
	int i;
	struct s_subblock* cur_subblock;
	struct s_Congestion* cur_congestion;
	struct s_clb* cur_clb;

	for( i=0; i<num_blocks; ++i )
	{
		cur_subblock = Find_Subblock( block[i].name );
		if( cur_subblock != NULL )
		{
			cur_subblock->index = i;
		}
	}

	/* Verification that every subblockk was found */
	cur_congestion = inc_global.congestion_map;
	while( cur_congestion != NULL )
	{
		cur_clb = cur_congestion->clbs;
		while( cur_clb != NULL )
		{
			cur_subblock = cur_clb->subblocks;
			while( cur_subblock != NULL )
			{
				if( cur_subblock->index == -1 )
				{
					printf( "A subblock that we are re-clustering does not exist in the original BLIF" );
					exit(-1);
				}
				cur_subblock = cur_subblock->next;
			}
			cur_clb = cur_clb->next;
		}
		cur_congestion = cur_congestion->next;
	}
}
/* Create the floorplan based on an IMD file and the new depopulated clustering */
void create_floorplan( char* out_fp )
{
	FILE* fp;
	int x_min,y_min,x_max,y_max,i;
	struct s_Congestion* cur_cong;
	struct s_clb* cur_clb;
	fp = fopen( out_fp, "w+" );
	if( fp == NULL )
		{
		printf( "Unable to open file to write out the floorplan!\n" );
		}
	/* Write key stone blocks */
	fprintf( fp, ".NUMELEMENTS %i\n", inc_global.num_map_elems );
	fprintf( fp, ".SIZE %i, %i\n", inc_global.prev_x_size, inc_global.prev_y_size );

	/* Go through each element of the congestion MAP and calculate the floorplan */
	cur_cong = inc_global.congestion_map;
	i=0;
	while( cur_cong != NULL )
	{
		x_min = 99999999; y_min = 99999999; x_max = -1; y_max = -1; 
		cur_clb = cur_cong->clbs;
		while( cur_clb != NULL )
		{
			/* Find if any parameters will need updating */
			if( cur_clb->x < x_min )
				x_min = cur_clb->x;
			if( cur_clb->y < y_min )
				y_min = cur_clb->y;
			if( cur_clb->x > x_max )
				x_max = cur_clb->x;
			if( cur_clb->y > y_max )
				y_max = cur_clb->y;

			cur_clb = cur_clb->next;
		}
		/* Print the bounding box */
		fprintf( fp, ".ELEMENT __dpop%i_ %i %i %i %i 1\n", i, x_min, y_min, x_max-x_min+1, y_max-y_min+1 );
		cur_cong = cur_cong->next;
		i++;
	}
	fclose( fp );
}

/* Function to free the allocated memory from the congestion structure */
void free_congestion_structure()
{
	struct s_Congestion* cur_cong;
	struct s_clb* cur_clb;
	struct s_subblock* cur_subblock;
	void* temp;

	cur_cong = inc_global.congestion_map;
	while( cur_cong != NULL )
	{
		cur_clb = cur_cong->clbs;
		while( cur_clb != NULL )
		{
			cur_subblock = cur_clb->subblocks;
			while( cur_subblock != NULL )
			{
				temp = cur_subblock;
				cur_subblock = cur_subblock->next;
				free( temp );
			}
			temp = cur_clb;
			cur_clb = cur_clb->next;
			free( temp );
		}
		temp = cur_cong;
		cur_cong = cur_cong->next;
		free( temp );
	}
}

/* Utility functions */
/* This function copies the clustering results of this round into the final result */
void copy_clustering_results( int* cluster_of_block, int* inc_cluster_of_block )
{
	int i; 
	for(i=0; i<num_blocks; ++i )
	{
		/* Only check for subblocks we care about */
		if( inc_cluster_of_block[i] != NEVER_CLUSTER &&
			cluster_of_block[i] != NEVER_CLUSTER )
		{
			/* FOR DEBUG */
			if( inc_cluster_of_block[i] != NO_CLUSTER )
			{
				printf("TRYING to copy a recluster subblock twice!!\n");
				exit(0);
			}
            /* Added for residue, we might not cluster all subblocks */
            if( cluster_of_block[i] == NO_CLUSTER )
                   continue;

            /* Copy the result over */
		    inc_cluster_of_block[i] = cluster_of_block[i]; 
		}
	}
}

/* Setup the labels for the incremental cluster of blocks, which stores how we re-clustered 
   Blocks not involved with reclustering are marked "NEVER_CLUSTER", blocks involved with re-clustering
   are marked as NO_CLUSTER 
*/
void setup_inc_cluster_of_block( int* inc_cluster_of_block )
{
	int i;
	struct s_Congestion* cur_cong;
	struct s_clb* cur_clb;
	struct s_subblock* cur_subblock;

	for( i=0; i<num_blocks; ++i )
	{
		inc_cluster_of_block[i] = NEVER_CLUSTER;
	}
	cur_cong = inc_global.congestion_map;
	while( cur_cong != NULL )
	{
		cur_clb = cur_cong->clbs;
		while( cur_clb != NULL )
		{
			cur_subblock = cur_clb->subblocks;
			while( cur_subblock != NULL )
			{
				/* debug code */
				if( strcmp( block[cur_subblock->index].name, cur_subblock->name ) != 0 )
				{
					printf( "Overwriting the wrong subblock for reclustering !!\n" );
					exit(0);
				}
				inc_cluster_of_block[cur_subblock->index] = NO_CLUSTER;
				cur_subblock = cur_subblock->next;
			}
			cur_clb = cur_clb->next;
		}
		cur_cong = cur_cong->next;
	}
}
/* Setup the labels for which subblocks we want to recluster this round 
   Blocks not involved with this congestion layer are marked as "NEVER_CLUSTER"
   Blocks involved with this congestion layer are marked as "NO_CLUSTER" 
*/
void setup_cluster_of_block( int* cluster_of_block, struct s_Congestion* cur_cong )
{
	int i;
	struct s_clb* cur_clb;
	struct s_subblock* cur_subblock;

	for( i=0; i<num_blocks; ++i )
	{
		cluster_of_block[i] = NEVER_CLUSTER; /* All blocks initially set to never cluster */
	}

	cur_clb = cur_cong->clbs;
	while( cur_clb != NULL )
	{
		cur_subblock = cur_clb->subblocks;
		while( cur_subblock != NULL )
		{
			/* debug code */
			if( strcmp( block[cur_subblock->index].name, cur_subblock->name ) != 0 )
			{
				printf( "Overwriting the wrong subblock for reclustering !!\n" );
				exit(0);
			}
			cluster_of_block[cur_subblock->index] = NO_CLUSTER;
			cur_subblock = cur_subblock->next;
		}
		cur_clb = cur_clb->next;
	}
}

/* This function copies the consolidated re-clustering results back into cluster of block */
void copy_final_cluster_of_block_results( int* cluster_of_block, int* inc_cluster_of_block )
{
	int i;
	for( i=0; i<num_blocks; ++i )
	{
		if( inc_cluster_of_block[i] == NEVER_CLUSTER )
			cluster_of_block[i] = NOT_RECLUSTERED;
		else
			cluster_of_block[i] = inc_cluster_of_block[i];
	}
}

void copy_residue_luts(struct s_Congestion* cong, int* cluster_of_block, int cluster_size )
{
    int i; 
    struct s_subblock* cur_subblock;
    struct s_subblock*  new_subblock;
    struct s_clb* cur_clb;

    cur_clb = cong->clbs;
    cur_subblock = cur_clb->subblocks;
    if( cur_subblock != NULL )
    {
        while( cur_subblock->next != NULL )
            cur_subblock = cur_subblock->next;  /* Fast forward to the end of the list*/
    }

    /* Loop through the subblocks that still need to be reclustered and add it to the free list */
    for( i=0; i<num_blocks; ++i )
    {
        if( cluster_of_block[i] == NO_CLUSTER )
        {
        new_subblock = (struct s_subblock*) malloc( sizeof( struct s_subblock) );
        new_subblock->index = i;
        new_subblock->next = NULL;
        strcpy( new_subblock->name, block[i].name );
        cong ->num_subblocks ++;
        cur_clb->num_subblocks ++;
        cong->new_cluster_size += cluster_size;

        /* Setup linked list */
        if( cur_subblock == NULL )
            cur_clb->subblocks = new_subblock;
        else
            cur_subblock->next = new_subblock;

        /* Move the linked list pointer*/
        cur_subblock = new_subblock;
        }
    }
}
