/*
// Functions specific to the Incremental Placement Flow
// By David Leong, October 2005
//
*/
#define BUFSIZE 300

struct s_subblock
{
	char name[BUFSIZE];
	struct s_subblock* next;  /* Make this a Linked list */
	int index;
};

struct s_clb
{
	char name[BUFSIZE]; /* Name for this clb */
	int x;         /* Previously placed X and Y location*/
	int y; 
	struct s_subblock* subblocks; /* Subblocks that make up this CLB */
	int num_subblocks;     /* The number of subblocks this CLB has*/

	struct s_clb* next;  /* Linked list */
};

struct s_Congestion
{
	int num_clbs_in_type; /* The number of CLBs in this type */
	int new_cluster_size; /* The new cluster Size */
	struct s_clb* clbs; /* List of CLBs */
	
	struct s_Congestion* next;

	int cluster_base;  /* Cluster number that we need to append name to */
	int cluster_end;

    int num_subblocks; /* Total number of subblocks in this congestion map */
    int residue;               /* If this guy was a residue block */
};

struct s_incremental
{
	struct s_Congestion* congestion_map;
    struct s_Congestion* residue;

	int prev_x_size; 
	int prev_y_size;
	int num_map_elems;
};
extern struct s_incremental inc_global;

/* Functions  Specific to congestion map */
struct s_Congestion* parse_congestion_file( char* filename );
void parse_net_file( char* filename, struct s_Congestion* cong );
void parse_imd_file( char* imd_filename, struct s_Congestion* cong );
void set_subblock_index_for_congestion_map();
void create_floorplan( char* out_fp );
void free_congestion_structure();

/* Utility functions for clustering */
void copy_clustering_results( int* cluster_of_block, int* inc_cluster_of_block );
void setup_inc_cluster_of_block( int* inc_cluster_of_block );
void setup_cluster_of_block( int* cluster_of_block, struct s_Congestion* cur_cong );
void copy_final_cluster_of_block_results( int* cluster_of_block, int* inc_cluster_of_block );
void copy_residue_luts(struct s_Congestion* cong, int* cluster_of_block, int cluster_size);	
